/**
 * \file Decomp.hpp
 *
 * Header file to declare classes for the geometric decomposition
 *
 *
 * \author Redouane Boumghar, Hung Cao
 *
 * \ingroup lgl
 *
 **/

#ifndef LGL_DECOMPOSITION_HPP
#define LGL_DECOMPOSITION_HPP

#include <vector>

#include <lgl/FlexGrid.hpp>
#include <lgl/Location.hpp>
#include <lgl/GeoData.hpp>
#include <lgl/Attributes.hpp>

namespace jafar { //namespace jafar
	namespace lgl { //namespace lgl


		//typedef struct DecompCluster {
		class DecompCluster { //{{{
			private:
				GeoData* raster;
				Attributes* att;
			public:
				/** Size on X axis in pixels */
				int xsize;
				/** Size on Y axis in pixels */
				int ysize;
				
				/** Constructor
				 *
				 * No attribute will be created on this one
				 */
				DecompCluster(GeoData* _rast=NULL);

				/** Constructor
				 *
				 * The creation of Attributes will be done if tlx and tly are both positive.
				 *
				 * @param _xsize Cluster number of pixels on X axis
				 * @param _ysize Cluster number of pixels on Y axis
				 * @param tlx the information source pixel top left X coordinate
				 * @param tly the information source pixel top left Y coordinate
				 */
				DecompCluster(int _xsize,int _ysize, GeoData* _rast=NULL, int tlx=-1, int tly=-1);
					
				/** Destructor */
				~DecompCluster();
					
				void set(int _xsize, int _ysize) {
					xsize=_xsize;
					ysize=_ysize;
				}
				
				bool valid() { return ((xsize != 0) && (ysize != 0)); }
				
				bool isNegative() { return ((xsize < 0) && (ysize < 0)); }

				bool operator==(const  DecompCluster &rhs) const;

				DecompCluster& operator=(const DecompCluster& rdc);

				//==================== Attributes
				Attributes* getAttributes()
				{//{{{
					return att;
				}//}}}

				/**  Checking of the given attribute has been set */
                bool hasBeenSet(attributeKey at_)
                { //{{{
					if (att == NULL) { return false; }
                    return att->hasbeenset[at_];
                } //}}}

                /** Setting certainty table for at_ attribute
                 *
                 * @param reset_ default is true as in the definition of attributes
                 */
                bool setCertainty(attributeKey at_, double value_, bool reset_=true)
                { //{{{
					if (att == NULL) { return false; }
                    att->set(at_, value_, reset_);
					return true;
                } //}}}

                double getEntropy(attributeClass _ac) const
                { //{{{
					if (att == NULL) { return -1.0; }
                    return att->infotropy[_ac];
                } //}}}

                double getCertainty(attributeKey _at) const
                { //{{{
					if (att == NULL) { return -1.0; }
                    return att->certainty[_at];
                } //}}}


		//} DecompCluster;
		}; //}}}

		/**
		* The geometric decomposition (quadtree, bintree or any rectangular-pattern based decomposition)
		*/
		class Decomp {

			public:
					
				typedef FlexGrid<DecompCluster*> DecompT;

				enum SIDE { CLUSTER_TOP_SIDE, CLUSTER_RIGHT_SIDE, CLUSTER_BOTTOM_SIDE, CLUSTER_LEFT_SIDE };
					
				Decomp(GeoData* _raster);

				~Decomp();

				int xbegin() const;
				int ybegin() const;
				int xend() const;
				int yend() const;
				int xsize() const;
				int ysize() const;

				int getNbClusters() const;

				/** FlexGrid at() function */
				DecompCluster* at(int i, int j);

				/** Create each cluster informing xsize,ysize on top left and -xsize,-ysize on bottom right */
				bool setCluster(const RasterRect& rect);
				bool setLeafCluster(const RasterRect& rect);
					
				// set the uniform cluster for the raster
				// rect.xsize must be divisible by raster.xsize rect.ysize must be divisible by raster.ysize
				bool setUniformClusters(const RasterRect& rect);
					
				/**
				 * Get all input/output navigation points of this cluster represented by it root cell (i,j)
				 * in this politic, the robot can navigate only from the edges of the clusters
				 * To determine the IN and OUT points/locations, we use the center of edges of adjacent clusters
				 *
				 * The points found are expressed in navigation coordinates (1
				 * in decomposition frame is 2 in navigation frame) making the
				 * edges' centers "countable".
				 */
				bool getInOutPoints(const RasterCellIndex& root_cell, std::vector<NavLocation>& inout) const;

				bool consistancyCheck() const;
					
			private :
					
				// the raster ..
				GeoData* raster;
					
				DecompT decomp;

				/** Set to NULL all the cells inside of an object passed in the FlexGrid template
				 * for us it is a DecompCluster*
				 *
				 * @param rect  RasterRect object to indicate coordinates and size of the cluster
				 */
				bool clearCluster(const RasterRect& rect);

				/** Clear the whole Decomposition */
				bool clear();
						
				friend class Decomp_iterator;
				friend class NavGraph;
					
		};

		std::ostream& operator<<(std::ostream& out, const Decomp& decomp);

		/**
		 * Iterate over all clusters of the decomposition
		 * by iterating over grid cell from (xbegin,ybegin),(xend-1,yend-1)
		 */
		class Decomp_iterator 
		{
			private:
				Decomp* data;
				int i,j;
					

				void increment();
				void incrementNextCluster();

			public:
				Decomp_iterator(const Decomp* _data);

				Decomp_iterator(const Decomp_iterator& _it);

				void begin();
					
				bool at_end() const;
				bool at_begin() const;

				// return the value for the iterator
				DecompCluster* operator*() const;
					
				// prefix increment the iterator pointer to the next cluster
				Decomp_iterator& operator++();

				// postfix increment the iterator pointer to the next cluster
				Decomp_iterator operator++(int);

				void getIndex(int& _i, int& _j) const;

		};


	}
}

#endif /* LGL_DECOMPOSITION_HPP */
