/**
 * \file LGLPursuitViewer.hpp
 *
 * Header file to declare Qt drawing objects and functions 
 * Used as a small IHM for the pursuit Module.
 *
 * \author Cyril Robin
 *
 * \ingroup lgl
 *
 **/

/* $Id: lglViewer.hpp 4094 2009-09-04 09:32:58Z hcao $ */

#ifdef HAVE_MODULE_QDISPLAY // lgl viewer is on if HAVE_MODULE_QDISPLAY is defined

#ifndef LGL_DISPLAY_PURSUIT_HPP
#define LGL_DISPLAY_PURSUIT_HPP

#include <qdisplay/init.hpp>
#include <qdisplay/Viewer.hpp>
#include <qdisplay/ImageView.hpp>
#include <qdisplay/Shape.hpp>

#include <lgl/PursuitGraph.hpp>
#include <lgl/PursuitStats.hpp>

#include <ostream>
#include <string>
#include <vector>

namespace jafar {
	namespace lgl {


		class LGLPursuitViewer ;

		/*  specific viewer, herited from qdisplay */
		class PursuitController : public jafar::qdisplay::Viewer 
		{///{{{
			
			private :
				LGLPursuitViewer *master ;
			
			protected : 
				void setMaster( LGLPursuitViewer &_master);

			signals:
				void keyPressEvent(QKeyEvent *event);

			friend class LGLPursuitViewer ;

		};///}}}

		class PursuitViewer : public jafar::qdisplay::Viewer 
		{///{{{

			private :
				double scale ;
			
			public :
				PursuitViewer(double _scale) :
					scale(_scale)
				{ }
			
			protected :
				void mouseMoveEvent ( QMouseEvent * event );

			friend class LGLPursuitViewer ;

		};///}}}

		struct Protagonist
		{///{{{
			//position
			int x ;
			int y ;

			//display
			jafar::qdisplay::Shape *shape1;
			jafar::qdisplay::Shape *shape2;
			jafar::qdisplay::Shape *shape3;
		};///}}}

		
		class LGLPursuitViewer 
		{///{{{
			public:
				
				// Constructors and Destructor 
				LGLPursuitViewer( PursuitGraph &_PG);
				
				~LGLPursuitViewer();

				// Set functions
				void setRobotAccess	( std::string _robotAccessPath	);
				void setTargetAccess( std::string _targetAccessPath	);
				void setRobotVisib	( std::string _robotVisibPath	);

				void setRobotPosition	( int _x, int _y);
				void setTargetPosition	( int _x, int _y);

				void setHorizon	( unsigned int _horizon );
				
				/* start the game 
				 *
				 * return false if encounter any problem while launching the game.
				 **/
				bool start () ;

				/* Reaction to target Movement (+ _x, + _y )
				 */
				void move ( int _x, int _y) ;

				/* Display a given Path on the robot accessibility window 
				 */
				void displayPath( const std::vector<Cell_ID> &path ) ;

				/* Display a given Path to a given set on the robot accessibility window
				 */
				void displayPath2Set( const std::vector<Cell_ID> &path , const std::set < Cell_ID> &set) ;


			private:

				/** Pursuit Graph
				 */
				PursuitGraph *myPG ;

				/* Controler */
				jafar::lgl::PursuitController* ctrl ;

				/** QDisplay viewer used to draw shapes */
				PursuitViewer* viewer1;
				PursuitViewer* viewer2;
				PursuitViewer* viewer3;

				/** Path for bitmap files 
				 */
				std::string robotAccessPath 	;
				std::string targetAccessPath	;
				std::string robotVisibPath 	;

				/* Position of the robot and the target */
				Protagonist robot ;
				Protagonist target ;

				/* temporal horizon used in the PursuitGraph */
				unsigned int horizon;

				/* Scale for display */
				double scale ;

				/* risk view */
				std::vector< jafar::qdisplay::Shape *> risks ;

				/* past positions view */
				std::vector< jafar::qdisplay::Shape *> viewTargetPastPositions ;
				std::vector< jafar::qdisplay::Shape *> viewRobotPastPositions ;

				/* enable / disable the display of the risks
				 */
				bool visibleRisks;

				/* enable / disable the display of the past positions
				 */
				bool visiblePastPositions;

				/* statistics 
				 */
				PursuitStats stats ;
				
				/* Target and Robot past positions
				 */
				std::vector< std::pair<int, int > > targetPastPositions ;
				std::vector< std::pair<int, int > > robotPastPositions ;

				/* start the display 
				 *
				 * return false if data (file path) are missing, true otherwise.
				 **/
				bool startDisplay () ;

				/* Refresh the display (especially the protagonist position )
				 */
				bool refresh();

				/* Display the risk positions
				 */
				void displayrisks();

				/* Hide the risk positions
				 */
				void hiderisks();

				/* Display the target and robot past positions
				 */
				void displaypastpositions();

				/* Hide the target and robot past positions
				 */
				void hidepastpositions();

				/* Display the statistics
				 */
				void displaystats();

				/* Display the "help"
				 */
				void displayhelp();

			friend class PursuitController ;

		};///}}}

	}
}

#endif // LGL_DISPLAY_PURSUIT_HPP
#endif // QDisplay enabled 
