/**
 * \file NavGraph.hpp
 *
 * Header file declaring the navigation graph object and functions
 *
 *
 * \author Redouane Boumghar, Hung Cao
 *
 * \ingroup lgl
 *
 **/

#ifndef LGL_NAVIGATION_GRAPH_HPP
#define LGL_NAVIGATION_GRAPH_HPP

#include <vector>
#include <math.h>

#include <lgl/GeoData.hpp>
#include <lgl/Decomp.hpp>
#include <lgl/Quadtree.hpp>
#include <lgl/PositionManager.hpp>
#include <lgl/CostManager.hpp>
#include <lgl/GraphPath.hpp>
#include <lgl/Graphs.hpp>
#include <lgl/NavGEdge.hpp>
#include <lgl/NavGNode.hpp>
#include <lgl/NavAStar.hpp>

#include <boost/noncopyable.hpp>

using namespace boost;

#ifndef PRUNE_NAVGRAPH

#define PRUNING_ENABLED true
#define PRUNING_DISABLED false

#define DEFAULT_PRUNING_THRESHOLD 0.5

#endif

namespace jafar { // namespace jafar
  namespace lgl { // namespace lgl

		/** Navigation path element. It is a a position with its associated attributes.
		 *
		 * These attributes represents the state of the previous edge that was
		 * traversed to reach this position.
		 */
		typedef std::pair<NavLocation, Attributes> NavVertex;

		/** The Navigation path is a std::list of navigation elements. */
		typedef GraphPath<NavVertex> NavPath;

	/**
	 * NavGraph build and manages the graph over a raster
	 */
	class NavGraph : public boost::noncopyable {

		/** Cost type definition */
		typedef double COST_TYPE;

	public:

		enum verbose_tag { 
			VERBOSE_SILENT=0,
			VERBOSE_ERRORS,
			VERBOSE_WARNINGS,
			VERBOSE_INFO,
			VERBOSE_DEBUG
		};

		/** Tell the nav graph what to print out */
		verbose_tag verbose;

		/** The navigation graph container with edge, node and cost definitions */
		typedef NavGraphT<NavGEdge,NavGNode,COST_TYPE> NavGraphContainer;

		/** A* star search path planner definition */
		typedef NavAStar<NavGraphContainer::GraphContainer,NavGraphContainer::Vertex,COST_TYPE> AStarPlanner;

		/** Frontier element */
		typedef std::pair<NavGraphContainer::Vertex, double> frontierPoint;

		/** Default Constructor */
		NavGraph();

		/** Constructor
		 *
		 * @param _posManager  The position manager only depends on the geo data and the quad tree so they are not mandatory.
		 * @param _costManager The cost manager only depends on the geo data and the quad tree so they are not mandatory.
		 * @param _quadtree    The quadtree is not a mandatory parameter as it only depends on the geodata
		 *
		 *
		 */
		NavGraph(GeoData* _geodata,
			Quadtree<double,GeoData>* _quadtree	=	NULL,
			bool			 _doNavGraphPruning	=	PRUNING_DISABLED,
			double			  _pruningThreshold	=	DEFAULT_PRUNING_THRESHOLD,
			PositionManager*		_posManager	=	NULL,
			CostManager*		   _costManager =	NULL,
			//void (*_displayHook)(const NavGraphContainer::Vertex&, int)		=	NULL),
			boost::function<void(const NavGNode& , int )>	displayHook		=	NULL
			);

		/** Destructor */
		~NavGraph();

		/** Return the pointer on the graph container usable in boost macros */
		NavGraphContainer* getGraph();

		/** Return the quadtree pointer */
		Quadtree<double,GeoData>* getQuadtree();

		/** Simply clears the graph and recalls buildGraph
		 * @see buildGraph()
		 */
		bool rebuildGraph();

		/** Calculates path with AStar according to the Edge::weight
		 *
		 * @param locstart Start location
		 * @param locend   Goal location
		 * @param navpath  Resulting navigation path
		 */
		bool findPath(const NavLocation& locstart, const NavLocation& locend, NavPath& navpath);

		/** Fast find path without any other calculus 
		 *
		 * @see findPath
		 */
		bool fastFindPath(const NavLocation& locstart, const NavLocation& locend, NavPath& navpath);

		/** Calculates path with AStar according to the Edge::weight and returns path cost */
		bool findPath(const NavLocation& locstart, const NavLocation& locend, NavPath& navpath, double& pCost);

		/** Fast find path without any other calculus. And set the path cost.
		 *
		 * @see findPath
		 */
		bool fastFindPath(const NavLocation& locstart, const NavLocation& locend, NavPath& navpath, double& pCost);

		/** Calculates path with AStar according to the Edge::gru
		 *
		 * @param start Starting location
		 * @param end   Goal location
		 * @param navpath Resulting path object.
		 * @param authDist Adviced distance for the wanted path. Usually
		 * used as the authorized distance before a meeting point. It sets
		 * a limit in time according to the space-time meeting and the
		 * motion capabilities of the vehicule requiring the path.
		 */
		bool findPathOnUtility(const NavLocation& start, const NavLocation& end, NavPath& navpath, double authDist = 0.0);

		/** Selects corresponding node at location and returns navigation cost */
		double navCostAtNode(const NavLocation& _loc);

		/** Caculates navigation cost */
		double navCostAtNode(const NavGraphContainer::Vertex& vn);

		/** Get the nominal path overall cost */
		double getPathCost();

		/** Get the cooperative path overall cost */
		double getCoopPathCost();

		/** Get the number of vertices in the graph */
		int getNumberOfVertices();

		/** Get the number of edges in the graph */
		int getNumberOfEdges();

		/** Get a copy of the corresponding node at the given navigation location
		 *
		 * This method should not be used if your goal is to modify the
		 * graph contents. Rather than this you should use:
		 * vertex = graph->getVertex(NavLocation)
		 * and use the vertex (index) directly with the graph container:
		 * (*(graph->getGCtn()))[vertex]
		 */
		void getNavGNodeFromNavLocation(NavGNode& nnode, const NavLocation& v);

		/** Returns the decomposition cluster that has the greatest information entropy
		 * in the direct neighborhood around vertex A It also gives a pointer to
		 * the cluster in order to retreive its attribute and gives its position (i,j) in
		 * term of raster cell index.
		 *
		 */
		void getOffClusterFromVertex(const NavGraphContainer::Vertex& A, DecompCluster** dcluster, int& i, int& j);

		/** Few lines to get the navigation location from a graph vertex */
		NavLocation getNavLocFromVertex(const NavGraphContainer::Vertex& A);

		/** Return true of vertex a and b are spatially close
		 * according to "closeDistance" object variable.
		 */
		bool areClose( const NavGraphContainer::Vertex& a, const NavGraphContainer::Vertex& b);

		/** Returns the euclidean distance between two vertices */
		double euclideanDistance( const NavGraphContainer::Vertex& a, const NavGraphContainer::Vertex& b);

		/** Output binary ppm showing the beautiful path alternative field
		 */
		bool saveAnyField(const std::string& outfile, attributeKey _at);

		/** Reset all graph nodes potential (Goal Reaching Potential) */
		bool resetGoalReachingPotential();

		/** Calculates navigation utility from entropy and GRP */
		bool updateNavigationUtility();

		/** Return the sparsly interpolated grid of path alternative potential */
		double* getPotential();

		/** Set the parameter to enable the navigation utility calculation
		 *
		 * @see calculate_navutil
		 */
		void calculateNavigationUtility(bool bsetter);

		/** Set the parameter to enable GRP calculation
		 *
		 * @see calculate_grp
		 */
		void calculateGoalReachingPotential(bool bsetter);

		/** Dis/Enable GRP to be transported to the Geodata
		 *
		 * it costs time but it is very useful for visualisation
		 *
		 * @see set_geodata_grp
		 */
		void setGeodataPotential(bool bsetter);

		/** Dis/Enable GRU to be transported to the Geodata
		 *
		 * it costs time but it is very useful for visualisation
		 *
		 * @see set_geodata_gru
		 */
		void setGeodataUtility(bool bsetter);

		/** Function to precalculate terrain classes cost factors*
		 *
		 * These costs are vehicule-dependant as they represent the rate
		 * differences between motion speed of a robot on different types
		 * of terrain.
		 */
		void calculateTerrainFactor();

		/** Add a node to the graph
		 *
		 * The connexion is assured within its cluster (or clusters if on a
		 * cluster's border) by the creation of all edges to its neighbor
		 * nodes.
		 * - Does not check if there is already a node at this location.
		 * - Equal weights are set on both edge-directions.
		 *
		 * @param atThisLoc  Navigation location at which the node will be added
		 * @param vloc will be set as the corresponding graph vertex
		 */
		void addNode(const NavLocation& atThisLoc,  NavGraphContainer::Vertex& vloc);

		/** Check if this node is on the frontier between known and unknown environment
		 *
		 * The frontier nodes should have at least one high entropy connexion.
		 */
		bool isOnFrontier(NavGraphContainer::Vertex& vloc);

		/** Sort-Insert a new vertex in the frontier point list
		 *
		 * @param vfrontier Vertex to add
		 * @param vcost its associated navigation cost (after path planning)
		 * @return false if vcost is the highest meaning vfrontier has been added at the end of the list.
		 */
		bool extendFrontiers(NavGraphContainer::Vertex& vfrontier, double vcost);

		/** Enable or disabled the pruning of obstacle vertices and edges in the navigation graph
		 *
		 * @param _doNavGraphPruning PRUNING_ENABLED (=true) or PRUNING_DISABLED
		 */
		void setNavGraphPruningOption(bool _doNavGraphPruning) ;

		/** Tell if the pruning option is enabled or not
		 *
		 * Tell if the obstacle pruning option while building the navigation graph
		 * is enabled (true) or not (false).
		 */
		bool getNavGraphPruningOption();

		/** Enable or disabled the pruning of paths on pruned zones
		 *
		 * @param _doPathPruning true of false
		 */
		void setPathPruningOption(bool _doPathPruning) ;
	
		/** Tell if the path pruning option is enabled or not
		 *
		 * 	True if enabled.
		 */
		bool getPathPruningOption();

		/** Change the threshold value for the obstacle pruning option
		 *
		 * @param Pruning  DEFAULT_PRUNING_THRESHOLD = 0.5
		 */
		void setPruningThreshold(double _pruningThreshold) ;

		/** Give the current pruningThreshold value
		 */
		double getPruningThreshold();

		/** Returns the point to the position manager so it can be used outside */
		PositionManager* getPositionManager();

		/** Returns the current list of frontier points 
		 *
		 * A frontier point is defined as a graph node at the interface between
		 * well-known and poorly-known lands.
		 *
		 */
		std::list<frontierPoint> getFrontiers();

		/** Return the navigation success potential NSP of the corresponding vertex
		 *
		 * @param v Vertex on which we want the NSP
		 * @return the NSP of vertex v
		 */
		double getNSP(const NavGraphContainer::Vertex& v);

		/** Return calculate_grp */
		bool isCalculateGRP();

		/** Return set_geodata_grp */
		bool isSetGeodataPotential();

	protected:

		// the raster data representing the terrain
		GeoData* geodata;

		/** Quadtree associated to the geodata */
		Quadtree<double,GeoData>* quadtree;

		/** If true, the navigation graph will prune the obstacle edges and vertices.
		 * PRUNING_ENABLED = true
		 * PRUNING_DISABLED = false
		 */
		bool doNavGraphPruning ;

		/** Threshold while pruning obstacles in the navigation graph
		 *  a DEFAULT_PRUNING_THRESHOLD value is fixed (0.5)
		 *  NB : between 0 and 1 ; a high value means less tolerance
		 */
		double pruningThreshold ;

		/** Tells the path finder to cut path search to the closest position
		 * while looking for path between undisconnected positions (when start
		 * or goal is in a pruned zone).
		 *
		 * By default it is set as the same as doNavGraphPruning. If false the
		 * path won't be calculated at all if one of the position is in a
		 * pruned zone.
		 *
		 * @see doNavGraphPruning
		 */
		bool doPathPruning;

		/** Associated position manager */
		PositionManager* posManager;

		/** Associated cost manager */
		CostManager* costManager;

		/** THe graph containing all edges and vertices built upon the quadtree decomposition of the geodata */
		NavGraphContainer* graph;

		/** Planner to find path in the graph thanx to A Star. */
		AStarPlanner* planner;

		/** Outside viewer which some functions are used to display what's happening in boost astar */
		boost::function<void(const NavGNode& , int )> displayHook;

		void setDisplayHookFunction(boost::function<void(const NavGNode&, int)>_displayHook );

	private:

		/** Tag to tell if quadtree is auto produced */
		bool m_quadtree_madeinhere;

		/** Tag to tell if cost manager has been auto produced */
		bool m_costmanager_madeinhere;
		/** Tag to tell if the position manager has been auto produced */
		bool m_posmanager_madeinhere;

		/** Distance to determine that a point is close or not from the goal */
		double m_closeDistance;

		/**
		 * Build graph over the terrain using the quadtree/decomp information
		 *
		 */
		bool buildGraph();

		/** If True NavGraph will update geodata internal maps with the goal reaching utility.
		 *
		 * @see set_geodata_grp
		 */
		bool set_geodata_gru;

		/** If True NavGraph will update geodata internal maps using the goal reaching potential value.
		 *
		 * @see set_geodata_gru
		 */
		bool set_geodata_grp;

		/** If True the Path utilitywill be calculated after each path planning.  */
		bool calculate_navutil;

		/** If True the Goal Reaching Potential will be calculated after each path planning.  */
		bool calculate_grp;

		/** Terrain classes weight for navigation cost calculation
		 * RElation between speeds of operations
		 */
		double terrainFactor[((T_OBSTACLE-T_FREE)+1)];

		/** Resulting path cost */
		double pathCost;

		/** Resultingi cooperation path cost */
		double coopPathCost;

		/** Odered list of frontier points and their associated cost
		 *
		 * Lowest costs on the front() and highest cost on the back().
		 */
		std::list<frontierPoint> frontiers;


	};

}
}

#endif
