/**
 * \file PursuitGNode.hpp
 *
 * Header file for the pursuit graph node class
 * 
 *
 * \author Cyril Robin
 *
 * \ingroup lgl
 *
 **/

#ifndef LGL_PURSUIT_GRAPH_NODE_HPP
#define LGL_PURSUIT_GRAPH_NODE_HPP

#include <lgl/CellID.hpp>

#include <set>
#include <vector>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>



namespace jafar {
	namespace lgl {
		
		typedef enum {	OK,								// Nothing special, situation is under control
						RISK,							// The target is unseen, but still managed
						DANGER_HIDDEN,					// The target is hiding from the robot (enter not visible enough area)
						DANGER_LOST,					// Too much potential position for the target : it overwhelms the robot capacities ;
						            					// or the robot cannot re-catch the target in a finite time (inaccessible area, etc.)
						DANGER_UNSEEN,					// The risk term reached its maximal value (the target has been unseen for too much time)
						DANGER_DISAPPEARED				// The target "disappeared" (it's probable that the target model does not fit the real target)
			} RISK_LEVEL;
		/**
		 * PursuitGNode is the class for representing each node in the pursuit graph.
		 * Each node corresponds to a cell of the environment (Cell_ID) (which is the robot position) 
		 * and a set of cells (Cell_ID) which are the target potential positions. 
		 */
		class PursuitGNode {
 		///{{{ 
		
			private :
				std::vector< boost::weak_ptr<PursuitGNode> > children;

			public : 
  				/** Robot position */
				Cell_ID robotPos;

				/** Risk node */
				RISK_LEVEL risk;
				
  				/** Target potential positions */
				std::set<Cell_ID> targetPos;

  				/** Time associated with the node */
				unsigned int time;
		
  				/** Time associated with the risk (term of the risk) */
				unsigned int riskTerm;
		
				/** default constructor */
				PursuitGNode() :
					robotPos (0),
					risk (OK),
					time(0),
					riskTerm(0)
				{ }
			
				/**
				 * @param robotPos  position of the robot (Cell_ID)
				 * @param risk		is the node a risk node  
				 */
				PursuitGNode( Cell_ID _robotPos, RISK_LEVEL _risk, unsigned long int _time=0, unsigned long int _riskTerm=0 ) :
					robotPos(_robotPos),
					risk(_risk),
					time(_time),
					riskTerm(_riskTerm)
				{ }

				/**
				 * @param robotPos  position of the robot (Cell_ID)
				 * @param risk		is the node a risk node  
				 * @param targetPos	potential positions of the target (std::set<Cell_ID>)
				 */
				PursuitGNode(Cell_ID _robotPos, RISK_LEVEL _risk, std::set<Cell_ID> _targetPos, unsigned long int _time=0, unsigned long int _riskTerm=0 ) :
					robotPos(_robotPos),
					risk(_risk),
					targetPos(_targetPos),
					time(_time),
					riskTerm(_riskTerm)
				{ }
				
				~PursuitGNode() { }

				/** change the children of the node
				 */
				void setChildren ( const std::vector< boost::shared_ptr<PursuitGNode> > &newChildren ) ;

				/** erase the children vector 
				 */
				void noMoreChildren() ;

				/** add a new child to the node (ie a link to a brand new node)
				 */
				void addChild ( const boost::shared_ptr<PursuitGNode> & newChild ) ;

				void addChild ( const boost::weak_ptr<PursuitGNode> & newChild ) ;

				/** get the children of the node */
				const std::vector< boost::weak_ptr<PursuitGNode> > & getChildren () const ; 
				
				/** operator == 
				 * /!\ does not take into account the "time" parameter...
				 */
				bool operator== (const PursuitGNode &rhs) const ; 

				/** test equality 
				 * (total equility, including time
				 * */
				bool equal (const PursuitGNode &rhs) const ; 

		};///}}}

		/* comparator using temporal order */ 
		bool before( const boost::weak_ptr<PursuitGNode> &lhs, const boost::weak_ptr<PursuitGNode> &rhs) ; 
	
		/* output stream for PursuitGNode */
		std::ostream& operator<<(std::ostream& out, PursuitGNode const &rhs);

	}
}

#endif /* LGL_PURSUIT_GRAPH_NODE_HPP */
