/**
 * \file Decomp.cpp
 *
 * 
 *
 * \author Redouane Boumghar, Hung Cao
 *
 * \ingroup lgl
 *
 **/

#include <lgl/Decomp.hpp>

#include <ostream>
#include <vector>

using namespace std;

namespace jafar {
	namespace lgl {
	 
		/**
		* Decomp_Iterator methods
		*/

		Decomp_iterator::Decomp_iterator(const Decomp* _data):
		data((Decomp*)_data) 
		{
			// Initial position
			begin();
		}

		Decomp_iterator::Decomp_iterator(const Decomp_iterator& _it):
		data(_it.data),
		i(_it.i),
		j(_it.j)
		{
			// nothing more to do
		}

		void Decomp_iterator::begin() {
			i=data->xbegin();
			j=data->ybegin();
			if (data->at(i,j) == NULL)
			{
				incrementNextCluster();  
			}
		}

		void Decomp_iterator::increment() {

			//  any of cells with i==xend is the end cell
			if (at_end())
			{ return; }

			// increment i,j														
			//(j=0 i=0) iiiii +>
			//    j
			//    j
			//   V+V
			i++;
			if (i >= data->xend()) {
				i = data->xbegin();
				j++;
			}

			incrementNextCluster();

		}

		void Decomp_iterator::incrementNextCluster() {/*{{{*/
			DecompCluster* curCluster = NULL;

			// increment i and j until reaching the end or reaching a cluster (cell not NULL)
			while (j < data->yend() ) { // && i < data->xend()) {
				curCluster = data->at(i,j);
				if ((curCluster!=NULL)&&(!curCluster->isNegative()))
				{ 
					// i and j are positionned on the right cluster
					return;
				}

				// increment i,j														
				i++;
				if (i >= data->xend()) {
					i = data->xbegin();
					j++;
				}
			}

		}  /*}}}*/


	 bool Decomp_iterator::at_end() const {
		// test if the cursor is at the end of the grid
		if ( i >= data->xend() || j >= data->yend() )
		  return true;
		else
		  return false;
	 }

	 bool Decomp_iterator::at_begin() const {/*{{{*/
		if ((i==data->xbegin())&&(j==data->ybegin()))
		  return true;
		else
		  return false;
	 }/*}}}*/

	 // return the value
	 DecompCluster* Decomp_iterator::operator*() const {/*{{{*/
		return data->at(i,j);
	 }/*}}}*/

	 Decomp_iterator& Decomp_iterator::operator++() {/*{{{*/
		increment();
		return (*this);
	 }/*}}}*/

	 Decomp_iterator Decomp_iterator::operator++(int) {
		//  std::cout <<"operator ++ i= " << i << "," << j << std::endl;
		Decomp_iterator tmp(*this);
		increment();
		return (tmp);
	 }

	 void Decomp_iterator::getIndex(int& _i, int& _j) const {/*{{{*/
		_i=i;
		_j=j;
	 }/*}}}*/


//----------------------------------------------------------------------------//

	 /**
	  * DecompCluster class implementation
	  */

	 DecompCluster::DecompCluster(GeoData* raster_):
		raster(raster_),
		xsize(0),
		ysize(0)
	 {
		 att=NULL;
	 }

	DecompCluster::DecompCluster(int _xsize,int _ysize, GeoData* raster_, int tlx, int tly):
		raster(raster_),
		xsize(_xsize),
		ysize(_ysize) 
	{
		if ( tlx < 0 || tly < 0)
		{
			// For DecompCluster pointing on the bottom right of raster cells
			att=NULL;
		} else {
			att = new Attributes();
			//--- Updating node's certainties
			int katt = 0;
			int kgeo = GEODATA_TFLAT_BAND;
			double geovalue = 0.0;
			for (katt = ATT_BEGIN; katt <= T_LIMIT; katt++)
			{
				geovalue = raster->getValueFromBand((GEODATA_BAND)kgeo, tlx, tly);
				if (geovalue < 0.0)
				{
					geovalue = 0.0;
				}
				setCertainty((attributeKey)katt, geovalue);
				kgeo++;
			}

			//--- Distributing certainties assuring normalization
			att->distributeCertainties();

		}
	}//}}}
	
	DecompCluster::~DecompCluster() 
	{//{{{
		delete att;
	}//}}}
	
/*
	 void DecompCluster::set(int _xsize, int _ysize) {
		xsize=_xsize;
		ysize=_ysize;
	 }

	 bool DecompCluster::valid() {
		return ((xsize!=0)&&(ysize!=0));
	 }

	 bool DecompCluster::isNegative() {
		return ((xsize<0)&&(ysize<0));
	 }
*/

	 bool DecompCluster::operator==(const  DecompCluster &rhs) const
	 {
		return (xsize == rhs.xsize) && (ysize == rhs.ysize);
	 }  

	 DecompCluster& DecompCluster::operator=(const DecompCluster& rdc) {
		this->xsize=rdc.xsize;
		this->ysize=rdc.ysize;
		return *this;
	 }

	 std::ostream& operator<< (std::ostream& out, const DecompCluster& cluster) {
		out << "(" << cluster.xsize <<":" << cluster.ysize << ")";
		return out;
	 }

//----------------------------------------------------------------------------//

	 /**
	  * Decomp class implementation
	  */
	Decomp::Decomp(GeoData* _raster):
		raster(_raster)
	{
		// resize and init with NULL value
		// Old dimensions are kept
		// if resize from xr=0, yr=0, xs=10, ys=10
		// with xr=10, yr=14, xs=10, ys=10
		// Then we will get  xr=0, yr=0, xs=20, ys=24
		decomp.resize(raster->getxroot(), raster->getxroot()+raster->getxsize(),
		              raster->getyroot(), raster->getyroot()+raster->getysize(), NULL);

		// set the main cluster
		decomp.at(raster->getxroot(),raster->getyroot()) = new DecompCluster(raster->getxsize(),raster->getysize(), raster, raster->getxroot(),raster->getyroot());
		if ((raster->getxsize()>1)||(raster->getysize()>1))
		{
			decomp.at(raster->getxsize()-1,raster->getysize()-1) = new DecompCluster(-raster->getxsize(),-raster->getysize());
		}
	}


	 Decomp::~Decomp() {
		int i = 0;
		int j = 0;
		for (i = raster->getxroot(); i < (raster->getxroot()+raster->getxsize());i++) {
			for (j = raster->getyroot(); j < (raster->getyroot()+raster->getysize());j++) {
				delete decomp.at(i,j);
			}
		}
	 }

	 int Decomp::xbegin() const {
		return decomp.xbegin();
	 }

	 int Decomp::ybegin() const {
		return decomp.ybegin();
	 }

	 int Decomp::xend() const {
		return decomp.xend();
	 }

	 int Decomp::yend() const {
		return decomp.yend();
	 }

	 int Decomp::xsize() const {
		return decomp.x_size();
	 }

	 int Decomp::ysize() const {
		return decomp.y_size();
	 }

	 DecompCluster* Decomp::at(int i, int j) {
		return decomp.at(i,j);
	 }

	 bool Decomp::clearCluster(const RasterRect& rect) { //{{{
		// clear all the cluster
		int i,j;
		for (i=rect.xtopleft;i<=rect.xbottomright;i++) {
		  for (j=rect.ytopleft;j<=rect.ybottomright;j++) {
			 if (decomp.at(i,j) != NULL)
				delete (DecompCluster*)decomp.at(i,j);
			 decomp.at(i,j)=NULL;
		  }
		}
		return true;
	 } //}}}

	 bool Decomp::clear() { //{{{
		// clear all the decomp
		int i,j;
		for (i=xbegin(); i<=xend(); i++) {
			for (j=ybegin(); j<=yend(); j++) {
				if (decomp.at(i,j) != NULL)
				{
					delete (DecompCluster*)decomp.at(i,j);
				}
				decomp.at(i,j)=NULL;
			}
		}

		// Reset the main cluster 
		decomp.at(xbegin(),ybegin())= new DecompCluster(xsize(),ysize());
		decomp.at(xend(),yend())= new DecompCluster(-xsize(),-ysize());

		return true;
	 } //}}}

	 int Decomp::getNbClusters() const { //{{{

		int result=0;
		Decomp_iterator it(this);

		for(it.begin();!it.at_end();++it)
		{
			result++;
		}

		return result;

	 } //}}}


	 bool Decomp::setCluster(const RasterRect& rect) { //{{{

		clearCluster(rect);

		// Set the root cell (top left) with information about the cluster size
		decomp.at(rect.xtopleft, rect.ytopleft) = new DecompCluster(rect.xsize(),rect.ysize(), raster, rect.xtopleft, rect.ytopleft);
		// For larger-than-a-pixel cluster inform the bottom right pixel/cell with minus size.
		if ((rect.xsize()>1)||(rect.ysize()>1))
		{
			// Set the bottom right cell with -xisze and -ysize so we know
			// where to find the root cell from here.
			decomp.at(rect.xbottomright, rect.ybottomright) = new DecompCluster(-rect.xsize(),-rect.ysize());
		}

		return true;
	 } //}}}

	 bool Decomp::setLeafCluster(const RasterRect& rect) { //{{{

		clearCluster(rect);

		// Set the root cell (top left) with information about the cluster size
		decomp.at(rect.xtopleft, rect.ytopleft) = new DecompCluster(
		    rect.xsize(),
			rect.ysize(),
			raster,
			rect.xtopleft,
			rect.ytopleft);
		// For larger-than-a-pixel cluster inform the bottom right pixel/cell with minus size.
		if ((rect.xsize()>1)||(rect.ysize()>1))
		{
			// Set the bottom right cell with -xisze and -ysize so we know
			// where to find the root cell from here.
			decomp.at(rect.xbottomright, rect.ybottomright) = new DecompCluster(-rect.xsize(),-rect.ysize());
		}

		return true;
	 } //}}}

	 bool Decomp::setUniformClusters(const RasterRect& rect) {/*{{{*/
		int clusterXSize = rect.xsize();
		int clusterYSize = rect.ysize();
		int i = 0;
		int j = 0;

		if ((xsize()%clusterXSize!=0)||(ysize()%clusterYSize!=0)) {
			return false;
		}
		clear();

		for (i=0; i < xsize(); i+=clusterXSize) {
			for (j=0; j < ysize(); j+=clusterYSize) {
				decomp.at(i,j) = new DecompCluster(clusterXSize,clusterYSize,
					raster,
					rect.xtopleft, rect.ytopleft);
				if ((clusterXSize>1)||(clusterYSize>1))
				{
					decomp.at(i+clusterXSize-1,j+clusterYSize-1) = new DecompCluster(-clusterXSize,-clusterYSize);
				}
			}
		}
		return true;
	 } /*}}}*/

	 // get the IN and OUT location for the cluster
	 bool Decomp::getInOutPoints(const RasterCellIndex& root_cell, vector<NavLocation>& inout) const 
	 { //{{{

		if (!raster->validIndex(root_cell)) {
			//#ifdef LGL_VVERBOSE
			std::cout << "#EEE# [Decomp::getInOutPoints] Invalid index" << std::endl;
			//#endif
			return false;
		}

		if (decomp.at(root_cell.i,root_cell.j)==NULL)
		{ 
			//#ifdef LGL_VVERBOSE
			std::cout << "#EEE# [Decomp::getInOutPoints] No cluster at provided indexes." << std::endl;
			//#endif
			return false;
		}

		// Prepare size parameters
		int xsize = 0;
		int ysize = 0; 
		// Clean the NavLocation List 
		inout.clear();

		// Getting the size of the cluster to look at
		xsize = decomp.at(root_cell.i,root_cell.j)->xsize;
		ysize = decomp.at(root_cell.i,root_cell.j)->ysize;

		DecompCluster* adjCluster;
		int curXPos,curYPos;
		int adjClusterXEnd, adjClusterYEnd;
		// the start and end index of the current adjacent cluster
		int jstart = 0;
		int jend   = 0;
		int istart = 0;
		int iend   = 0;
		int cur_i  = 0;
		int cur_j  = 0;

		//--- Now scanning left, right, bottom and top sides to find in and out points with adjacent clusters. ---//

		//----------- INOUT Points with LEFT adjacent clusters
		// Setting limits jstart and jend to the begin and end of the left side.
		jstart = root_cell.j;
		jend   = root_cell.j + ysize;
		// Reference in i and j for the search of adjacent clusters.
		cur_i = root_cell.i-1; 
		cur_j = root_cell.j;
		// Position in the navigation frame reference
		curXPos = root_cell.i*2;  


		// Case where the cluster is on the left side of the overall raster
		// In this case we only add the middle edge navigation point.
		// Else we look for all adjacent clusters.
		if (cur_i < raster->getxroot()) {
			// Adding the middle edge navigation point.
			inout.push_back( NavLocation(curXPos, jstart+jend) );
		} else {
			// Iterate over each adjacent cells on the left
			// Get the start and end of the frontier between the current cluster and the adjacent neighbor.
			// Take the frontier middle position as the IN-OUT point
			while (cur_j < jend)
			{
				adjCluster = decomp.at(cur_i , cur_j);			 
				if (adjCluster != NULL) {
					// Test if adjCluster is the bottom right corner of a cluster
					adjClusterYEnd = adjCluster->isNegative()?cur_j:(cur_j + adjCluster->ysize - 1);				
					cur_j = min( (jend - 1), adjClusterYEnd);				
					// j*2 because each raster cell correponds to 2x2 units in deplacement (location grid)
					// y position is computed by : (jstart*2 + (jend+1)*2)/2 = jstart+jend+1
					inout.push_back(NavLocation(curXPos, (jstart + cur_j + 1)));

					// looking for next adjacent cluster
					cur_j++;
					jstart = cur_j;
				} else {
					cur_j++;
				}
			}

			// In case we didn't find a last cluster in perfectly jointed cells
			// then we have to add a navigation point in the middle of what's left of the side-with-no-clusters
			// because we know that there is a cluster and we don't to look for its end.
			if (cur_j > jstart)
			{
				//cur_j--;
				//inout.push_back(NavLocation(curXPos, (jstart + cur_j + 1)));
				inout.push_back(NavLocation(curXPos, (jstart + cur_j )));
			}
		}

		//----------- INOUT Points with RIGHT adjacent clusters
		// Setting Indexes
		cur_i = root_cell.i+xsize;
		cur_j = root_cell.j;
		// Setting limits
		jstart = root_cell.j;
		jend   = root_cell.j + ysize;
		// Position in the navigation frame
		curXPos=(cur_i)*2;  

		// Border case
		if (cur_i >= raster->getxsize()) {
			inout.push_back(NavLocation(curXPos, (jstart + jend)));
		} else {
			// Looking for adjacent cluster on right side
			while ( cur_j < jend)
			{
				adjCluster = decomp.at(cur_i, cur_j);			 
				if (adjCluster != NULL) {
					adjClusterYEnd = adjCluster->isNegative()?cur_j:(cur_j + adjCluster->ysize - 1);				
					cur_j = min( (jend - 1), adjClusterYEnd);				
					// Adding corresponding navigation point
					inout.push_back(NavLocation(curXPos, (jstart + cur_j + 1)));

					// Looking for next adjacent cluster
					cur_j++;
					jstart = cur_j;
				} else {
					cur_j++;
				}
			}
			if (cur_j >jstart) {
				inout.push_back(NavLocation(curXPos, (jstart+cur_j)));
			}
		}

		//----------- INOUT Points with BOTTOM adjacent clusters
		// Setting limits
		istart = root_cell.i;
		iend   = root_cell.i + xsize;
		// Setting indexes
		cur_i = root_cell.i;
		cur_j = root_cell.j + ysize;
		// Y fixed Position in the navigation frame
		curYPos=(cur_j)*2;  

		if (cur_j >= (raster->getyroot() + raster->getysize())) {
			inout.push_back(NavLocation( (istart+iend), curYPos));
		} else {
			while ( cur_i < iend) {
				adjCluster = decomp.at( cur_i, cur_j);			 
				if (adjCluster != NULL) {
					adjClusterXEnd = adjCluster->isNegative()?cur_i:(cur_i + adjCluster->xsize - 1);
					cur_i = min( (iend-1), adjClusterXEnd);				
					inout.push_back(NavLocation((istart+cur_i+1), curYPos));
					cur_i++;
					istart=cur_i;
				} else {
					cur_i++;
				}
			}

			if (cur_i > istart) {
				inout.push_back(NavLocation( (istart+cur_i), curYPos));
			}
		}


		//----------- INOUT Points with TOP adjacent clusters
		// Setting limits
		istart = root_cell.i;
		iend   = root_cell.i + xsize;
		// Setting indexes
		cur_i = root_cell.i;
		cur_j = root_cell.j - 1;
		// Y fixed Position in the navigation frame
		curYPos=(root_cell.j)*2;  

		if (cur_j < raster->getyroot()) {
			inout.push_back(NavLocation( (istart+iend), curYPos));
		} else {
			// scanning for in-out points on top side
			while (cur_i < iend) {
				adjCluster=decomp.at(cur_i, cur_j);			 
				if (adjCluster!=NULL) {
					adjClusterXEnd = adjCluster->isNegative()?cur_i:(cur_i+adjCluster->xsize-1);			 
					cur_i = min( (iend-1),adjClusterXEnd);				
					inout.push_back(NavLocation( (istart+cur_i+1),curYPos));
					cur_i++;
					istart=cur_i;
				} else {
					cur_i++;
				}
			}

			// One last point in case some cells weren't associated with an adjacent cluster.
			if (cur_i > istart) {
				inout.push_back(NavLocation((istart+cur_i), curYPos));
			}
		}

		return true;

	 } //}}}

	 ostream& operator<<(ostream& out, const Decomp& decomp) { /*{{{*/ 

		Decomp_iterator it(&decomp);
		int i = 0;
		int j = 0;
		double xc = 0.0;
		double yc = 0.0;
		double xs = 0.0;
		double ys = 0.0;
		double xmin = 0.0;
		double ymin = 0.0;
		double xmax = 0.0;
		double ymax = 0.0;

		// Initialization
		it.begin();
		it.getIndex(i,j);
		xmin = (double)i;
		ymin = (double)j;
		xmax = xmin;
		ymax = ymin;

		// Printing the decomposition information
		out << endl << "# Decomposition of " <<  decomp.getNbClusters() << " clusters" << endl;;
		out << "set key off" << endl;
		for(it.begin();!it.at_end();++it) {
			// Printing the current gnuplot rectangle
			it.getIndex(i,j);
			xs = (double)((*(*it)).xsize);
			ys = (double)((*(*it)).ysize);
			xc = ((double)i) + xs/2.0;
			yc = ((double)j) + ys/2.0;
			out << "# Rectangle at (" << i << ", " << j << ") with size " << (*(*it)) << endl;
			out << "    set object rect center " << xc << ", " << -yc << " size " << xs << ", " << ys << endl;
			// Setting limits
			if ( xmin > (double)i)
			{
				xmin = (double)i;
			}

			if ( ymin > (double)j)
			{
				ymin = (double)j;
			}

			if ( xmax < ((double)i + xs))
			{
				xmax = (double)i + xs;
			}

			if ( ymax < ((double)j + ys))
			{
				ymax = (double)j + ys;
			}
		}

		out << "plot [" << xmin <<":"<< xmax <<"] ["<< -ymax <<":"<< ymin <<"] 0"<< endl;
		out << "pause -1" <<endl<<endl;


		return out;
	} /*}}}*/

	 bool Decomp::consistancyCheck() const {
		int i,j,xroot,yroot,xsize,ysize;
		Decomp_iterator it(this);		
		for(it.begin();!it.at_end();++it) {
		  //
		  it.getIndex(xroot,yroot);
		  xsize=(*it)->xsize;
		  ysize=(*it)->ysize;

		  if (decomp.at(xroot,yroot)==NULL) {
			 std::cout << "[Decomp] decomp cluster iteration problem" << std::endl;
			 return false;
		  }

		  for (i=xroot+1;i<xroot+xsize;i++) {
			 for (j=yroot+1;j<yroot+ysize;j++) {
				if ((decomp.at(i,j)!=NULL)&&((i!=xroot+xsize-1)||(j!=yroot+ysize-1)))
				  return false;
			 }
		  }

		}
		return true;
	 }

  }
}
