/**
 * \file LGLPursuitViewer.cpp
 *
 * Source file that defines a small IHM for the Pursuit Module
 *
 * \author Cyril Robin 
 *
 * \ingroup lgl
 *
 **/

#ifdef HAVE_MODULE_QDISPLAY // lgl viewer is on if HAVE_MODULE_QDISPLAY is defined

#include <lgl/LGLPursuitViewer.hpp>
#include <lgl/PursuitGNode.hpp>

#include <exception>
#include <iostream>

#include <QGraphicsScene>
#include <QGraphicsView>

#ifndef SWITCH_CONNEXITY
#define SWITCH_CONNEXITY 0
// use with "findAccessibilxwity()" functions
// 0 is for 9-connexity
// 1 is for 5-connexity ( '+' cross)
#endif

#ifndef SCALE
//#define SCALE 4.0
#define SCALE 8.0
#endif

#ifndef TIME_CONSTRAINT
#define TIME_CONSTRAINT true
#endif

#if TIME_CONSTRAINT
	#ifndef DEFAULT_TMAX
	#define DEFAULT_TMAX 800.0
	#endif
#endif

#ifndef SAVE_FILE
#define SAVE_FILE true
#include <fstream>
//TODO change this global variable..
 ofstream fout ;
#endif


using namespace jafar::qdisplay;
using namespace std;

namespace jafar {
	namespace lgl {

		/****************************************************************/
		/************************ PURSUIT_CONTROLER *********************/
		/****************************************************************/
		
		void PursuitController::keyPressEvent ( QKeyEvent * event )
		{///{{{
			
			switch (event->key()) {

				// Control of the Target {{{
				#if SWITCH_CONNEXITY

				case Qt::Key_Return:
					std::cout<<"-> You stay here !"<<std::endl;
					master->move( +0, +0 );
					break;
				case Qt::Key_Left:
					std::cout<<"-> You go left !"<<std::endl;
					master->move( -1, +0 );
					break;
				case Qt::Key_Right:
					std::cout<<"-> You go right !"<<std::endl;
					master->move( +1, +0 );
					break;
				case Qt::Key_Down:
					std::cout<<"-> You go down !"<<std::endl;
					master->move( +0, +1 );
					break;
				case Qt::Key_Up:
					std::cout<<"-> You go up !"<<std::endl;
					master->move( +0, -1 );
					break;

				#else 

				case Qt::Key_S:
					std::cout<<"-> You stay here !"<<std::endl;
					master->move( +0, +0 );
					break;
				case Qt::Key_Q:
					std::cout<<"-> You go left !"<<std::endl;
					master->move( -1, +0 );
					break;
				case Qt::Key_D:
					std::cout<<"-> You go right !"<<std::endl;
					master->move( +1, +0 );
					break;
				case Qt::Key_X:
					std::cout<<"-> You go down !"<<std::endl;
					master->move( +0, +1 );
					break;
				case Qt::Key_Z:
					std::cout<<"-> You go up !"<<std::endl;
					master->move( +0, -1 );
					break;
				case Qt::Key_A:
					std::cout<<"-> You go up and left !"<<std::endl;
					master->move( -1, -1 );
					break;
				case Qt::Key_E:
					std::cout<<"-> You go up and right !"<<std::endl;
					master->move( +1, -1 );
					break;
				case Qt::Key_W:
					std::cout<<"-> You go down and left !"<<std::endl;
					master->move( -1, +1 );
					break;
				case Qt::Key_C:
					std::cout<<"-> You go down and right !"<<std::endl;
					master->move( +1, +1 );
					break;
				#endif
				///}}}

				// Option {{{
				case Qt::Key_H:
					std::cout <<"-> Displaying help : " << std::endl ;
					master->displayhelp() ;

					break;

				case Qt::Key_I:
					std::cout <<"-> Displaying Statistics : " << std::endl ;
					master->displaystats() ;

					break;

				case Qt::Key_P:
					if ( ! master->visiblePastPositions) 
					{
						std::cout<<"-> Displaying the target and robot past positions..."<<std::endl;
						master->displaypastpositions();
					} else {
						std::cout<<"-> Hiding the target and robot past positions..."<<std::endl;
						master->hidepastpositions();
					}
					
					break;

				case Qt::Key_Space:
					if ( ! master->visibleRisks) 
					{
						std::cout<<"-> Displaying the risk positions..."<<std::endl;
						master->displayrisks();
					} else {
						std::cout<<"-> Hiding the risk positions..."<<std::endl;
						master->hiderisks();
					}
					
					break;

				case Qt::Key_Escape:
					std::cout<<"-> Exiting the game... :-("<<std::endl;

					#if SAVE_FILE
					  fout.close();
					  std::cout<<"Log file saved."<<std::endl;
					#endif

					std::cout<<"Good bye! See you soon!"<<std::endl;

					exit(0);

					break;
				default :
					std::cout<<"-> Not a valid command !!!"<<std::endl;
				///}}}
			}

			std::cout << std::endl << ">>> Next Action ? " << std::flush ;

		}///}}}

		void PursuitController::setMaster( LGLPursuitViewer &_master)
		{///{{{
			master = &_master;
		}///}}}

		/****************************************************************/
		/************************ PURSUIT_VIEWER ************************/
		/****************************************************************/

		//void PursuitViewer::MouseGraphicsScene::mouseMoveEvent ( QGraphicsSceneMouseEvent * event )
		//void PursuitViewer::mouseMoveEvent ( QGraphicsSceneMouseEvent * event )
		void PursuitViewer::mouseMoveEvent ( QMouseEvent * event )
		{///{{{

			QGraphicsView::mouseMoveEvent(event);

			setTitleWithMouseCoordinates( event->x() * 2.0 / scale, event->y() * 2.0 / scale );

		}///}}}

		/****************************************************************/
		/*********************** LGL_PURSUIT_VIEWER *********************/
		/****************************************************************/

		// Constructor
		LGLPursuitViewer::LGLPursuitViewer(PursuitGraph &_PG ):
			myPG (&_PG)
		{ //{{{
		
			scale = SCALE ;

			// viewers with screen coordinates (corner 0,0 is at the top left, x is the horizontal axis)
			viewer1 = new PursuitViewer(scale);
			viewer2 = new PursuitViewer(scale);
			viewer3 = new PursuitViewer(scale);
			
			viewer1->setTitle("Robot Accessibility");
			viewer2->setTitle("Target Accessibility");
			viewer3->setTitle("Robot Visibility");

			// controller
			ctrl 	= new jafar::lgl::PursuitController();
			ctrl->setMaster(*this);
			ctrl->grabKeyboard();

			ctrl->setTitle("Controller");

			// robot initialisation
			setRobotPosition(-1.0 * scale,-1.0*scale);

			robot.shape1 = new Shape(Shape::ShapeRectangleFilled, robot.x * scale /2, robot.y * scale /2, 1.0 * scale, 1.0 * scale);
			robot.shape2 = new Shape(Shape::ShapeRectangleFilled, robot.x * scale /2, robot.y * scale /2, 1.0 * scale, 1.0 * scale);
			robot.shape3 = new Shape(Shape::ShapeRectangleFilled, robot.x * scale /2, robot.y * scale /2, 1.0 * scale, 1.0 * scale);

			robot.shape1->setColor(0,0,250);
			robot.shape2->setColor(0,0,250);
			robot.shape3->setColor(0,250,0);

			robot.shape1->setFontSize( 1 * scale);
			robot.shape2->setFontSize( 1 * scale);
			robot.shape3->setFontSize( 1 * scale);

			robot.shape1->setLabel("R1");
			robot.shape2->setLabel("R1");
			robot.shape3->setLabel("R1");

			robot.shape1->setFontColor(0,0,250);
			robot.shape2->setFontColor(0,0,250);
			robot.shape3->setFontColor(0,250,0);


			// target initialisation
			setTargetPosition(-1.0 * scale,-1.0*scale);

			target.shape1 = new Shape(Shape::ShapeEllipseFilled, target.x * scale /2, target.y * scale /2, 1.0 * scale, 1.0 * scale);
			target.shape2 = new Shape(Shape::ShapeEllipseFilled, target.x * scale /2, target.y * scale /2, 1.0 * scale, 1.0 * scale);
			target.shape3 = new Shape(Shape::ShapeEllipseFilled, target.x * scale /2, target.y * scale /2, 1.0 * scale, 1.0 * scale);

			target.shape1->setColor(0,0,100);
			target.shape2->setColor(0,0,100);
			target.shape3->setColor(250,250,250);

			target.shape1->setFontSize( 1 * scale);
			target.shape2->setFontSize( 1 * scale);
			target.shape3->setFontSize( 1 * scale);

			target.shape1->setLabel("T");
			target.shape2->setLabel("T");
			target.shape3->setLabel("T");

			target.shape1->setFontColor(0,0,100);
			target.shape2->setFontColor(0,0,100);
			target.shape3->setFontColor(250,250,250);

			// Horizon
			horizon = 10 ;

			//risk
			visibleRisks = false ;

		} //}}}

		LGLPursuitViewer::~LGLPursuitViewer()
		{///{{{
			delete viewer1;
			delete viewer2;
			delete viewer3;

			for (unsigned int i = 0 ; i < risks.size() ; i++ ) 
				delete risks.at(i);

		}///}}}
	
		bool LGLPursuitViewer::start() 
		{///{{{

			std::cout << "   *********************************** [LGL_PURSUIT_VIEWER] ***********************************   " << std::endl ;

			// Save the game data in a file is asked for
			// TODO Add option for name, directory, etc.
			#if SAVE_FILE

			  fout.open("./output.csv") ; 
			
			  std::cout << " /!\\ Log game data in ./output.csv..." << std::endl ;

			#endif

			bool b = true;

			timeval t;
			double d;

			gettimeofday(&t,NULL);
			d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0;

			if (myPG == NULL)
			{
				std::cerr << "No Pursuit Graph specified. Cannot run the game." << std::endl;
				return(!b);
			}
			
			std::cout << "#---III---  The robot  stands in (" << robot.x  << "," << robot.y  << ")." << std::endl;
			std::cout << "#---III---  The target stands in (" << target.x << "," << target.y << ")." << std::endl;

			// Initialization of the display
			b = startDisplay() ;
			if (!b) return(b);

			// Initialize the situation
			myPG->computeAll(robot.x, robot.y, target.x, target.y, horizon);

			refresh();

			gettimeofday(&t,NULL);
			d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d;
			
			std::cout << "#---III---  Game initialized in " << d << "ms." << std::endl;

			std::cout << "#---III---  [ temporal horizon is "<< myPG->gethorizon() 
								<< " with front = "<< myPG->getfront().size() 
								<< " - risks = " << myPG->getrisk().size() 
								<<" - danger = " << myPG->getdanger().size() 
								<< " ] " << std::endl ; 

			std::cout << std::endl;

			// statistics
			stats.update(myPG, double ( d / myPG->gethorizon() ) );

			robotPastPositions.push_back ( std::make_pair( robot.x,  robot.y  ) );
			targetPastPositions.push_back( std::make_pair( target.x, target.y ) );

			std::cout << "   *********************************** [LGL_PURSUIT_VIEWER] ***********************************   " << std::endl ;

			// Start the game
			std::cout << "   /!\\ Now, it's up to you : use keyboard to control the target ! (Press \"H\" Key for some help )" << std::endl ;

			std::cout << std::endl << ">>> Next Action ? " << std::flush ;

			return( b );

		}///}}}

		bool LGLPursuitViewer::startDisplay() 
		{///{{{
			bool b = true ;
			
			// robot accessibility 
			try {
				ImageView* imgV1 = new ImageView(robotAccessPath);
				imgV1->setPos( 0 , 0 );	// top-left
				imgV1->setScale(scale);
				viewer1->setImageView(imgV1);
			} 
			catch (const std::exception& e)
			{
				b = false ;
				std::cerr << "Couldn't add robot accessibility on the pursuit viewer." << std::endl;
				//throw;
			}

			// target accessibility 
			try {
				ImageView* imgV2 = new ImageView(targetAccessPath);
				imgV2->setPos( 0 , 0 );	// top-right
				imgV2->setScale(scale);
				viewer2->setImageView(imgV2);
			} 
			catch (const std::exception& e)
			{
				b = false ;
				std::cerr << "Couldn't add target accessibility on the pursuit viewer." << std::endl;
				//throw;
			}
			
			// robot visibility 
			try {
				ImageView* imgV3 = new ImageView(robotVisibPath);
				imgV3->setPos( 1 , 1 );	// bottow-left
				imgV3->setScale(scale);
				viewer3->setImageView(imgV3);
			} 
			catch (const std::exception& e)
			{
				b = false ;
				std::cerr << "Couldn't add target accessibility on the pursuit viewer." << std::endl;
				//throw;
			}

			viewer1->setGeometry( 0,								0,									myPG->getxsize() *  scale/2, myPG->getysize() * scale/2 );
			viewer2->setGeometry( 0,								myPG->getysize() * scale/2 + 70,	myPG->getxsize() *  scale/2, myPG->getysize() * scale/2 );
			viewer3->setGeometry( myPG->getxsize() * scale/2 + 20,	0,									myPG->getxsize() *  scale/2, myPG->getysize() * scale/2 );

			ctrl->setGeometry(    myPG->getxsize() * scale/2 + 20, 	myPG->getysize() * scale/2 + 70, 200, 100);

			// Display the robot position
			robot.shape1->setPos(robot.x, robot.y);
			robot.shape2->setPos(robot.x, robot.y);
			robot.shape3->setPos(robot.x, robot.y);

			target.shape1->setPos(target.x, target.y);
			target.shape2->setPos(target.x, target.y);
			target.shape3->setPos(target.x, target.y);

			viewer1->addShape(robot.shape1);
			viewer2->addShape(robot.shape2);
			viewer3->addShape(robot.shape3);

			viewer1->addShape(target.shape1);
			viewer2->addShape(target.shape2);
			viewer3->addShape(target.shape3);

			return (b);

		}///}}}

		bool LGLPursuitViewer::refresh()
		{///{{{{

			// Controler
			switch( myPG->getroot()->risk )
			{
				case OK 					: 	ctrl->setBackgroundColor(   0, 255,  0 ); break;
				case RISK 					: 	ctrl->setBackgroundColor( 255, 255,  0 ); break;
				case DANGER_HIDDEN 			: 	ctrl->setBackgroundColor( 255,   0,  0 ); break;
				case DANGER_LOST 			: 	ctrl->setBackgroundColor( 255,   0,  0 ); break;
				case DANGER_UNSEEN 			: 	ctrl->setBackgroundColor( 255,   0,  0 ); break;
				case DANGER_DISAPPEARED 	: 	ctrl->setBackgroundColor(   0,   0,  0 ); break;

				default 	: 	ctrl->setBackgroundColor(  255, 255, 255); 
			}
				
			// Update risks and danger
			for (unsigned int i = 0 ; i < risks.size() ; i++ ) 
				delete risks.at(i);
			risks.clear();
			
			std::set< Cell_ID>::const_iterator itS ;
			int ysize = myPG->getysize();

			// Dangers
			///{{{
			std::set<Cell_ID> _danger;
			for (unsigned int i = 0 ; i < myPG->getdanger().size() ; i++ ) 
			{
				boost::shared_ptr< PursuitGNode> ptr = (myPG->getdanger().at(i)).lock() ;
				if (ptr)
					_danger.insert(ptr->targetPos.begin(), ptr->targetPos.end() );
			}

			for (itS = _danger.begin() ; itS != _danger.end() ; ++ itS)
			{
				for ( int j=0 ; j<3; j++)
				{
					risks.push_back(new Shape( Shape::ShapeEllipseFilled, (*itS / ysize)*scale/2, (*itS % ysize)*scale/2, 0.7*scale, 0.7*scale, 0) );
					risks.at(risks.size()-1)->setColor( 240, 240, 40 );
					//risks.push_back(new Shape( Shape::ShapeCross, (*itS / ysize)*scale/2, (*itS % ysize)*scale/2, 0.6*scale, 0.6*scale, 45) );
					//risks.at(risks.size()-1)->setColor( 150, 150,  50 );
					if (!visibleRisks)	
						risks.at(risks.size()-1)->setVisible(false);
				}
				viewer1->addShape(risks.at(risks.size()-3));
				viewer2->addShape(risks.at(risks.size()-2));
				viewer3->addShape(risks.at(risks.size()-1));
			}
			///}}}

			// Risks
			///{{{
			std::set<Cell_ID> _risk;
			for (unsigned int i = 0 ; i < myPG->getrisk().size() ; i++ ) 
			{
				boost::shared_ptr< PursuitGNode> ptr = (myPG->getrisk().at(i)).lock() ;
				if (ptr)
					_risk.insert(ptr->targetPos.begin(), ptr->targetPos.end() );
			}

			for (itS = _risk.begin() ; itS != _risk.end() ; ++ itS)
			{
				for ( int j=0 ; j<3; j++)
				{
					risks.push_back(new Shape( Shape::ShapeRectangleFilled, (*itS / ysize)*scale/2, (*itS % ysize)*scale/2, 0.4*scale, 0.4*scale, 0) );
					risks.at(risks.size()-1)->setColor( 200, 30, 120 );
					//risks.push_back(new Shape( Shape::ShapeCross, (*itS / ysize)*scale/2, (*itS % ysize)*scale/2, 0.2*scale, 0.2*scale, 0) );
					//risks.at(risks.size()-1)->setColor( 100, 100, 100 );
					if (!visibleRisks)	
						risks.at(risks.size()-1)->setVisible(false);
				}
				viewer1->addShape(risks.at(risks.size()-3));
				viewer2->addShape(risks.at(risks.size()-2));
				viewer3->addShape(risks.at(risks.size()-1));
			}
			///}}}
			
			// Target Past Positions
			///{{{
			if (!targetPastPositions.empty() )
			{
				for ( int j=0 ; j<3; j++)
				{
					viewTargetPastPositions.push_back( new Shape(	Shape::ShapeEllipseFilled, 
																	(targetPastPositions.back().first ) * scale /2, 
																	(targetPastPositions.back().second) * scale /2, 
																	0.5*scale, 0.5*scale)
														);
					viewTargetPastPositions.back()->setColor( 100, 100, 100 );
					if (!visiblePastPositions)	
						viewTargetPastPositions.back()->setVisible(false);
				}
				viewer1->addShape(viewTargetPastPositions.at(viewTargetPastPositions.size()-3));
				viewer2->addShape(viewTargetPastPositions.at(viewTargetPastPositions.size()-2));
				viewer3->addShape(viewTargetPastPositions.at(viewTargetPastPositions.size()-1));
			}
			///}}}

			// Robot Past Positions
			///{{{
			if (!robotPastPositions.empty() )
			{
				for ( int j=0 ; j<3; j++)
				{
					viewRobotPastPositions.push_back( new Shape(	Shape::ShapeRectangleFilled, 
																	(robotPastPositions.back().first ) * scale /2, 
																	(robotPastPositions.back().second) * scale /2, 
																	0.5*scale, 0.5*scale)
														);
					viewRobotPastPositions.back()->setColor( 180, 180, 180 );
					if (!visiblePastPositions)	
						viewRobotPastPositions.back()->setVisible(false);
				}
				viewer1->addShape(viewRobotPastPositions.at(viewRobotPastPositions.size()-3));
				viewer2->addShape(viewRobotPastPositions.at(viewRobotPastPositions.size()-2));
				viewer3->addShape(viewRobotPastPositions.at(viewRobotPastPositions.size()-1));
			}
			///}}}

			// Move the robot position
			robot.shape1->setPos(robot.x * scale /2, robot.y * scale /2);
			robot.shape2->setPos(robot.x * scale /2, robot.y * scale /2);
			robot.shape3->setPos(robot.x * scale /2, robot.y * scale /2);

			// Move the target position
			target.shape1->setPos(target.x * scale /2, target.y * scale /2);
			target.shape2->setPos(target.x * scale /2, target.y * scale /2);
			target.shape3->setPos(target.x * scale /2, target.y * scale /2);

			return (true);
		}///}}}

		void LGLPursuitViewer::displayrisks()
		{///{{{
			visibleRisks = true ;
			for (unsigned int i=0 ; i< risks.size() ; i++)
				risks.at(i)->setVisible(true);
		}///}}}

		void LGLPursuitViewer::hiderisks()
		{///{{{
			visibleRisks = false ;
			for (unsigned int i=0 ; i< risks.size() ; i++)
				risks.at(i)->setVisible(false);
		}///}}}

		void LGLPursuitViewer::displaypastpositions()
		{///{{{
			visiblePastPositions = true ;
			
			for (unsigned int i=0 ; i< viewTargetPastPositions.size() ; i++)
				viewTargetPastPositions.at(i)->setVisible(true);
			
			for (unsigned int i=0 ; i< viewRobotPastPositions.size() ; i++)
				viewRobotPastPositions.at(i)->setVisible(true);

		}///}}}

		void LGLPursuitViewer::hidepastpositions()
		{///{{{
			visiblePastPositions = false ;
			
			for (unsigned int i=0 ; i< viewTargetPastPositions.size() ; i++)
				viewTargetPastPositions.at(i)->setVisible(false);
			
			for (unsigned int i=0 ; i< viewRobotPastPositions.size() ; i++)
				viewRobotPastPositions.at(i)->setVisible(false);

		}///}}}

		void LGLPursuitViewer::setRobotAccess( string _robotAccessPath )
		{///{{{
			robotAccessPath = _robotAccessPath ;
		}///}}}

		void LGLPursuitViewer::setTargetAccess( string _targetAccessPath )
		{///{{{
			targetAccessPath = _targetAccessPath ;
		}///}}}

		void LGLPursuitViewer::setRobotVisib( string _robotVisibPath )
		{///{{{
			robotVisibPath = _robotVisibPath ;
		}///}}}
	
		void LGLPursuitViewer::setRobotPosition( int _x, int _y)
		{///{{{
			robot.x = _x;
			robot.y = _y;
		}///}}}

		void LGLPursuitViewer::setTargetPosition( int _x, int _y)
		{///{{{
			target.x = _x;
			target.y = _y;
		}///}}}

		void LGLPursuitViewer::setHorizon( unsigned int _horizon )
		{///{{{
			horizon=_horizon;
		}///}}}

		void LGLPursuitViewer::move( int _x, int _y)
		{///{{{

			#if TIME_CONSTRAINT
				double tMax = DEFAULT_TMAX ;   
			#endif


			timeval t;
			double d, dl ;
			double dd;
			


			gettimeofday(&t,NULL);
			d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0;

			// Check validity of the movement
			int newTargetPosx, newTargetPosy;
			
			newTargetPosx = target.x + 2 * _x ;
			newTargetPosy = target.y + 2 * _y ;

			newTargetPosx = newTargetPosx < 0 ? 0 : newTargetPosx ;
			newTargetPosy = newTargetPosy < 0 ? 0 : newTargetPosy ;

			newTargetPosx = (unsigned int) newTargetPosx >= myPG->getxsize() ? myPG->getxsize()-2 : (newTargetPosx/2) * 2 + 1 ;
			newTargetPosy = (unsigned int) newTargetPosy >= myPG->getysize() ? myPG->getysize()-2 : (newTargetPosy/2) * 2 + 1 ;

			// check with geodata
			if ( myPG->getnavigator()->gettarGeoData()->getValueFromBand( GEODATA_TFLAT_BAND, newTargetPosx/2, newTargetPosy/2 ) > 0.5 ) 
			{
	
				// Updating target position
				target.x = newTargetPosx ;
				target.y = newTargetPosy ;

				// Computing

				gettimeofday(&t,NULL); dl = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0;
					myPG->updatePG ( target.x, target.y, robot.x, robot.y);
				gettimeofday(&t,NULL); dl = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - dl;
				//std::cout << " the limit for completeToHorizon is : " << (( tMax - dl > 0.0 ) ? tMax-dl : 0.0 )<< std::endl; 

				#if TIME_CONSTRAINT
					myPG->completeToHorizonBefore( ( tMax - dl > 0.0 ) ? tMax - dl : 0.0 ); 
				#else
					myPG->completeToHorizon( );
				#endif

				// Refreshing the display
				gettimeofday(&t,NULL); dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0;
					refresh();

				gettimeofday(&t,NULL); d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d; 
				dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - dd;

				//myPG->disp_risk();
				//myPG->disp_danger();

				// output	
				std::cout 	<< "You move to (" << target.x << "," << target.y << ") and " ; 

				switch ( myPG->getroot()->risk )
				{
					case OK	:
						std::cout << "the robot follows you in (" << robot.x << "," << robot.y << ") and is watching you... "; break;
					case RISK :
						std::cout << "the robot follows you in (" << robot.x << "," << robot.y << ") but cannot see you ! "; break ;
					case DANGER_HIDDEN :	
						std::cout << "the robot failed ! (you hide ;-) ) "; break;
					case DANGER_LOST :
						std::cout << "the robot failed ! (cannot trace you anymore :P ) "; break;
					case DANGER_UNSEEN : 	
						std::cout << "the robot failed ! (hasn't seen you for too much time...) "; break;
					case DANGER_DISAPPEARED : 	
						std::cout << "the robot failed ! (you just disappeared ! oO ) "; break;
					default :
							;//throw exception ?
				}

				std::cout 	<< "[ temporal horizon is "<< myPG->gethorizon() 
								<< " with front = "<< myPG->getfront().size() 
								<< " - risks = " << myPG->getrisk().size() 
								<<" - danger = " << myPG->getdanger().size() 
							<< " ] " << std::endl ; 
				std::cout 	<< "[computed in " << d << " ms"
								<< " including " << dl << " ms for updatePG" 
								<< " and " << dd << " ms for display." 
							<< "]" << std::endl;

				// statistics
				stats.update( myPG, d );

				robotPastPositions.push_back ( std::make_pair( robot.x,  robot.y  ) );
				targetPastPositions.push_back( std::make_pair( target.x, target.y ) );

				// Log file (when asked for)
				#if SAVE_FILE
				   fout	 		<< target.x << " " << target.y  	// target coordinates
				   		<< " " 	<< robot.x 	<< " " << robot.y 			// robot coordinates
				   		<< " " 	<< myPG->getroot()->risk		// current state
					 	<< " "	<< myPG->gethorizon() 			// current temporal horizon
						<< " " 	<< myPG->getfront().size()		// nbr of front nodes 
						//<< " " << myPG->getrisk().size()		// nbr of risks nodes 
						<< " " 	<< myPG->getdanger().size()		// nbr of danger nodes 
						<< " " 	<< d 							// time required by the computation
						<< " " 	<< dl							// time required for the PG update 
						<< std::endl;
				#endif


			} else {
				std::cout<< "You are NOT allowed to go there ! (" << newTargetPosx << "," << newTargetPosy << ") - Please choose another action." << std::endl;
			}

		}///}}}

		void LGLPursuitViewer::displayPath( const std::vector< Cell_ID> &path)
		{///{{{
			for (unsigned int i = 0 ; i < path.size() ; i++)
			{
				Shape* stone = new Shape(Shape::ShapeEllipseFilled, (path.at(i)/myPG->getysize()) * scale /2, (path.at(i)%myPG->getysize() ) * scale /2, 0.5*scale, 0.5*scale);
				stone->setColor(100,100,100);

				viewer1->addShape(stone);
			}
		}///}}}

		void LGLPursuitViewer::displayPath2Set( const std::vector< Cell_ID> &path, const std::set< Cell_ID> &set)
		{///{{{
		
			std::set< Cell_ID>::const_iterator itc;
			for ( itc = set.begin() ; itc != set.end() ; ++ itc )
			{
				Shape* stone = new Shape(Shape::ShapeEllipseFilled, (*itc/myPG->getysize()) * scale /2, (*itc%myPG->getysize() ) * scale /2, 0.7*scale, 0.7*scale);
				stone->setColor(100,100,200);

				viewer1->addShape(stone);
			}

			for (unsigned int i = 0 ; i < path.size() ; i++)
			{
				Shape* stone = new Shape(Shape::ShapeEllipseFilled, (path.at(i)/myPG->getysize()) * scale /2, (path.at(i)%myPG->getysize() ) * scale /2, 0.5*scale, 0.5*scale);
				stone->setColor(200,200,200);

				viewer1->addShape(stone);
			}
		}///}}}

		void LGLPursuitViewer::displaystats()
		{///{{{
			std::cout << stats ;
		}///}}}

		void LGLPursuitViewer::displayhelp()
		{///{{{
			std::cout << " Use keyboard to control the target !" << std::endl ;
				// control command
				#if SWITCH_CONNEXITY
				std::cout << "\t\t ( Up     \t=>\t Move Up                )" << std::endl ;
				std::cout << "\t\t ( Down   \t=>\t Move Down              )" << std::endl ;
				std::cout << "\t\t ( Left   \t=>\t Move Left              )" << std::endl ;
				std::cout << "\t\t ( Right  \t=>\t Move Right             )" << std::endl ;
				std::cout << "\t\t ( Return \t=>\t Stay here              )" << std::endl ;
				#else
				std::cout << "\t\t ( Z      \t=>\t Move Up                )" << std::endl ;
				std::cout << "\t\t ( X      \t=>\t Move Down              )" << std::endl ;
				std::cout << "\t\t ( Q      \t=>\t Move Left              )" << std::endl ;
				std::cout << "\t\t ( D      \t=>\t Move Right             )" << std::endl ;
				std::cout << "\t\t ( S      \t=>\t Stay here              )" << std::endl ;
				std::cout << "\t\t ( A      \t=>\t Move Up and Left       )" << std::endl ;
				std::cout << "\t\t ( E      \t=>\t Move Up and Right      )" << std::endl ;
				std::cout << "\t\t ( W      \t=>\t Move Down and Left     )" << std::endl ;
				std::cout << "\t\t ( C      \t=>\t Move Down and Right    )" << std::endl ;
				#endif

				std::cout << std::endl ;

				// option command
				std::cout << "\t\t ( H  	\t=>\t Display help           )" << std::endl ;
				std::cout << "\t\t ( I  	\t=>\t Display statistics     )" << std::endl ;
				std::cout << "\t\t ( Space  \t=>\t Display/Hide the risks )" << std::endl ;
				std::cout << "\t\t ( Escape \t=>\t Exit                   )" << std::endl ;
				
		}///}}}

	}
}

#endif
