/**
 * \file LGLViewer.cpp
 *
 * Source file that defines Qt drawing functions
 *
 * \author Redouane Boumghar, Hung Cao
 *
 * \ingroup lgl
 *
 **/

#ifdef HAVE_MODULE_QDISPLAY // lgl viewer is on if HAVE_MODULE_QDISPLAY is defined

#include <lgl/LGLViewer.hpp>
#include <lgl/NavGNode.hpp>

#include <set>

using namespace jafar::qdisplay;

namespace jafar {
	namespace lgl {

typedef void(*FUNCAST)(void*);
		// Default constructor
		LGLViewer::LGLViewer():
		withUTM(false),
		decomp(NULL),
		graph(NULL),
		scaleFactor(1.0),
		n_path_drawn(0),
		visibility(NULL),
		highcolor(0),
		lowcolor(0)
		{ //{{{
			// a viewer with screen coordinates (corner 0,0 is at the top left, x is the horizontal axis)
			viewer = new jafar::qdisplay::Viewer();
		} //}}}

		// Specialized constructor
		LGLViewer::LGLViewer(Decomp* _decomp, NavGraph::NavGraphContainer* _graph, double sf):
		withUTM(false),
		decomp(_decomp),
		graph(_graph),
		scaleFactor(sf),
		n_path_drawn(0),
		visibility(NULL),
		highcolor(0),
		lowcolor(0)
		{//{{{
			// a viewer with screen coordinates (corner 0,0 is at the top left, x is the horizontal axis)
			viewer = new jafar::qdisplay::Viewer();
		}//}}}

		// Specialized constructor (visibility)
		LGLViewer::LGLViewer(VisibilityMap *_visibility, Decomp* _decomp, NavGraph::NavGraphContainer* _graph, double sf):
		withUTM(false),
		decomp(_decomp),
		graph(_graph),
		scaleFactor(sf),
		n_path_drawn(0),
		visibility(_visibility),
		highcolor(0),
		lowcolor(0)
		{//{{{
			// a viewer with screen coordinates (corner 0,0 is at the top left, x is the horizontal axis)
			viewer = new jafar::qdisplay::Viewer();
		}//}}}

		LGLViewer::~LGLViewer()
		{
			delete viewer;
		}

		jafar::qdisplay::Viewer* LGLViewer::getQViewer() {
			return viewer;
		}

		void LGLViewer::setDecomp(Decomp* _decomp)
		{/*{{{*/
			this->decomp = _decomp;
		}/*}}}*/

		void LGLViewer::setGraph(NavGraph::NavGraphContainer* _graph)
		{/*{{{*/
			this->graph = _graph;
		}/*}}}*/

		void LGLViewer::setScaleFactor(double _sf)
		{/*{{{*/
			this->scaleFactor = _sf;
		}/*}}}*/

		void LGLViewer::setImage(const std::string& filename, double xoffset, double yoffset)
		{/*{{{*/
			ImageView* imgV = new ImageView(filename);
			imgV->setPos(xoffset, yoffset);
			imgV->setScale(2.0);
			viewer->setImageView(imgV);
		}/*}}}*/

		void LGLViewer::displayPath(NavPath& foundPath)
		{/*{{{*/
			typedef NavPath::const_iterator path_iterator;
			NavLocation prevPos;
			NavLocation curPos;
			bool drawedger = false;
			// Draw grey for first node
			lowcolor = 20;
			highcolor = 20;
			for(path_iterator lit = foundPath.begin(); lit != foundPath.end(); ++lit) {
				curPos = (*lit).first;
				// Drawing edge first to define colors
				if (drawedger)
				{
					drawEdge(prevPos, curPos, DRAW_NODE_PATH);
				} else { drawedger = true; }
				// Displaying nodes with the same edge color.
				// Only after the first node has been drawn.
				displayHook(curPos, DRAW_NODE_PATH);
				prevPos = curPos;
			}
			n_path_drawn++;
			if (n_path_drawn >= NMAX_PATHS)
			{
				n_path_drawn = 0;
			}
		}/*}}}*/

		void LGLViewer::displayFrontiers(std::list<NavGraph::frontierPoint> frontierList)
		{/*{{{*/
			// Condition to continue
			if (frontierList.size() <= 0) { return; }
			// List iterator
			std::list<NavGraph::frontierPoint>::iterator itfrt;
			// Draw grey for first node
			lowcolor = 20;
			highcolor = 20;
			// Looking for the minimum cost in the frontier list
			// We are not taking into account the first -1.0 values at the
			// begining of the list as they just represent out of A* bounds
			// frontiers and they have not been set with a real cost.
			double mintmp = -1.0;
			for(itfrt = frontierList.begin(); itfrt != frontierList.end(); ++itfrt) {
				if ((*itfrt).second >= 0.0)
				{
					if (mintmp == -1.0)
					{
						mintmp = (*itfrt).second;
					} else {
						if ((*itfrt).second < mintmp)
						{ mintmp = (*itfrt).second; }
					}

				}
			}
			// Setting the maximum cost in the frontier list
			double maxtmp = (frontierList.back()).second;

			for(itfrt = frontierList.begin(); itfrt != frontierList.end(); ++itfrt) {
				displayFrontierPoint((*itfrt), mintmp, maxtmp);
			}
		}/*}}}*/

		bool LGLViewer::displayDecomp()
		{ //{{{

			// In case the LGL Viewer was constructed without "decomp"
			if (decomp == NULL)
			{
				return false;
			}

			// Init.
			int xroot = 0;
			int yroot = 0;
			// Parameters for qdisplay have to be in double precision
			double xcenter = 0;
			double ycenter = 0;
			double xsize = 0;
			double ysize = 0;

			// Show the quadtree decomposition
			Shape* curCluster;
			Decomp_iterator it(decomp);		
			// ++it only increments on positive sized clusters
			for(it.begin(); !(it.at_end()); ++it) {
				// Display this "it" node or leaf from the quadtree decomp.
				it.getIndex(xroot,yroot);

				// For comprehension: we draw the decomposition in Navigation coordinates reference
				// so we introduce (*2.0) factor to translate position from the decomposition raster frame.
				// into the position manager (navigation) frame
				xsize = (double)((*it)->xsize)*2.0*scaleFactor;
				ysize = (double)((*it)->ysize)*2.0*scaleFactor;		
				xcenter = ((double)xroot)*2.0*scaleFactor + xsize/2.0;
				ycenter = ((double)yroot)*2.0*scaleFactor + ysize/2.0;

				curCluster = new Shape(Shape::ShapeRectangle, xcenter, ycenter, xsize, ysize);
				curCluster->setColor(30,50,70);
				viewer->addShape(curCluster);
			}

			return true;
		} //}}}
		
		bool LGLViewer::displayGraph(bool show_edges, bool show_nodes)
		{ /*{{{*/

			if (graph == NULL)
			{
				return false;
			}

			// print the graph		
			NavGraph::NavGraphContainer::GraphContainer* g = graph->getGCtn();

			//--- print all edges
			if (show_edges)
			{
				Line* curEdge;
				NavGraph::NavGraphContainer::Edge e;
				BGL_FORALL_EDGES(e, *g, NavGraph::NavGraphContainer::GraphContainer) {
					curEdge = new Line(
							   (*g)[source(e,*g)].pos.pos_x * scaleFactor,
							   (*g)[source(e,*g)].pos.pos_y * scaleFactor,
							   (*g)[target(e,*g)].pos.pos_x * scaleFactor,
							   (*g)[target(e,*g)].pos.pos_y * scaleFactor);
					curEdge->setColor(120,120,255);
					viewer->addLine(curEdge);
				}
			}

			//--- print all vertices
			if (show_nodes)
			{
				Shape *curVertex;
				//NavGraph::NavGraphContainer::Vertex v;
				BGL_FORALL_VERTICES(v,*g, adjacency_list<>) {
					curVertex = new Shape(Shape::ShapeCross,
								(*g)[v].pos.pos_x * scaleFactor,
								(*g)[v].pos.pos_y * scaleFactor,
								0.6,0.6);
					curVertex->setColor(255,40,40);
					viewer->addShape(curVertex);
				}
			}

			return true;
		} /*}}}*/

		bool LGLViewer::displayVisibilityMap(const int x, const int y, bool show_viewlines, bool show_nodes, bool show_rays)
		{ /*{{{*/

			if (visibility == NULL)
				return false;

			// get the data from the visibilityMap
			const VisibleArea*  myView = visibility->getPtVisibility(x,y) ;
			int xx, yy, ys, vl, n ;
			double alpha;
			ys = myView->getysize() ;
			xx = myView->getOriginalVertex() / ys ; 
			yy = myView->getOriginalVertex() - ( (myView->getOriginalVertex()/ys) * ys ) ;

			vl 		= visibility->getviewLength();
			n 		= visibility->getn();
			alpha 	= visibility->getalpha();
			
			// rounded central node
			Shape *roundedCentralVertex;
			roundedCentralVertex = new Shape(Shape::ShapeRectangleFilled, 	
					xx * scaleFactor,
					yy * scaleFactor,
					2.0,2.0);
			roundedCentralVertex->setColor(230,50,0);
			viewer->addShape(roundedCentralVertex);
			
			// central node
			Shape *centralVertex;
			centralVertex = new Shape(Shape::ShapeRectangleFilled, 	
					x * scaleFactor,
					y * scaleFactor,
					0.5,0.5);
			centralVertex->setColor(255,150,0);
			viewer->addShape(centralVertex);

			//for each node visible from the central node
			std::set<Cell_ID>::const_iterator it;
			for ( it = myView->visibleArea.begin(); it != myView->visibleArea.end() ; ++it )
			{//{{{
				if (show_viewlines) 
				{
					Line* curEdge;
					curEdge = new Line(
						   xx * scaleFactor,
						   yy * scaleFactor,
						   (*it / ys) * scaleFactor,
						   (*it - (*it / ys) * ys) * scaleFactor);
					curEdge->setColor(80,130,100);
					viewer->addLine(curEdge);
				}

				if(show_nodes)
				{
					Shape *curVertex;
					curVertex = new Shape(Shape::ShapeCross,
						    (*it / ys) * scaleFactor,
						   	(*it - (*it / ys) * ys) * scaleFactor,
							0.6,0.6);
					curVertex->setColor(50,250,50);
					viewer->addShape(curVertex);

				}

			}//}}}

			// viewLength circle
			Shape *vCircle;
			vCircle= new Shape(Shape::ShapeEllipse, 
					xx * scaleFactor,
					yy * scaleFactor,
					4*vl,4*vl );
			vCircle->setColor(255,150,0);
			viewer->addShape(vCircle);

			// rays
			if (show_rays)
			{
				// for each ray
				for (int i=0 ; i< n ; i++ )
				{
					Line* curRay;
					curRay = new Line(
						   xx * scaleFactor,
						   yy * scaleFactor,
						   ( xx + cos ( i * alpha ) * 2 * vl ) * scaleFactor,
						   ( yy + sin ( i * alpha ) * 2 * vl ) * scaleFactor);
					curRay->setColor(255,150,0);
					viewer->addLine(curRay);
				}

			}

			return true;
		} /*}}}*/

		bool LGLViewer::displayVisibleArea(const VisibleArea &myArea, bool show_viewlines, bool show_nodes, bool show_rays)
		{ /*{{{*/

			if (visibility == NULL)
				return false;

			// get the data from the visibilityMap
			int xx, yy, ys, vl, n ;
			double alpha;
			ys = myArea.getysize() ;
			xx = myArea.getOriginalVertex() / ys ; 
			yy = myArea.getOriginalVertex() - ( (myArea.getOriginalVertex()/ys) * ys ) ;

			vl 		= visibility->getviewLength();
			n 		= visibility->getn();
			alpha 	= visibility->getalpha();
			
			// rounded central node
			Shape *roundedCentralVertex;
			roundedCentralVertex = new Shape(Shape::ShapeRectangleFilled, 	
					xx * scaleFactor,
					yy * scaleFactor,
					2.0,2.0);
			roundedCentralVertex->setColor(230,50,0);
			viewer->addShape(roundedCentralVertex);
			
			//for each node visible from the central node
			std::set<Cell_ID>::const_iterator it;
			for ( it = myArea.visibleArea.begin(); it != myArea.visibleArea.end() ; ++it )
			{//{{{
				if (show_viewlines) 
				{
					Line* curEdge;
					curEdge = new Line(
						   xx * scaleFactor,
						   yy * scaleFactor,
						   (*it / ys) * scaleFactor,
						   (*it - (*it / ys) * ys) * scaleFactor);
					curEdge->setColor(80,130,100);
					viewer->addLine(curEdge);
				}

				if(show_nodes)
				{
					Shape *curVertex;
					curVertex = new Shape(Shape::ShapeCross,
						    (*it / ys) * scaleFactor,
						   	(*it - (*it / ys) * ys) * scaleFactor,
							0.6,0.6);
					curVertex->setColor(50,250,50);
					viewer->addShape(curVertex);

				}

			}//}}}

			// viewLength circle
			Shape *vCircle;
			vCircle= new Shape(Shape::ShapeEllipse, 
					xx * scaleFactor,
					yy * scaleFactor,
					4*vl,4*vl );
			vCircle->setColor(255,150,0);
			viewer->addShape(vCircle);

			// rays
			if (show_rays)
			{
				// for each ray
				for (int i=0 ; i< n ; i++ )
				{
					Line* curRay;
					curRay = new Line(
						   xx * scaleFactor,
						   yy * scaleFactor,
						   ( xx + cos ( i * alpha ) * 2 * vl ) * scaleFactor,
						   ( yy + sin ( i * alpha ) * 2 * vl ) * scaleFactor);
					curRay->setColor(255,150,0);
					viewer->addLine(curRay);
				}

			}

			return true;
		} /*}}}*/

		bool LGLViewer::displayPGFront(const PursuitGraph &_PG)
		{///{{{

			std::vector< boost::weak_ptr <PursuitGNode> >::const_iterator itcb1, itce1 ; 
			itcb1 = _PG.getfront().begin();
			itce1 = _PG.getfront().end();
			
			unsigned int ys = _PG.getysize();

			for ( ; itcb1!=itce1 ; ++itcb1 )
			{
			
				//lock
				boost::shared_ptr <PursuitGNode> pt = itcb1->lock();

				//display
				if ( pt )
				{
					// Robot position
					Cell_ID robotID = pt->robotPos;
					Shape *rPos;
					rPos = new Shape(Shape::ShapeRectangleFilled,
						    ( robotID / ys) * scaleFactor,
						   	( robotID % ys) * scaleFactor,
							2.0,2.0);
					rPos->setColor(130,200,50);
					viewer->addShape(rPos);

					// target positions
					std::set< Cell_ID >::const_iterator itcb2, itce2 ; 
					itcb2 = pt->targetPos.begin();
					itce2 = pt->targetPos.end();
					for ( ; itcb2 != itce2 ; ++ itcb2)
					{
						//Shape *tPos;
						//tPos = new Shape(Shape::ShapeCross,
						//		( *itcb2 / ys) * scaleFactor,
						//		( *itcb2 % ys) * scaleFactor,
						//		0.6,0.6);
						//tPos->setColor(50,250,50);
						//viewer->addShape(tPos);
						Shape *tPos;
						tPos = new Shape(Shape::ShapeEllipse,
								( *itcb2 / ys) * scaleFactor,
								( *itcb2 % ys) * scaleFactor,
								1,1);
						tPos->setColor(100,250,100);
						viewer->addShape(tPos);

						Line* tLink;
						tLink = new Line(
								( robotID / ys) * scaleFactor,
								( robotID % ys) * scaleFactor,
								( *itcb2 / ys) * scaleFactor,
								( *itcb2 % ys) * scaleFactor);
						tLink->setColor(255,150,0);
						viewer->addLine(tLink);

					}
				}
			}

			return true;

		}///}}}

		bool LGLViewer::displayPGRisk(const PursuitGraph &_PG)
		{///{{{

			std::vector< boost::weak_ptr <PursuitGNode> >::const_iterator itcb1, itce1 ; 
			itcb1 = _PG.getrisk().begin();
			itce1 = _PG.getrisk().end();

			unsigned int ys = _PG.getysize();
			
			for ( ; itcb1!=itce1 ; ++itcb1 )
			{
			
				//lock
				boost::shared_ptr <PursuitGNode> pt = itcb1->lock();

				//display
				if ( pt )
				{
					// Robot position
					//Cell_ID robotID = pt->robotPos;
					//Shape *rPos;
					//rPos = new Shape(Shape::ShapeRectangleFilled,
					//	    ( robotID / ys) * scaleFactor,
					//	   	( robotID % ys) * scaleFactor,
					//		2.0,2.0);
					//rPos->setColor(250,75,75);
					//viewer->addShape(rPos);

					// target positions
					std::set< Cell_ID >::const_iterator itcb2, itce2 ; 
					itcb2 = pt->targetPos.begin();
					itce2 = pt->targetPos.end();
					for ( ; itcb2 != itce2 ; ++ itcb2)
					{
						Shape *tPos;
						tPos = new Shape(Shape::ShapeEllipse,
								( *itcb2 / ys) * scaleFactor,
								( *itcb2 % ys) * scaleFactor,
								2,2);
						tPos->setColor(100,200,255);
						viewer->addShape(tPos);

						//Line* tLink;
						//tLink = new Line(
						//		( robotID / ys) * scaleFactor,
						//		( robotID % ys) * scaleFactor,
						//		( *itcb2 / ys) * scaleFactor,
						//		( *itcb2 % ys) * scaleFactor);
						//tLink->setColor(255,150,0);
						//viewer->addLine(tLink);

					}
				}
			}

			return true;

		}///}}}

		bool LGLViewer::displayPGraph(const PursuitGraph &_PG, bool showTarget )
		{///{{{

			unsigned int ys = _PG.getysize();
			unsigned long int tmax = _PG.gettime();
			double ratio ;

			// display root
			Cell_ID robotID = _PG.getroot()->robotPos;
			Shape *rootPos;
			rootPos = new Shape(Shape::ShapeCross,
				( robotID / ys) * scaleFactor,
				( robotID % ys) * scaleFactor,
				2.0,2.0, 45);
			rootPos->setColor(255,255,255);
			viewer->addShape(rootPos);

			rootPos = new Shape(Shape::ShapeRectangle,
				( robotID / ys) * scaleFactor,
				( robotID % ys) * scaleFactor,
				2.0,2.0);
			rootPos->setColor(255,255,255);
			viewer->addShape(rootPos);

			Cell_ID targetID = * (_PG.getroot()->targetPos.begin());
			rootPos = new Shape(Shape::ShapeCross,
				( targetID / ys) * scaleFactor,
				( targetID % ys) * scaleFactor,
				2.0,2.0, 45);
			rootPos->setColor(255,255,255);
			viewer->addShape(rootPos);
			rootPos = new Shape(Shape::ShapeEllipse,
				( targetID / ys) * scaleFactor,
				( targetID % ys) * scaleFactor,
				2.0,2.0);
			rootPos->setColor(255,255,255);
			viewer->addShape(rootPos);


			// display graph (robot temporal positions)
			std::set< boost::shared_ptr <PursuitGNode> >::const_iterator itN ; 
			for ( itN = _PG.getsetOfNodes().begin(); itN != _PG.getsetOfNodes().end(); ++ itN )
			{
				// ratio
				ratio = 0.5 * (*itN)->time / ( (double) tmax) ;

				// robot position
				Cell_ID robotID = (*itN)->robotPos;
				Shape *rPos;
				rPos = new Shape(Shape::ShapeRectangleFilled,
					( robotID / ys) * scaleFactor,
					( robotID % ys) * scaleFactor,
					ratio, ratio);
				rPos->setColor(0,80,0);
				viewer->addShape(rPos);
				
				// target positions
				if (showTarget)
				{
					std::set< Cell_ID >::const_iterator itT ; 
					itT = (*itN)->targetPos.begin();
					for ( ; itT != (*itN)->targetPos.end(); ++ itT)
					{
						Shape *tPos;
						tPos = new Shape(Shape::ShapeEllipse,
								( *itT / ys) * scaleFactor,
								( *itT % ys) * scaleFactor,
								2*ratio, 2*ratio);
						tPos->setColor(0,250,250);
						viewer->addShape(tPos);
					}
				}
			}

			// highlight the front nodes
			std::vector< boost::weak_ptr <PursuitGNode> >::const_iterator itN2 ; 
			itN2 = _PG.getfront().begin();
			for ( ; itN2!=_PG.getfront().end() ; ++itN2 )
			{
				//lock
				boost::shared_ptr <PursuitGNode> pt = itN2->lock();

				//display
				if ( pt )
				{
					Cell_ID robotID = pt->robotPos;
					Shape *rPos;
					rPos = new Shape(Shape::ShapeRectangleFilled,
						( robotID / ys) * scaleFactor,
						( robotID % ys) * scaleFactor,
						0.2,0.2);
					rPos->setColor(250,160,120);
					viewer->addShape(rPos);
				}
			 }


			// display risk
			itN2 = _PG.getrisk().begin();
			for ( ; itN2!=_PG.getrisk().end() ; ++itN2 )
			{
				//lock
				boost::shared_ptr <PursuitGNode> pt = itN2->lock();

				//display
				if ( pt )
				{
					// ratio
					//ratio = 0.5 * pt->time / ( (double) tmax) ;

					std::set< Cell_ID >::const_iterator itT ; 
					itT = pt->targetPos.begin();
					for ( ; itT != pt->targetPos.end(); ++ itT)
					{
						Shape *tPos;
						tPos = new Shape(Shape::ShapeCross,
								( *itT / ys) * scaleFactor,
								( *itT % ys) * scaleFactor,
								1.5, 1.5, 45);
						tPos->setColor(250,250,250);
						viewer->addShape(tPos);
					}
				}
			 }

			// display danger
			itN2 = _PG.getdanger().begin();
			for ( ; itN2!=_PG.getdanger().end() ; ++itN2 )
			{
				//lock
				boost::shared_ptr <PursuitGNode> pt = itN2->lock();

				//display
				if ( pt )
				{
					// ratio
					//ratio = 0.5 * pt->time / ( (double) tmax) ;

					std::set< Cell_ID >::const_iterator itT ; 
					itT = pt->targetPos.begin();
					for ( ; itT != pt->targetPos.end(); ++ itT)
					{
						Shape *tPos;
						tPos = new Shape(Shape::ShapeCross,
								( *itT / ys) * scaleFactor,
								( *itT % ys) * scaleFactor,
								1.5, 1.5, 45);
						tPos->setColor(250,200,50);
						viewer->addShape(tPos);
					}
				}
			}

			return ( true );

		}///}}}

		bool LGLViewer::displayGraphWithUTM(bool show_edges, bool show_nodes)
		{ /*{{{*/
			if (graph == NULL)
			{
				return false;
			}

			// Telling other functions that we are using UTM
			withUTM = true;

			// Displaying the graph		
			NavGraph::NavGraphContainer::GraphContainer* g = graph->getGCtn();

			//--- print all edges
			if (show_edges)
			{
				Line* curEdge;
				BGL_FORALL_EDGES(e, *g, NavGraph::NavGraphContainer::GraphContainer) {
					curEdge = new Line(
							   (*g)[source(e,*g)].pos.utm_x,
							   (*g)[source(e,*g)].pos.utm_y,
							   (*g)[target(e,*g)].pos.utm_x,
							   (*g)[target(e,*g)].pos.utm_y);
					curEdge->setColor(0,102,255);
					viewer->addLine(curEdge);
				}
			}

			//--- print all vertices
			if (show_nodes)
			{
				Shape *curVertex;
				BGL_FORALL_VERTICES(v,*g, adjacency_list<>) {
					curVertex = new Shape(Shape::ShapeCross,
								(*g)[v].pos.utm_x,
								(*g)[v].pos.utm_y,
								0.6,0.6);
					curVertex->setColor(255,0,0);
					viewer->addShape(curVertex);
				}
			}

			return true;
		} /*}}}*/

		void LGLViewer::drawEdge(const NavGNode& prevNode, const NavGNode& curNode, DRAW_NODE_STATE dstate)
		{//{{{
			if ((&curNode) == NULL || (&prevNode) == NULL)
			{
				return;
			}

			drawEdge(prevNode.pos, curNode.pos, dstate);
		}//}}}

		void LGLViewer::drawEdge(const NavLocation& prevPos, const NavLocation& curPos, DRAW_NODE_STATE dstate)
		{//{{{
			Line *curLine;

			if ( withUTM )
			{
				curLine = new Line(
							  prevPos.utm_x,
							  prevPos.utm_y,
							  curPos.utm_x,
							  curPos.utm_y);
			} else {
				curLine = new Line(
				                  prevPos.pos_x * scaleFactor,
								  prevPos.pos_y * scaleFactor,
				                  curPos.pos_x * scaleFactor,
								  curPos.pos_y * scaleFactor);
			}

			switch(dstate) {
				case DRAW_NODE_PATH: // examine vertex is in the guessed path
					lowcolor  = 40;
					highcolor = 200;
					// Changing color at each new path from green to orange
					highcolor -= n_path_drawn *  (highcolor / NMAX_PATHS); 
					lowcolor  += n_path_drawn * ((255-lowcolor) / NMAX_PATHS);
					// Alternative pick of color 
					if ( (n_path_drawn/2.0) == (double)((int)n_path_drawn/2))
					{
						curLine->setColor(lowcolor, lowcolor, highcolor);
					} else {
						curLine->setColor(highcolor, lowcolor, lowcolor);
					}
					break;
				case DRAW_NODE_TRACK: // examine vertex is in the guessed path
					curLine->setColor(250,250,80);
					break;
				default:
					curLine->setColor(0,0,0);
			}

			if (curLine != NULL)
			{
				viewer->addLine(curLine);
			}

		}//}}}

		void LGLViewer::displayHook(const NavGNode& curNode, DRAW_NODE_STATE dstate)
		{//{{{

			if ((&curNode) == NULL)
			{
				return;
			}

			displayHook(curNode.pos, dstate);

		}//}}}

		void LGLViewer::displayHook(const NavLocation& curPos, DRAW_NODE_STATE dstate)
		{/*{{{*/
			Shape *curBack;
			Shape *curShape;
			Shape *curCore;
			Shape *curRadius;
			double tmp = 0.0;

			if ( withUTM )
			{
				curBack = new Shape(Shape::ShapeEllipseFilled,
								  curPos.utm_x,
				                  curPos.utm_y,
								  0.4, 0.4);
				curShape = new Shape(Shape::ShapeEllipseFilled,
								  curPos.utm_x,
				                  curPos.utm_y,
								  0.35, 0.35);
				curCore = new Shape(Shape::ShapeEllipseFilled, 
				                  curPos.utm_x * scaleFactor,
								  curPos.utm_y * scaleFactor,
								  0.3, 0.3);
				if(curPos.safeRadius  < 0.0) { tmp = 0.0; }
				else { tmp = 2.0*curPos.safeRadius; }
				curRadius = new Shape(Shape::ShapeEllipse,
								  curPos.utm_x,
								  curPos.utm_y,
								  tmp, tmp);
			} else {
				curBack = new Shape(Shape::ShapeEllipseFilled, 
				                  curPos.pos_x * scaleFactor,
								  curPos.pos_y * scaleFactor,
								  0.5, 0.5);
				curShape = new Shape(Shape::ShapeEllipseFilled, 
				                  curPos.pos_x * scaleFactor,
								  curPos.pos_y * scaleFactor,
								  0.4, 0.4);
				curCore = new Shape(Shape::ShapeEllipseFilled, 
				                  curPos.pos_x * scaleFactor,
								  curPos.pos_y * scaleFactor,
								  0.3, 0.3);
				if(curPos.safeRadius  < 0.0) { tmp = 0.0; }
				else { tmp = 2.0*curPos.safeRadius; }
				curRadius = new Shape(Shape::ShapeEllipse,
								  curPos.pos_x,
								  curPos.pos_y,
								  tmp, tmp);
			}


			switch(dstate) {
				case DRAW_NODE_OPEN: // discover_vertex has just been added to the OPEN List
					curBack->setColor(40,250,40);
					curShape->setColor(40,40,40);
					if ( (n_path_drawn/2.0) == (double)((int)n_path_drawn/2))
					{ curCore->setColor(lowcolor, lowcolor, highcolor); }
					else { curCore->setColor(highcolor, lowcolor, lowcolor); }
					break;
				case DRAW_NODE_VISITED: // examine vertex has just been popped out the OPEN List
					curBack->setColor(50,50,250);
					curShape->setColor(250,190,40);
					if ( (n_path_drawn/2.0) == (double)((int)n_path_drawn/2))
					{ curCore->setColor(lowcolor, lowcolor, highcolor); }
					else { curCore->setColor(highcolor, lowcolor, lowcolor); }
					break;
				case DRAW_NODE_PATH: // examine vertex is in the guessed path
					curBack->setColor(50,50,10);
					curShape->setColor(200,255,200);
					if ( (n_path_drawn/2.0) == (double)((int)n_path_drawn/2))
					{ curCore->setColor(lowcolor, lowcolor, highcolor); }
					else { curCore->setColor(highcolor, lowcolor, lowcolor); }
					break;
				case DRAW_NODE_TRACK: // examine vertex is in the guessed path
					curBack->setColor(200,200,200);
					curShape->setColor(250,250,80);
					if ( (n_path_drawn/2.0) == (double)((int)n_path_drawn/2))
					{ curCore->setColor(lowcolor, lowcolor, highcolor); }
					else { curCore->setColor(highcolor, lowcolor, lowcolor); }
					break;
				default:
					curShape->setColor(0,0,0);
			}

			curRadius->setColor(50,180,250);

			// Finishing display
			if (curShape != NULL)
			{
				viewer->addShape(curRadius);
				viewer->addShape(curBack);
				viewer->addShape(curShape);
				viewer->addShape(curCore);
			}

		}/*}}}*/

		void LGLViewer::displayFrontierPoint(const NavGraph::frontierPoint& F, double normalizer1, double normalizer2)
		{/*{{{*/
			Shape *curBack;
			Shape *curShape;
			Shape *curCore;

			// Manual corrections
			double coreSizeFactor = 0.0;
			bool tooStrongError = false;

			if (F.second == -1.0) 
			{
				// No cost has been calculated on this frontier point
				coreSizeFactor = 1.0;
			} else {
				if (normalizer1 <= 0.0 ) { normalizer1 = 0.0; }
				if (normalizer2 <= 0.0 ) { normalizer2 = 1000.0; }
				// Avoiding too narrow scales.
				if (normalizer1 == normalizer2) { normalizer1 = 0.0; normalizer2 = 1000.0; }
				// Setting the core size factor
				if ( F.second > normalizer2 || F.second == std::numeric_limits<double>::max() )
				{ coreSizeFactor = 1.0; tooStrongError = true; }
				else
				{ coreSizeFactor = (F.second - normalizer1) / (normalizer2-normalizer1); }
			}

			NavGraph::NavGraphContainer::GraphContainer* graphee = graph->getGCtn();
			NavLocation curPos = ((*graphee)[F.first]).pos;

			if ( withUTM )
			{
				curBack = new Shape(Shape::ShapeRectangleFilled, 
								  curPos.utm_x,
				                  curPos.utm_y,
								  0.4, 0.4);
				curShape = new Shape(Shape::ShapeRectangleFilled, 
								  curPos.utm_x,
				                  curPos.utm_y,
								  0.38, 0.38);
				curCore = new Shape(Shape::ShapeRectangleFilled, 
				                  curPos.utm_x * scaleFactor,
								  curPos.utm_y * scaleFactor,
								  0.35*coreSizeFactor, 0.35*coreSizeFactor);
			} else {
				curBack = new Shape(Shape::ShapeRectangleFilled, 
				                  curPos.pos_x * scaleFactor,
								  curPos.pos_y * scaleFactor,
								  0.5, 0.5);
				curShape = new Shape(Shape::ShapeRectangleFilled, 
				                  curPos.pos_x * scaleFactor,
								  curPos.pos_y * scaleFactor,
								  0.45, 0.45);
				curCore = new Shape(Shape::ShapeRectangleFilled, 
				                  curPos.pos_x * scaleFactor,
								  curPos.pos_y * scaleFactor,
								  0.4*coreSizeFactor, 0.4*coreSizeFactor);
			}



			// Dark grey rectangle ring
			curBack->setColor(40,40,40);
			// Yellow second ring/core
			curShape->setColor(240,240,40);
			if (F.second == -1.0) 
			{
				// grey core if no cost has been calculated
				curCore->setColor(40,40,40);
			} else {
				if (tooStrongError)
				{
					// light grey core if cost value is too strong
					curCore->setColor(180,180,180);
				} else {
					// Red core
					curCore->setColor(240,40,40);
				}
			}

			// Finishing display
			if (curShape != NULL)
			{
				viewer->addShape(curBack);
				viewer->addShape(curShape);
				viewer->addShape(curCore);
			}

		}/*}}}*/

	}
}

#endif
