/**
 * \file NavGraph.cpp
 *
 * Source file defining the navigation graph object
 *
 * \author Redouane Boumghar, Hung Cao
 *
 * \ingroup lgl
 *
 **/

#include <lgl/NavGraph.hpp>

using namespace std;
using namespace boost;

namespace jafar { //namespace jafar
  namespace lgl { //namespace lgl
	 
		NavGraph::NavGraph():
			geodata(NULL),
			quadtree(NULL),
			doNavGraphPruning(PRUNING_DISABLED),
			pruningThreshold(DEFAULT_PRUNING_THRESHOLD),
			doPathPruning(PRUNING_DISABLED),
			posManager(NULL),
			costManager(NULL)
		{//{{{
			verbose = VERBOSE_WARNINGS;
			m_quadtree_madeinhere = false;
			m_costmanager_madeinhere = false;
			m_posmanager_madeinhere = false;
			set_geodata_grp = true;
			set_geodata_gru = true;
			calculate_grp   = true;
			calculate_navutil = true;
			pathCost = -1.0;
			coopPathCost = -1.0;
		}//}}}

		NavGraph::NavGraph(GeoData* _geodata,
		  Quadtree<double,GeoData>* _quadtree,
		  bool _doNavGraphPruning,
		  double _pruningThreshold, 
		  PositionManager* _posManager,
		  CostManager* _costManager,
		  boost::function<void(const NavGNode&, int)> _displayHook 
		  ):
			geodata(_geodata),
			quadtree(_quadtree),
			doNavGraphPruning(_doNavGraphPruning),
			pruningThreshold(_pruningThreshold),
			doPathPruning(_doNavGraphPruning),
			posManager(_posManager),
			costManager(_costManager)
		{ //{{{	
			verbose = VERBOSE_WARNINGS;
			m_quadtree_madeinhere = false;
			m_costmanager_madeinhere = false;
			m_posmanager_madeinhere = false;
			set_geodata_grp = true;
			set_geodata_gru = true;
			calculate_grp   = true;
			calculate_navutil = true;
			pathCost = -1.0;
			coopPathCost = -1.0;

			// Checking geodata
			if (geodata == NULL)
			{
				std::cout << "#EEE# [NavGraph] No geodata available to create navigation graph." << std::endl;
				return;
			}

			// display hook
			this->displayHook = _displayHook;

			// Create quadtree from geodata
			if (quadtree == NULL)
			{
				m_quadtree_madeinhere = true;
				quadtree = new Quadtree<double, GeoData>(geodata);
			}

			// Create position manager in case not given
			if (posManager == NULL)
			{
				m_posmanager_madeinhere = true;
				posManager =  new PositionManager(geodata, quadtree); 
			}

			// Create cost manager in case not given
			if (costManager == NULL && posManager != NULL)
			{
				m_costmanager_madeinhere = true;
				costManager = new CostManager(geodata, quadtree, posManager);
				costManager->setObstacleLimit(0.8);
			}

			// create an object graph
			graph = new NavGraphContainer((geodata->getxsize()*2+1),(geodata->getysize()*2+1));
		
			// create an A* planner
			planner = new AStarPlanner();

			// In PosManager manner
			m_closeDistance = 40.0;

			// Initializing terrain classes weights/factor
			//terrainFactor = NULL;

			// build the graph
			if (!buildGraph())
			{
				std::cout << "#EEE# [NavGraph] Could not create navigation graph." << std::endl;
				return;
			}
		} //}}}
	 
	 	// Destructor
		NavGraph::~NavGraph() 
		{ // {{{ 
			// terrainFactor is already deleted by the garbage collector
			delete planner;
			delete graph;
			if (m_costmanager_madeinhere)
			delete costManager; 
			if (m_posmanager_madeinhere)
			delete posManager;
			if (m_quadtree_madeinhere)
			delete quadtree; 
		} // }}}

		/**
		 * Build graph over the terrain using the quadtree/decomp information
		 * 
		 */
		bool NavGraph::buildGraph()
		{ /*{{{*/
			int i = 0;
			int j = 0;
			int m = 0; // to avoid messing up with index i
			int n = 0; // to avoid messing up with index j
			int m_end = 0;
			int n_end = 0;
			int nbLocs;
			vector<NavLocation> inOut;
			RasterCellIndex clusterRootIndex;
			Decomp *decomp = quadtree->getDecomp();
			NavGEdge edge;
			NavGNode node;
			NavGraphContainer::Vertex tmp_v;
			double weight = 0.0;
			// UTM origin to take from geodata
			double utmXRoot = 0.0;
			double utmYRoot = 0.0;
			// Cell resolutions to take from geodata
			double cellXRes = 0.0;
			double cellYRes = 0.0;

			// Setting an easy  usable graph container
			NavGraph::NavGraphContainer::GraphContainer* graphee = graph->getGCtn();

			// Fill terrainFactor array with weight accross terrain classes
			//calculateTerrainFactor(); // UNUSED

			// To find the borders of the cell to which in and out points belong
			RasterRect qnl_rect;

			// Get the geo reference affine (UTM like) transformation
			geodata->getAffineTransform(utmXRoot, utmYRoot, cellXRes, cellYRes);

			// for each cluster inside the environnment
			Decomp_iterator it(decomp);
			for(it.begin();!it.at_end();++it) {
				// for each cluster in the decomp
				// Decomp Cluster curCluster=*(it);
				it.getIndex(m, n);

				// faster way => get the geo data for the cluster instead of computing it every time 
		  
				// get vector of INOUT points from this cluster
				clusterRootIndex.set(m, n);

				// Do prune the navigation graph if the option is enabled
				if ( doNavGraphPruning && ((*it)->getCertainty(T_OBSTACLE) > pruningThreshold) ) 
					continue ;
				
				if (decomp->getInOutPoints(clusterRootIndex,inOut)) {
					// NODES ---------- Populate the graph with these in and out points
					nbLocs = (int)inOut.size();
					for (i=0;i<nbLocs;i++) {
						// ---- Creating --- and Getting the corresponding vertex
						tmp_v = graph->addVertex(inOut[i]);
						// Set the UTM position according to geodata's parameters
						((*graphee)[tmp_v]).pos.updateUTMPos(utmXRoot, utmYRoot, cellXRes, cellYRes);
						//inOut[i].updateUTMPos(utmXRoot, utmYRoot, cellXRes, cellYRes);
					}

					// EDGES ----------- Getting the rasterRect corresponding to the current leaf
					if (quadtree->getClusterForCell(clusterRootIndex, qnl_rect))
					{
						// Transforms cluster indexes into cluster position
						m = m*2;
						n = n*2;
						m_end = (1+ qnl_rect.xend()) * 2;
						n_end = (1+ qnl_rect.yend()) * 2;
						// For each cluster positions connect to every other
						for (i=0;i<nbLocs;i++) {
							for (j=i+1;j<nbLocs;j++) {
								// Checking if navigation locations (gotten from getInOutPoints) are
								// connectable or not. Position on the same cluster edge cannot be connected
								// as we won't be able to assign a weigh to the edge (vertex(position)-connector).
								if ( !((inOut[i].pos_x == m) && (inOut[j].pos_x == m))
								&&   !((inOut[i].pos_y == n) && (inOut[j].pos_y == n))
								&&   !((inOut[i].pos_x == m_end) && (inOut[j].pos_x == m_end))
								&&   !((inOut[i].pos_y == n_end) && (inOut[j].pos_y == n_end))
								)
								{
									//--- build edge from i to j
									edge.cluster = clusterRootIndex;
									// Calculate edge's weight and set it
									costManager->weight(clusterRootIndex, inOut[i], inOut[j], weight);
									edge.weight = weight;
									// Copy this Edge Bundle to the Graph on both sides
									graph->addEdge(inOut[i],inOut[j],edge);
									graph->addEdge(inOut[j],inOut[i],edge);

									//--- build edge from j to i
									// Cluster didn't change
									//edge.cluster=clusterRootIndex;
									//
									// the weight didn't change.
									//costManager->weight(inOut[j],inOut[i], navCostAtNode(inOut[i]), weight);
									//edge.weight=weight;
									// Copy this Edge Bundle to the Graph
									//graph->addEdge(inOut[j],inOut[i],edge);
								}
							}
						}
						//			 graph->connectInOutLocs(inOut);
					}
				}
				inOut.clear();
			}

			return true;
		} /*}}}*/

		bool NavGraph::rebuildGraph()
		{//{{{
			graph->clear();
			return buildGraph();
		}//}}}

		NavGraph::NavGraphContainer* NavGraph::getGraph() 
		{ //{{{
			return graph;
		} //}}}
	 
		Quadtree<double,GeoData>* NavGraph::getQuadtree() 
		{ //{{{
			return quadtree;
		} //}}}
	 
		//---------------------------------------------------- SETTERS
		void NavGraph::calculateNavigationUtility(bool bsetter)
		{//{{{
			calculate_navutil = bsetter;
		}//}}}

		void NavGraph::calculateGoalReachingPotential(bool bsetter)
		{//{{{
			calculate_grp = bsetter;
		}//}}}

		void NavGraph::setGeodataPotential(bool bsetter)
		{//{{{
			set_geodata_grp = bsetter;
		}//}}}

		void NavGraph::setGeodataUtility(bool bsetter)
		{//{{{
			set_geodata_gru = bsetter;
		}//}}}

		void NavGraph::setDisplayHookFunction(boost::function<void(const NavGNode&, int)> _displayHook)
		{/*{{{*/
			this->displayHook =  _displayHook;
		}/*}}}*/

		//---------------------------------------------------- GETTERS
		double NavGraph::getPathCost()
		{//{{{
			 return pathCost;
		}//}}}

		double NavGraph::getCoopPathCost()
		{//{{{
			 return coopPathCost;
		}//}}}

		int NavGraph::getNumberOfVertices()
		{/*{{{*/
			 return (int)(boost::num_vertices(*(graph->getGCtn())));
		}/*}}}*/

		int NavGraph::getNumberOfEdges()
		{/*{{{*/
			 return (int)(boost::num_edges(*(graph->getGCtn())));
		}/*}}}*/

		void NavGraph::getNavGNodeFromNavLocation(NavGNode& nnode, const NavLocation& nl)
		{/*{{{*/
			graph->getVertexBundle(graph->getVertex(nl), nnode);
		}/*}}}*/

		NavLocation NavGraph::getNavLocFromVertex(const NavGraphContainer::Vertex& A)
		{ //{{{
			NavGraphContainer::GraphContainer* graphee = graph->getGCtn();
			return ((*graphee)[A]).pos;
		} //}}}

		void NavGraph::getOffClusterFromVertex(const NavGraphContainer::Vertex& A, DecompCluster** outdcluster, int& i, int& j)
		{ //{{{
			NavGraphContainer::GraphContainer* graphee = graph->getGCtn();
			DecompCluster* maxdcluster = NULL;
			DecompCluster* dcluster = NULL;
			double maxClusterEntropy = 0.0;
			NavGraphContainer::out_edge_iterator ei, edge_end;
			for (tie(ei,edge_end) = out_edges(A, *graphee); ei != edge_end; ++ei)
			{
				if ( posManager->getCluster( ((*graphee)[(*ei)]).cluster,   &dcluster ) )
				{
					// Return the neighbor cluster with the max entropy
					if ( maxdcluster == NULL) { maxdcluster = dcluster; }
					if ((dcluster->getAttributes())->infotropy[jafar::lgl::T_SYMBOL] >= maxClusterEntropy)
					{
						maxClusterEntropy = (dcluster->getAttributes())->infotropy[jafar::lgl::T_SYMBOL];
						maxdcluster = dcluster;
						i = ((*graphee)[(*ei)]).cluster.i;
						j = ((*graphee)[(*ei)]).cluster.j;
					}
				}
			}
			(*outdcluster) = maxdcluster;
		} //}}}

		std::list<jafar::lgl::NavGraph::frontierPoint> NavGraph::getFrontiers()
		{//{{{
			return frontiers;
		}//}}}

		bool NavGraph::isSetGeodataPotential()
		{//{{{
			return set_geodata_grp;
		}//}}}

		bool NavGraph::isCalculateGRP()
		{ //{{{
			return calculate_grp;
		} //}}}

		//---------------------------------------------------- UTILS
		double NavGraph::euclideanDistance(const NavGraphContainer::Vertex& a, const NavGraphContainer::Vertex& b)
		{/*{{{*/
			NavGNode nodea;
			graph->getVertexBundle(a, nodea);
			NavGNode nodeb;
			graph->getVertexBundle(b, nodeb);

			double tmpx = nodea.pos.utm_x - nodeb.pos.utm_x;
			double tmpy = nodea.pos.utm_y - nodeb.pos.utm_y;
			return sqrt ( tmpx*tmpx + tmpy*tmpy );
		}/*}}}*/

		bool NavGraph::areClose(const NavGraphContainer::Vertex& a, const NavGraphContainer::Vertex& b)
		{/*{{{*/
			NavGNode nodea;
			graph->getVertexBundle(a, nodea);
			NavGNode nodeb;
			graph->getVertexBundle(b, nodeb);

			double tmpx = (double)nodea.pos.pos_x - (double)nodeb.pos.pos_x;
			double tmpy = (double)nodea.pos.pos_y - (double)nodeb.pos.pos_y;
			if ( m_closeDistance < 0.0 || sqrt ( tmpx*tmpx + tmpy*tmpy ) <= m_closeDistance )
			{
				return true;
			}
			return false;
		}/*}}}*/

		void NavGraph::addNode(const NavLocation& atThisLoc, NavGraphContainer::Vertex& vloc)
		{//{{{
			// Total number of connexion to do.
			int nbLocs = 0;
			double weight = 0.0;
			vector<NavLocation> graphEnterPoints;
			NavGraph::NavGraphContainer::GraphContainer* graphee = graph->getGCtn();
			NavGraphContainer::Edge elink;
			RasterCellIndex locClusterRootIndex;
			double utmXRoot;
			double utmYRoot;
			double cellXRes;
			double cellYRes;

			// Get the geo reference affine (UTM like) transformation
			geodata->getAffineTransform(utmXRoot, utmYRoot, cellXRes, cellYRes);

			// Localise node cluster root
			quadtree->getClusterForCell(RasterCellIndex((atThisLoc.pos_x/2), (atThisLoc.pos_y/2)), locClusterRootIndex);
			// Start Location does not exist. Finding nodes on cluster borders
			posManager->getGEnterPos(atThisLoc, graphEnterPoints);

			nbLocs = (int)graphEnterPoints.size();
			// Using any (the first one) enter points to check if it is
			// connected to any unborder-but-inside-cluster node
			if (nbLocs > 0)
			{
				NavGraphContainer::Vertex tmpv;
				int xmin = locClusterRootIndex.i * 2;
				int xmax = 0;
				int ymin = locClusterRootIndex.j * 2;
				int ymax = 0;
				// Getting the max bounds on possible coordinates
				for (int i=0;i<nbLocs;i++) {
					if (graphEnterPoints[i].pos_x > xmax)
					{ xmax = graphEnterPoints[i].pos_x; }
					if (graphEnterPoints[i].pos_y > ymax)
					{ ymax = graphEnterPoints[i].pos_y; }
				}
				//getInsidePoints(locClusterRootIndex, graphEnterPoints[0], graphEnterPoints);
				graph->getVertex(graphEnterPoints[0], tmpv, false);
				NavGraphContainer::out_edge_iterator ei, edge_end;
				for (tie(ei,edge_end) = out_edges(tmpv, *graphee); ei != edge_end; ++ei) {
					// Check if target node is in the cluster (not including borders)
					if ( 
						   ((*graphee)[target((*ei), *graphee)].pos.pos_x) > xmin
						&& ((*graphee)[target((*ei), *graphee)].pos.pos_x) < xmax
						&& ((*graphee)[target((*ei), *graphee)].pos.pos_y) > ymin
						&& ((*graphee)[target((*ei), *graphee)].pos.pos_y) < ymax
					   )
					{
						// Add this target node to the list of to-be-connected nodes
						graphEnterPoints.push_back((*graphee)[target((*ei), *graphee)].pos);
					}
				}
			}

			// Connecting : creation of edge and its source and target nodes if non-existant.
			graph->connect(atThisLoc, graphEnterPoints);

			// Getting the new node vertex
			graph->getVertex(atThisLoc, vloc, false);

			// Updating UTM coordinates according to the current geodata
			((*graphee)[vloc]).pos.updateUTMPos(utmXRoot, utmYRoot, cellXRes, cellYRes);

			//--- Modifying the newly created edges :  updating their weight
			nbLocs = (int)graphEnterPoints.size();
			for (int i=0;i<nbLocs;i++) {
				// Get handle on the current edge
				elink = graph->getEdge(vloc, graphEnterPoints[i]);
				((*graphee)[elink]).cluster = locClusterRootIndex;
				//costManager->weight(atThisLoc, graphEnterPoints[i], weight);
				costManager->weight(locClusterRootIndex, atThisLoc, graphEnterPoints[i], weight);
				((*graphee)[elink]).weight = weight;
				// Dual edge
				elink = graph->getEdge(graphEnterPoints[i], vloc);
				((*graphee)[elink]).cluster = locClusterRootIndex;
				//costManager->weight(graphEnterPoints[i], atThisLoc, weight);
				// Wait didn't change since we are in the same cluster
				((*graphee)[elink]).weight = weight;
			}
		}//}}}

		bool NavGraph::isOnFrontier(NavGraphContainer::Vertex& vloc)
		{//{{{
			DecompCluster* dcluster;
			NavGraph::NavGraphContainer::GraphContainer* graphee = graph->getGCtn();
			NavGraphContainer::out_edge_iterator ei, edge_end;
			int locState = 0; // 0: undefined, 1:unknown land, 2:known land
			for (tie(ei,edge_end) = out_edges(vloc, *graphee); ei != edge_end; ++ei)
			{
				if ( posManager->getCluster( ((*graphee)[(*ei)]).cluster,   &dcluster  ) )
				{
					if ((dcluster->getAttributes())->infotropy[jafar::lgl::T_SYMBOL] >= 0.6290)
					{//{{{
						if (locState == 0)
						{
							// Setting the first state as unknown land
							locState = 1;
						} else {
							if (locState != 1)
							{
								// State has changed so we must be on a frontier
								return true;
							}
						}
					}//}}}
					else
					{//{{{
						if (locState == 0)
						{
							// Setting the first state as known land
							locState = 2;
						} else {
							if (locState != 2)
							{
								// State has changed so we must be on a frontier
								return true;
							}
						}
					}//}}}
				}

			}
			return false;
		}//}}}

		void NavGraph::setNavGraphPruningOption(bool _doNavGraphPruning)
		{/*{{{*/
			doNavGraphPruning=_doNavGraphPruning ;
		}/*}}}*/

		bool NavGraph::getNavGraphPruningOption()
		{/*{{{*/
			return ( doNavGraphPruning );
		}/*}}}*/

		void NavGraph::setPathPruningOption(bool _doPathPruning)
		{//{{{
			doPathPruning = _doPathPruning ;
		}//}}}

		bool NavGraph::getPathPruningOption()
		{//{{{
			return ( doPathPruning );
		}//}}}

		void NavGraph::setPruningThreshold(double _pruningThreshold) 
		{/*{{{*/
			pruningThreshold=_pruningThreshold;
		}/*}}}*/ 

		double NavGraph::getPruningThreshold()
		{/*{{{*/
			return ( pruningThreshold ) ;
		}/*}}}*/

		PositionManager* NavGraph::getPositionManager() 
		{ //{{{
			return posManager;
		} //}}}

		double NavGraph::getNSP(const NavGraphContainer::Vertex& v)
		{//{{{
			// Check if Vertex exists is not really necessary now 
			// TODO: should be checked with a more dynamic graph.
			NavGraphContainer::GraphContainer* graphee = graph->getGCtn();
			return (*graphee)[v].grp;
		}//}}}

		//---------------------------------------------------------//
		bool NavGraph::findPath(const NavLocation& userStart, const NavLocation& userEnd, NavPath& userNavpath, double& pCost) 
		{ //{{{
			bool areturn = findPath(userStart, userEnd, userNavpath);
			pCost = pathCost;
			return areturn;
		} //}}}

		bool NavGraph::findPath(const NavLocation& userStart, const NavLocation& userEnd, NavPath& userNavpath) 
		{ //{{{
			NavGraphContainer::Vertex vstart, vend;
			NavGraphContainer::Edge elink;
			int n = 0;
			double vpot = 0.0;
			//double vutil_max= 0.0;
			double vpot_max = 0.0;
			NavGNode curNode;

			if (calculate_grp)
			{ //{{{
				// Clear frontiers
				frontiers.clear();

				// Unset and Reset each graph-node potential
				resetGoalReachingPotential();
				if ( set_geodata_grp )
				{
					geodata->resetUtility();
				}
			} //}}}

			
			// Connect start and goal point to the graph if they are not already part of the graph's vertex
			//------------------------------------------------------------- START 
			if (   ((userStart.pos_x + 1)/2) > geodata->getxsize() 
				|| ((userStart.pos_x + 1)/2) < geodata->getxroot() 
				|| ((userStart.pos_y + 1)/2) > geodata->getysize() 
				|| ((userStart.pos_y + 1)/2) < geodata->getyroot() )
			{
				// Start point is not contained in the world
				NavLocation LocStart = graph->getClosestLocation(userStart);
				if (verbose >= VERBOSE_WARNINGS)
				{
					std::cout << "#WWW# [NavGraph::findPath] Out of Graph Position : Path LocStart Closest : " 
							  << LocStart << std::endl;
				}
				graph->getVertex(LocStart, vstart, false);
			} else { //{{{
				if (! graph->vertexExists(userStart)) {
					if (verbose >= VERBOSE_INFO)
					{
						std::cout
						<< "#III#  [NavGraph::findPath] start position out of graph, creating position : "
					    << userStart << std::endl;
					}

					// Checking if pruning is on and adapting addition
					// But first lets get the decomposition cluster.
					DecompCluster* dcluster = NULL;
					posManager->getClusterUnordered(userStart, &dcluster);
					if (dcluster == NULL)
					{
						// This should normally be dealt by the oldest if-test parent
						NavLocation LocStart = graph->getClosestLocation(userStart);
						graph->getVertex(LocStart, vstart, false);
						if (verbose >= VERBOSE_WARNINGS)
						{
							std::cout
							<< "#WWW#  [NavGraph::findPath] no decomposition cluster for start position. Using closest point: "
							<< LocStart << std::endl;
						}
					} else {
						// If obstacle pruning is on
						if ( doNavGraphPruning && (dcluster->getCertainty(T_OBSTACLE) > pruningThreshold) ) 
						{
							if (doPathPruning)
							{
								NavLocation LocStart = graph->getClosestLocation(userStart);
								graph->getVertex(LocStart, vstart, false);
								if (verbose >= VERBOSE_WARNINGS)
								{
									std::cout
									<< "#WWW#  [NavGraph::findPath] can not add start position in pruned zone. Using closest point: "
									<< LocStart << std::endl;
								}
							} else {
								if (verbose >= VERBOSE_WARNINGS)
								{
									std::cout
									<< "#WWW#  [NavGraph::findPath] No possible path. Unconnected positions." << std::endl;
								}
								pathCost = -1.0;
								return false;
							}
						} else {
							addNode(userStart, vstart);
							if (verbose >= VERBOSE_INFO)
							{
								std::cout
								<< "#III#  [NavGraph::findPath] adding new start position at: " << userStart << std::endl;
							}
						}
					}
				} else {
					NavLocation LocStart = graph->getClosestLocation(userStart);
					if (verbose >= VERBOSE_INFO)
					{
						std::cout 
						<< "#III# [NavGraph::findPath] Using existing start position: " 
						<< LocStart << std::endl;
					}
					graph->getVertex(LocStart, vstart, false);
				}
			} //}}}

			//------------------------------------------------------------- GOAL
			if (   ((userEnd.pos_x + 1)/2) > geodata->getxsize() 
				|| ((userEnd.pos_x + 1)/2) < geodata->getxroot() 
				|| ((userEnd.pos_y + 1)/2) > geodata->getysize() 
				|| ((userEnd.pos_y + 1)/2) < geodata->getyroot() )
			{
				// Goal point is not contained in the world
				NavLocation LocEnd = graph->getClosestLocation(userEnd);
				if (verbose >= VERBOSE_WARNINGS)
				{
					std::cout << "#WWW# [NavGraph::findPath] Out of Graph Position : Path LocEnd Closest : "
							  << LocEnd << std::endl;
				}
				graph->getVertex(LocEnd, vend, false);
			} else { //{{{
				if (! graph->vertexExists(userEnd)) {
					if (verbose >= VERBOSE_INFO)
					{
						std::cout
						<< "#III#  [NavGraph::findPath] goal position out of graph, creating position : "
					    << userEnd << std::endl;
					}

					// Checking if pruning is on and adapting addition
					// But first lets get the decomposition cluster.
					DecompCluster* dcluster = NULL;
					posManager->getClusterUnordered(userEnd, &dcluster);
					if (dcluster == NULL)
					{
						// This should normally be dealt by the oldest if-test parent
						NavLocation LocEnd = graph->getClosestLocation(userEnd);
						graph->getVertex(LocEnd, vend, false);
						if (verbose >= VERBOSE_WARNINGS)
						{
							std::cout
							<< "#WWW#  [NavGraph::findPath] no decomposition cluster for goal position. Using closest point: "
							<< LocEnd << std::endl;
						}
					} else {
						// If obstacle pruning is on
						if ( doNavGraphPruning && (dcluster->getCertainty(T_OBSTACLE) > pruningThreshold) ) 
						{
							if (doPathPruning)
							{
								NavLocation LocEnd = graph->getClosestLocation(userEnd);
								graph->getVertex(LocEnd, vend, false);
								if (verbose >= VERBOSE_WARNINGS)
								{
									std::cout
									<< "#WWW#  [NavGraph::findPath] can not add goal position in pruned zone. Using closest point: "
									<< LocEnd << std::endl;
								}
							} else {
								if (verbose >= VERBOSE_WARNINGS)
								{
									std::cout
									<< "#WWW#  [NavGraph::findPath] No possible path. Unconnected positions." << std::endl;
								}
								pathCost = -1.0;
								return false;
							}
						} else {
							addNode(userEnd, vend);
							if (verbose >= VERBOSE_INFO)
							{
								std::cout
								<< "#III#  [NavGraph::findPath] adding new goal position at: " << userEnd << std::endl;
							}
						}
					}
				} else {
					NavLocation LocEnd = graph->getClosestLocation(userEnd);
					if (verbose >= VERBOSE_INFO)
					{
						std::cout 
						<< "#III# [NavGraph::findPath] Using existing goal position: " 
						<< LocEnd << std::endl;
					}
					graph->getVertex(LocEnd, vend, false);
				}
			} //}}}
	  
			//------------------------------------------------------------- A*
			// Getting weight map with boost get function
			NavGraphContainer::WeightMap weightmap( get(&NavGEdge::weight,*(graph->getGCtn())) );

			// Buffer Map to get the calculated cost F
			std::vector<double> resultingFullCost( boost::num_vertices(*(graph->getGCtn())), -1.0 );

			// Buffer predecessor map
			std::vector<NavGraphContainer::Vertex> predecessorMap( boost::num_vertices(*(graph->getGCtn())) );

			/*
			 * Buffer node colors to check which is open closed or neither
			 * default_color_type can be { white_color, gray_color, green_color, red_color, black_color };
			 * standing for {default_start, open, ?, ? , closed };
			 */
			std::vector<boost::default_color_type> colorMap( boost::num_vertices(*(graph->getGCtn())) );

			// List of open nodes: constructed on the fly when checking for nodes' color
			//std::list<NavGraphContainer::Vertex> backPointedList;
			std::list<NavGraphContainer::Vertex> backPointedList;

			// NOT GOOD DEFINITIONS
			//std::map<NavGraphContainer::Vertex, double> resultingFullCost;
			//std::map<NavGraphContainer::Vertex, NavGraphContainer::Vertex> predecessorMap;

			// Custom Heuristic
			nav_heuristic<NavGraphContainer::GraphContainer, COST_TYPE, NavGraphContainer::Vertex> 
			    navHeuristic (*(graph->getGCtn()),vend);

			//------------------------------------------------- A* SEARCH 
			GraphPath<NavGraphContainer::Vertex> graphpath =  planner->AStarSearch(
										navHeuristic,
										*(graph->getGCtn()),
										weightmap,
										vstart,
										vend,
										resultingFullCost,
										predecessorMap,
										colorMap,
										boost::ref(this->displayHook)
										);
  
  			//---- Saving the resulting path cost
			pathCost = resultingFullCost[vend];

			NavGraph::NavGraphContainer::GraphContainer* graphee = graph->getGCtn();
			//------------------------------------------------------- GRP
			if (calculate_grp)
			{ //{{{
				//--- calculating map path alternative potential  or Goal Reaching Potential
				vpot_max = 0.0;
				double goaldist = 0.0;
				BGL_FORALL_VERTICES(v, (*graphee), adjacency_list<>)
				{
					// Check if frontier vertice
					if (isOnFrontier(v))
					{
						if (colorMap[v] == white_color) // nodes is undiscovered
						{
							extendFrontiers(v, -1.0);
						} else {
							// Add this vertex in the ordered list of frontier vertices.
							extendFrontiers(v, resultingFullCost[v]);
						}
					}

					// Checking if v is an open vertex
					// These vertices are start points 
					//if (colorMap[v] == black_color) // examined nodes are black
					if (colorMap[v] == gray_color) // nodes in the open list are grey
					{

						if ( true) //areClose(v, vend) )
						{
							// and v is "m_closeDistance" to the goal
							vpot = resultingFullCost[v];
							goaldist = euclideanDistance(v, vend);
							vpot = vpot + goaldist;
							if(vpot > 0.0)
							{
								vpot = 1.0/(vpot);
							} else { vpot = 1.0/ATT_MIN_VAL; }
							
							// Saving max for future normalization
							// DO NOT TRY TO GET MAX HERE : vpot is accumulating
							// so no max will be found here.
							//
							// Back pointing v cost from v to all its predecessors
							for(NavGraphContainer::Vertex vi = v;; vi = predecessorMap[vi]) {
								// Adding this Node to the list of nodes which
								// potential is to be updated
								if ( !((*graphee)[vi].hasPotential) )
								{
									backPointedList.push_back(vi);
								}

								// Not copying the node. Directly working on the graph.
								(*graphee)[vi].setPotential(vpot);

								// end loop condition 
								if(predecessorMap[vi] == vi)
								{
									break;
								}
							}
						}
					}
				} 

				// The maximum potential is by definition on the start point.
				// (no need to calculate it elseway)
				vpot_max = (*graphee)[vstart].grp;

				//--- calculating map potential : normalization
				std::list<NavGraphContainer::Vertex>::iterator bplit;
				for(bplit = backPointedList.begin(); bplit != backPointedList.end(); ++bplit)
				{
					if ( (*graphee)[(*bplit)].hasPotential && vpot_max > 0.0)
					{
						// Normalizing and updating potential
						// Linear normalization
						vpot = (*graphee)[(*bplit)].grp/vpot_max;
						// Polynomial normalization
						//vpot = (*graphee)[(*bplit)].getCertainty(U_POT);
						//vpot = (vpot_max - vpot)/vpot_max;
						//vpot = 1.0 - vpot*vpot;

						(*graphee)[(*bplit)].hasPotential = false;
						(*graphee)[(*bplit)].setPotential(vpot);
					} else {
						// Resetting potential
						(*graphee)[(*bplit)].hasPotential = false;
						(*graphee)[(*bplit)].setPotential(0.0);
						(*graphee)[(*bplit)].hasPotential = false;
					}
				}
			} //}}}// eoi (calculate_grp)

			//--- Setting geodata
			if ( calculate_grp && set_geodata_grp )
			{ //{{{
				std::list<NavGraphContainer::Vertex>::iterator bplit;
				for(bplit = backPointedList.begin(); bplit != backPointedList.end(); ++bplit)
				{
					// Interpolating in geodata
					if (predecessorMap[(*bplit)] == (*bplit))
					{
						// Do nothing. We are on the starting point
						// or on a vertex that is not reachable from the start.
					} else {
						geodata->setMap(
							POTENTIAL_MAP,
							(*graphee)[(*bplit)].grp,
							(*graphee)[(*bplit)].pos.pos_x,
							(*graphee)[(*bplit)].pos.pos_y,
							(*graphee)[predecessorMap[(*bplit)]].grp,
							(*graphee)[predecessorMap[(*bplit)]].pos.pos_x,
							(*graphee)[predecessorMap[(*bplit)]].pos.pos_y );
					}
				}
			} //}}} eoi ( calculate_grp && set_geodata_grp )

			//--- Calculating utility (goal reaching utility)
			if (calculate_grp && calculate_navutil)
			{ //{{{
				if (verbose >= VERBOSE_INFO)
				{
					std::cout << "#III# [NavGraph::findPath] Updating Utility." << std::endl;
				}
				// Calculate and Set Graph utility for next path search on utility
				updateNavigationUtility();
			} //}}}


			//-----------------------------------------------/
			// Copying nodes to fill the final navigation path without the start point
			// which is the current position of the vehicule.
			typedef GraphPath<NavGraphContainer::Vertex>::const_iterator path_iterator;  
			userNavpath.clear();
  
  			n = 0;
			NavGraphContainer::Vertex vcurr = vstart;
			NavGraphContainer::Vertex vprev = vstart;
			NavVertex navV;
			DecompCluster* dcluster;
			for(path_iterator lit=graphpath.begin(); lit!=graphpath.end(); ++lit) {	 
				vcurr = (*lit);
				// Avoiding the first vertex to not put the start point in the resulting path
				if (n != 0)
				{
					// Get a bundled copy of the path node
					graph->getVertexBundle(vcurr,curNode);
					navV.first = curNode.pos;

					// Get edge descriptor from vprev to vcurr
					elink = graph->getEdge(vprev, vcurr);
					posManager->getCluster(((*graphee)[elink]).cluster, &dcluster);
					navV.second = *(dcluster->getAttributes());

					userNavpath.push_back(navV);
				} else {
					n = 1;
				}
				vprev = vcurr;
			}
 
			return true;
		} //}}}

		//---------------------------------------------------------//
		bool NavGraph::fastFindPath(const NavLocation& userStart, const NavLocation& userEnd, NavPath& userNavpath, double& pCost) 
		{ //{{{
			bool areturn = fastFindPath(userStart, userEnd, userNavpath);
			pCost = pathCost;
			return areturn;
		} //}}}

		bool NavGraph::fastFindPath(const NavLocation& userStart, const NavLocation& userEnd, NavPath& _path) 
		{ //{{{
			// Saving states
			bool ffp_setgeodatagrp = set_geodata_grp;
			bool ffp_setgeodatagru = set_geodata_gru;
			bool ffp_calculategrp  = calculate_grp;
			bool ffp_calculatenavutil = calculate_navutil;
			// Setting state for fast path search
			set_geodata_grp = false;
			set_geodata_gru = false;
			calculate_grp   = false;
			calculate_navutil = false;
			// Running the thus fast search
			bool areturn = findPath(userStart, userEnd, _path);
			// reset state to previous one
			set_geodata_grp = ffp_setgeodatagrp;
			set_geodata_gru = ffp_setgeodatagru;
			calculate_grp   = ffp_calculategrp;
			calculate_navutil = ffp_calculatenavutil;
			return areturn;
		} //}}}

		//---------------------------------------------------------//
		bool NavGraph::findPathOnUtility(const NavLocation& userStart, const NavLocation& userEnd, NavPath& userNavpath, double authDist) 
		{ //{{{
			NavGraphContainer::Vertex vstart, vend;
			NavGraphContainer::Edge elink;
			vector<NavLocation> startEnterPoints, endOutPoints;
			int n = 0;
			NavGNode curNode;
			NavGraph::NavGraphContainer::GraphContainer* graphee = graph->getGCtn();

			// Connect start and goal point to the graph if they are not already part of the graph's vertex
			//------------------------------------------------------------- START 
			if (   ((userStart.pos_x + 1)/2) > geodata->getxsize() 
				|| ((userStart.pos_x + 1)/2) < geodata->getxroot() 
				|| ((userStart.pos_y + 1)/2) > geodata->getysize() 
				|| ((userStart.pos_y + 1)/2) < geodata->getyroot() )
			{
				// Start point is not contained in the world
				NavLocation LocStart = graph->getClosestLocation(userStart);
				if (verbose >= VERBOSE_WARNINGS)
				{
					std::cout << "#WWW# [NavGraph::findPathOnUtility] Out of Graph Position : Path start closest position: " 
							  << LocStart << std::endl;
				}
				graph->getVertex(LocStart, vstart, false);
			} else { //{{{
				if (! graph->vertexExists(userStart)) {
					if (verbose >= VERBOSE_INFO)
					{
						std::cout
						<< "#III#  [NavGraph::findPath] start position out of graph, creating position : "
					    << userStart << std::endl;
					}

					// Checking if pruning is on and adapting addition
					// But first lets get the decomposition cluster.
					DecompCluster* dcluster = NULL;
					posManager->getClusterUnordered(userStart, &dcluster);
					if (dcluster == NULL)
					{
						NavLocation LocStart = graph->getClosestLocation(userStart);
						graph->getVertex(LocStart, vstart, false);
						if (verbose >= VERBOSE_WARNINGS)
						{
							std::cout
							<< "#WWW#  [NavGraph::findPath] no decomposition cluster for start position. Using closest point: "
							<< LocStart << std::endl;
						}
					} else {
						// If obstacle pruning is on
						if ( doNavGraphPruning && (dcluster->getCertainty(T_OBSTACLE) > pruningThreshold) ) 
						{
							NavLocation LocStart = graph->getClosestLocation(userStart);
							graph->getVertex(LocStart, vstart, false);
							if (verbose >= VERBOSE_WARNINGS)
							{
								std::cout
								<< "#WWW#  [NavGraph::findPath] can not add start position in pruned zone. Using closest point: "
								<< LocStart << std::endl;
							}
						} else {
							addNode(userStart, vstart);
							if (verbose >= VERBOSE_INFO)
							{
								std::cout
								<< "#III#  [NavGraph::findPath] adding new start position at: " << userStart << std::endl;
							}
						}
					}
				} else {
					NavLocation LocStart = graph->getClosestLocation(userStart);
					if (verbose >= VERBOSE_INFO)
					{
						std::cout 
						<< "#III# [NavGraph::findPath] Using existing start position: " 
						<< LocStart << std::endl;
					}
					graph->getVertex(LocStart, vstart, false);
				}
			} //}}}

			//------------------------------------------------------------- GOAL
			if (   ((userEnd.pos_x + 1)/2) > geodata->getxsize() 
				|| ((userEnd.pos_x + 1)/2) < geodata->getxroot() 
				|| ((userEnd.pos_y + 1)/2) > geodata->getysize() 
				|| ((userEnd.pos_y + 1)/2) < geodata->getyroot() )
			{
				// Start point is not contained in the world
				NavLocation LocEnd = graph->getClosestLocation(userEnd);
				if (verbose >= VERBOSE_INFO)
				{
					std::cout << "#WWW# [NavGraph::findPathOnUtility] Position out of graph. Using Closest : "
							  << LocEnd << std::endl;
				}
				graph->getVertex(LocEnd, vend, false);
			} else { //{{{
				if (! graph->vertexExists(userEnd)) {
					if (verbose >= VERBOSE_INFO)
					{
						std::cout
						<< "#III#  [NavGraph::findPath] goal position out of graph, creating position : "
					    << userEnd << std::endl;
					}

					// Checking if pruning is on and adapting addition
					// But first lets get the decomposition cluster.
					DecompCluster* dcluster = NULL;
					posManager->getClusterUnordered(userEnd, &dcluster);
					if (dcluster == NULL)
					{
						NavLocation LocEnd = graph->getClosestLocation(userEnd);
						graph->getVertex(LocEnd, vend, false);
						if (verbose >= VERBOSE_WARNINGS)
						{
							std::cout
							<< "#WWW#  [NavGraph::findPath] no decomposition cluster for goal position. Using closest point: "
							<< LocEnd << std::endl;
						}
					} else {
						// If obstacle pruning is on
						if ( doNavGraphPruning && (dcluster->getCertainty(T_OBSTACLE) > pruningThreshold) ) 
						{
							NavLocation LocEnd = graph->getClosestLocation(userEnd);
							graph->getVertex(LocEnd, vend, false);
							if (verbose >= VERBOSE_WARNINGS)
							{
								std::cout
								<< "#WWW#  [NavGraph::findPath] can not add goal position in pruned zone. Using closest point: "
								<< LocEnd << std::endl;
							}
						} else {
							addNode(userEnd, vend);
							if (verbose >= VERBOSE_INFO)
							{
								std::cout
								<< "#III#  [NavGraph::findPath] adding new goal position at: " << userEnd << std::endl;
							}
						}
					}
				} else {
					NavLocation LocEnd = graph->getClosestLocation(userEnd);
					if (verbose >= VERBOSE_INFO)
					{
						std::cout 
						<< "#III# [NavGraph::findPath] Using existing goal position: " 
						<< LocEnd << std::endl;
					}
					graph->getVertex(LocEnd, vend, false);
				}
			} //}}}
	  
			//------------------------------------------------------------- A*
			// Getting weight map with boost get function
			NavGraphContainer::WeightMap weightmap( get(&NavGEdge::gru, *(graph->getGCtn())) );

			// Buffer Map to get the calculated cost F
			std::vector<double> resultingFullCost( boost::num_vertices(*(graph->getGCtn())), -1.0 );

			// Buffer predecessor map
			std::vector<NavGraphContainer::Vertex> predecessorMap( boost::num_vertices(*(graph->getGCtn())) );

			/*
			 * Buffer node colors to check which is open closed or neither
			 * default_color_type can be { white_color, gray_color, green_color, red_color, black_color };
			 * standing for {default_start, open, ?, ? , closed };
			 */
			std::vector<boost::default_color_type> colorMap( boost::num_vertices(*(graph->getGCtn())) );

			// List of open nodes: constructed on the fly when checking for nodes' color
			std::list<NavGraphContainer::Vertex> backPointedList;

			
			boost::astar_heuristic<NavGraphContainer::GraphContainer, COST_TYPE> navHeuristic;

			// Custom Heuristic
			if (authDist <= 0.0)
			{
				/*
				nav_heuristic<NavGraphContainer::GraphContainer, COST_TYPE, NavGraphContainer::Vertex> 
					navHeuristic (*(graph->getGCtn()),vend);
					*/
				navHeuristic = nav_heuristic<NavGraphContainer::GraphContainer, COST_TYPE, NavGraphContainer::Vertex>(*(graph->getGCtn()),vend);

			} else {
				navHeuristic = nav_heuristic_rdv< NavGraphContainer::GraphContainer, 
							   COST_TYPE, 
							   NavGraphContainer::Vertex> 
							   (*(graph->getGCtn()),vend, (COST_TYPE)authDist, predecessorMap);
			}

			/*
			 * A Star search
			 *
			 */
			GraphPath<NavGraphContainer::Vertex> graphpath =  planner->AStarSearch(
										navHeuristic,
										*(graph->getGCtn()),
										weightmap,
										vstart,
										vend,
										resultingFullCost,
										predecessorMap,
										colorMap,
										boost::ref(this->displayHook)
										);
	  
	  		// Save the cooperative path cost.
	  		coopPathCost = resultingFullCost[vend];

			// Copying nodes to fill the final navigation path without the start point
			// which is the current position of the vehicule.
			typedef GraphPath<NavGraphContainer::Vertex>::const_iterator path_iterator;  
			userNavpath.clear();
  
  			n = 0;
			NavGraphContainer::Vertex vcurr = vstart;
			NavGraphContainer::Vertex vprev = vstart;
			NavVertex navV;
			DecompCluster* dcluster;
			for(path_iterator lit=graphpath.begin(); lit!=graphpath.end(); ++lit) {	 
				vcurr = (*lit);
				// To avoid inserting the start point.
				if (n != 0)
				{
					// Get a bundled copy of the path node
					graph->getVertexBundle(vcurr,curNode);
					navV.first = curNode.pos;

					// Get edge descriptor from vprev to vcurr
					elink = graph->getEdge(vprev, vcurr);
					posManager->getCluster(((*graphee)[elink]).cluster, &dcluster);
					navV.second = *(dcluster->getAttributes());

					userNavpath.push_back(navV);
				} else {
					n = 1;
				}
				vprev = vcurr;
			}
 
			return true;
		} //}}}

		//---------------------------------------------------------//

		double NavGraph::navCostAtNode(const NavLocation& loc)
		{//{{{
			return navCostAtNode(graph->getVertex(loc));
		}//}}}

		double NavGraph::navCostAtNode(const NavGraphContainer::Vertex& vn)
		{/*{{{*/

			// For traversable land we ponder each attribute
			// with a cost associated to the speed of the vehicule
			// Flat land: 1.2m/s  --> cost 1
			// Slope land: 0.6m/s  --> cost 3
			// Rough land: 0.2m/s  --> cost 6
			// Higher Roughness --> costs multiplied by 2
			// Obstacles are considered infinite
			// We don't take into account zero values and normally we should never have
			// one at this point.

			// Graph Containter
			//NavGraph::NavGraphContainer::GraphContainer* graphee = graph->getGCtn();

			// Evaluating total weighting for esperance (denominator)
			double total_cost = 0.0; // final denominator
			//double current_cost  = 0.0;
				/*
			for(int k=T_FREE; k <= T_LIMIT; k++)
			{
				for(int m=k+1; m <= T_OBSTACLE; m++)
				{

					if ( m == T_OBSTACLE)
					{
						if (   ((*graphee)[vn]).getCertainty((attributeKey)k) >= ATT_MIN_VAL
							&& ((*graphee)[vn]).getCertainty((attributeKey)m) <= ATT_MAX_VAL)
						{
							// Exponential-like cost
							current_cost = ((*graphee)[vn]).getCertainty((attributeKey)k) 
							* ( 1.0 - ((*graphee)[vn]).getCertainty((attributeKey)m) );
							//current_cost = 1.0 / current_cost;
							current_cost = (terrainFactor[k]+terrainFactor[m]) / current_cost;
						} else {
							// Max cost = 1/ Min^2
							// with Min^2 << Min
							current_cost = (terrainFactor[k]+terrainFactor[m]) / (ATT_MIN_VAL*ATT_MIN_VAL);
						}
					} else {
						if (   ((*graphee)[vn]).getCertainty((attributeKey)k) >= ATT_MIN_VAL
							&& ((*graphee)[vn]).getCertainty((attributeKey)m) <= ATT_MAX_VAL)
						{

						}
					}
					total_cost = total_cost + current_cost;
				}
			}
				**************/
				/**************
			total_cost = (1.0 - ((*graphee)[vn]).getCertainty(T_FREE))
			+ (  (((*graphee)[vn]).getCertainty(T_OBSTACLE)) * (((*graphee)[vn]).getCertainty(T_OBSTACLE)) );
			if (((*graphee)[vn]).getCertainty(T_OBSTACLE) < ATT_MAX_VAL)
			{
				total_cost = (1.0/(1.0-(((*graphee)[vn]).getCertainty(T_OBSTACLE)))+1.0) * total_cost;
			} else { total_cost =  2.0/ATT_MIN_VAL; }
				*/


			return total_cost;
		}/*}}}*/

		bool NavGraph::resetGoalReachingPotential()
		{/*{{{*/
			NavGraph::NavGraphContainer::GraphContainer* graphee = graph->getGCtn();
			if (graphee != NULL)
			{
				BGL_FORALL_VERTICES(v, *(graphee), NavGraph::NavGraphContainer::GraphContainer)
				{
					// Resetting potential for this point
					((*(graphee))[v]).grp = 0.0;
					((*(graphee))[v]).hasPotential = false;
				}
				return true;
			}
			return false;
		}/*}}}*/

		bool NavGraph::updateNavigationUtility()
		{/*{{{*/
			NavGraph::NavGraphContainer::GraphContainer* graphee = graph->getGCtn();
			int edgeCX = 0;
			int edgeCY = 0;
			double edgeEnt = 0.0;
			double maxgru  = 0.0;
			double displaygru  = 0.0;
			DecompCluster* dcluster;

			BGL_FORALL_EDGES(edgee, *graphee, NavGraph::NavGraphContainer::GraphContainer)
			{
				(*graphee)[edgee].gru = 0.0;
				// Setting utility from the center of the edge
				// the gru is the inverse of the utility.
				// For navigation we want to maximise utility
				// and for astar_search we need to minimize (1.0 - utility)
				edgeCX = ((*graphee)[source(edgee,*graphee)].pos.pos_x
				                + (*graphee)[target(edgee,*graphee)].pos.pos_x)/4;
								// division by 4 = /2 for average and /2 for position frame
				edgeCY = ((*graphee)[source(edgee,*graphee)].pos.pos_y 
				                + (*graphee)[target(edgee,*graphee)].pos.pos_y)/4;

				// TODO : formalize this a bit more. To Be Clarifed !
				// For now we use the entropy as the weigting of the goal reaching potential.
				//edgeEnt = geodata->getFootPrintEntropy(edgeCX, edgeCY);
				posManager->getCluster(((*graphee)[edgee]).cluster, &dcluster);
				//if (dcluster == NULL) { return false; }
				edgeEnt = (dcluster->getAttributes())->infotropy[T_SYMBOL] / MAX_ENTROPY_BINARY;
				//edgeEnt = (dcluster->getAttributes())->infotropy[T_SYMBOL];
				//(*graphee)[edgee].gru =  (1.0 - edgeEnt) 
				//                        * (1.0 - (*graphee)[target(edgee,*graphee)].grp);
				//(*graphee)[edgee].gru = 1.0 / ((*graphee)[edgee].gru * (*graphee)[edgee].gru);

				//---- Additionnal GRU
				if (edgeEnt < 0.8) { // high values segmented
					edgeEnt = 1.0;
				} else {
					edgeEnt = (2.0 - edgeEnt - (*graphee)[target(edgee,*graphee)].grp)/2.0;
				}
				(*graphee)[edgee].gru = edgeEnt*edgeEnt;

				//--- Preparing normalization in case of drawing
				if ( set_geodata_gru )
				{
					// get the max(gru)
					if ((*graphee)[edgee].gru > maxgru)
					{
						maxgru = (*graphee)[edgee].gru;
					}
				}
			}

			// Normalizing for visualization
			if ( set_geodata_gru )
			{
				BGL_FORALL_EDGES(edgee, *graphee, NavGraph::NavGraphContainer::GraphContainer)
				{
					if (maxgru > 0.0)
					{
						displaygru = (*graphee)[edgee].gru / maxgru;
					}
					geodata->setMap(
						UTILITY_MAP,
						displaygru,
						(*graphee)[source(edgee,*graphee)].pos.pos_x,
						(*graphee)[source(edgee,*graphee)].pos.pos_y,
						displaygru,
						(*graphee)[target(edgee,*graphee)].pos.pos_x,
						(*graphee)[target(edgee,*graphee)].pos.pos_y);
				}
			}
			return true;
		}/*}}}*/

		void NavGraph::calculateTerrainFactor()
		{//{{{
			// Dynamic of traversable classes: number of classes
			// considered traversable.
			//int NFreeClasses     = 1 + (T_LIMIT - T_FREE);
			//int NOccupiedClasses = (T_OBSTACLE - T_LIMIT);

			//terrainFactor =(double*)calloc((T_OBSTACLE-T_FREE)+1, sizeof(double));

			// TODO
			// for now all ones
			terrainFactor[T_FREE] = 1.0;
			for(int k=T_FREE; k < T_OBSTACLE; k++)
			{
				terrainFactor[k+1] = 100.0 *  terrainFactor[k];
			}
		}//}}}

		bool NavGraph::extendFrontiers(NavGraphContainer::Vertex& vfrontier, double vcost)
		{//{{{
			frontierPoint insertor;
			std::list<frontierPoint>::iterator itfrt;
			// Sorted-Insertion
			insertor.first  = vfrontier;
			insertor.second = vcost;
			for(itfrt = frontiers.begin(); itfrt != frontiers.end(); ++itfrt)
			{
				if (vcost < (*itfrt).second)
				{
					frontiers.insert(itfrt, insertor);
					return true;
				}
			}

			// No good place but the back of the list.
			frontiers.push_back(insertor);
			return false;
		}//}}}

		//---------------------------------------------------------//
		bool NavGraph::saveAnyField(const std::string& outfile, attributeKey _at)
		{ //{{{
			// SAVING PPM gray Files 
			// P5
			// # Comments with 6 width and 2 height
			// 6 2
			// 255
			// binary rgb pixels

			std::ofstream myFile (outfile.c_str(), std::ios::out | std::ios::binary);
			
			if ( ! myFile.good() )
			{
				return false;
			}

			char* header= (char*)calloc(256, sizeof(char));
			sprintf(header, "P5\n%d %d\n255\n", geodata->getxsize(), geodata->getysize());

			myFile << header;
			NavLocation looky(0,0);
			NavLocation foundy(0,0);
			int k = 0;
			//int* content = (int*)calloc(geodata->getysize()*geodata->getxsize(), sizeof(int));
			char* contentBlur = (char*)calloc(geodata->getysize()*geodata->getxsize(), sizeof(char));
			// Init.
			for (k=0; k < geodata->getysize()*geodata->getxsize(); k++)
			{
				//content[k] = 0;
				contentBlur[k] = (char)0;
			}

			//--- Filling map content
			//--- Blurring map content
			int val = 0;
			int Nval = 0;
			int m = 0;
			int n = 0;
			int mi = 0;
			int ni = 0;
			k = 0;
			for (int j=0; j < geodata->getysize(); j++)
			{
				for (int i=0; i < geodata->getxsize(); i++)
				{
					val = 0;
					Nval = 0;
					m = i*2;
					n = j*2;
					for (mi = m; mi <= (m+2); mi++)
					{
						for (ni = n; ni <= (n+2); ni++)
						{
						}
					}
					
					if (Nval == 0)
					{
						contentBlur[k] = (char)0;
					} else {
						// Averaging interpolating somethinging
						contentBlur[k] = val/Nval;
					}
					k++;
				}
			}

			//--- Writing map content
			myFile.write(contentBlur, geodata->getysize()*geodata->getxsize());
			myFile.close();
			free(header);
			//free(content);
			free(contentBlur);
			return true;
			/*
			NavGNode curNode;
			NavGraphContainer::Vertex vfoundy;
			double vutil = 0.0;
			for (int j=0; j < geodata->getysize(); j++)
			{
				for (int i=0; i < geodata->getxsize(); i++)
				{
					// Look for cell centers and pray for good behaviour lol
					looky.setPos(i*2+1, j*2+1); 
					foundy = graph->getClosestLocation(looky);
					graph->getVertex(foundy, vfoundy, false);
					graph->getVertexBundle(vfoundy, curNode);
					vutil = curNode.getCertainty(U_UTIL); 
std::cout << " from " << curNode.pos << ": "<< vutil << "  or  " << curNode.getCertainty(U_UTIL) << "  or " <<  ((int)(vutil * 255.0)) << " into fiiiiillle " << (char)(vutil * 255.0) <<endl;
					myFile.put((char)(vutil * 255.0));
				}
			}
			*/
		} //}}}

	 
  }
}

