/**
 * \file PositionManager.cpp
 *
 * Source file that defines position manager functions
 *
 * \date 2009-09-04
 * \author Redouane Boumghar, Hung Cao
 *
 * \ingroup lgl
 *
 **/

#include <lgl/PositionManager.hpp>

using namespace std;

namespace jafar {
  namespace lgl {

	 PositionManager::PositionManager(GeoData* _geodata, Quadtree<double,GeoData>* _qtree): 
		geodata(_geodata),
		quadtree(_qtree)
	 {
		rect.setRect(_geodata->getxroot(),_geodata->getyroot(),_geodata->getxsize()*2+1,_geodata->getysize()*2+1);
		pos.resize(rect.xroot(),rect.xsize(),rect.yroot(),rect.ysize(),PosFeature());  
	 }
	 
	 
	 PositionManager::~PositionManager()
	 {/*{{{*/
	 	pos.clear();
	 }/*}}}*/

	 bool PositionManager::isAtEdge(const NavLocation& loc) const {
		return  ( 
		     ((loc.pos_x%2==0)&&(loc.pos_y%2==1)) 
		  || ((loc.pos_x%2==1)&&(loc.pos_y%2==0))
		  );
	 }

	 bool PositionManager::isAtCenter(const NavLocation& loc) const {
		return ((loc.pos_x%2==1)&&(loc.pos_y%2==1));
	 }

	 bool PositionManager::isAtBorder(const NavLocation& loc) const {
		return (loc.pos_x==rect.xroot()) 
		  ||  (loc.pos_x==rect.xend())
		  ||  (loc.pos_y==rect.yroot())
		  ||  (loc.pos_y==rect.yend())
		  ;
	 }

	 /**
	  * Get the adjacent raster cells for a given location 
	  */
	 bool PositionManager::getAdjRasterCells(const NavLocation& loc, std::vector<RasterCellIndex>& cells) const {
		
		cells.clear();
		
		// if the position is at the center of a cluster cell => only 1 adjacent raster cell
		// This should never happen (red)
		if (isAtCenter(loc)) {
			cells.push_back(RasterCellIndex((loc.pos_x-1)/2,(loc.pos_y-1)/2));
		} else if (isAtEdge(loc)) {
			// if the position is at the edge of the cluster cells => only 2 adjacent raster cells
			//  This is was usually happen in our navigation graphs (red)
			// if the loc is on a Horizontal edge
			if (loc.pos_x%2==1) {
				if (loc.pos_y > rect.yroot()) // Upper adjacent cell exists
				{ cells.push_back(RasterCellIndex((loc.pos_x-1)/2,loc.pos_y/2 - 1)); }
				if (loc.pos_y < rect.yend())  // lower adjacent cell exists
				{ cells.push_back(RasterCellIndex((loc.pos_x-1)/2,loc.pos_y/2)); }
			} else { // the loc is on a vertical edge
				if (loc.pos_x > rect.xroot()) // Left adjacent cell exists
				{ cells.push_back(RasterCellIndex(loc.pos_x/2-1,(loc.pos_y-1)/2)); }
				if (loc.pos_x < rect.xend()) // Right adjacent cell exists
				{ cells.push_back(RasterCellIndex(loc.pos_x/2,(loc.pos_y-1)/2)); }
			}	 
		} else { 
		// As we can not really define if we are on a center point, edge or corner
		// we systematiccally check for all 4 adjacent cells
		// 4 adjacent raster cells
		// topleft border
		if ((loc.pos_x>rect.xroot())&& (loc.pos_y>rect.yroot())) {
			cells.push_back(RasterCellIndex(loc.pos_x/2-1,loc.pos_y/2-1));
		}
		// topright
		if ((loc.pos_x>rect.xroot())&&(loc.pos_y<rect.yend())) {
			cells.push_back(RasterCellIndex(loc.pos_x/2-1,loc.pos_y/2));
		}
		// bottomright
		if ((loc.pos_x<rect.xend())&&(loc.pos_y<rect.yend())) {
			cells.push_back(RasterCellIndex(loc.pos_x/2,loc.pos_y/2));
		}
		// bottomleft
		if ((loc.pos_x<rect.xend())&&(loc.pos_y>rect.yroot())) {
			cells.push_back(RasterCellIndex(loc.pos_x/2,loc.pos_y/2-1));
		}

		}

		return true;
	 }

	 /**
	  * Get the adjacent clusters (represented by their origins) to a given location
	  */
	 bool PositionManager::getAdjClusters(const NavLocation& loc, vector<RasterCellIndex>& clusters) const {
		std::vector<RasterCellIndex> adjCells;
		RasterCellIndex curClusterCellRoot;
		bool exists;

		if (getAdjRasterCells(loc,adjCells)==false) {
			// This never happens for now.
			return false;
		}

		//
		for (unsigned int i=0; i < adjCells.size(); i++) {
			// get the cluster of the cell =>
			if (quadtree->getClusterForCell(adjCells[i], curClusterCellRoot)) {
				exists=false;
				for (unsigned int j=0;j<clusters.size();j++) {
					if (curClusterCellRoot==clusters[j]) {
						exists=true;
					}
				}
				
				if (!exists) {
					clusters.push_back(curClusterCellRoot);			 
				}
		  }
		}

		return true;
	 }


	 // get top left raster cell of the position
	 // if the position is at the topleft border, return the bottomright raster cell
	 bool PositionManager::getTopLeftRasterCell(const NavLocation& loc, RasterCellIndex& cell) 
	 { //{{{
		// if the location is at the left or top border of the environment
		if ((loc.pos_x==rect.xtopleft)
			 || (loc.pos_y==rect.ytopleft)) {

		  if ((loc.pos_x%2==0)&&(loc.pos_y%2==0)) {
			 if (loc.pos_x>rect.xtopleft)
				cell.set(loc.pos_x/2-1,loc.pos_y/2);
			 else if (loc.pos_y>rect.ytopleft)
				cell.set(loc.pos_x/2,loc.pos_y/2-1);
			 else
				cell.set(loc.pos_x/2,loc.pos_y/2);
			 // the loc is at the left or right of the get the 
		  } else if (loc.pos_x%2==0) {
			 cell.set(loc.pos_x/2,(loc.pos_y-1)/2);		  
		  } else 
			 cell.set((loc.pos_x-1)/2,loc.pos_y/2);		  
			 
		  return true;
		}
		
		if ((loc.pos_x%2==0)&&(loc.pos_y%2==0)) {
		  cell.set(loc.pos_x/2-1,loc.pos_y/2-1);
		  // the loc is at the left or right of the get the 
		} else if (loc.pos_x%2==0) {
		  cell.set(loc.pos_x/2-1,(loc.pos_y-1)/2);		  
		} else if (loc.pos_y%2==0) {
		  cell.set((loc.pos_x-1)/2,loc.pos_y/2-1);
		} else
		  cell.set((loc.pos_x-1)/2,(loc.pos_y-1)/2);

		return true;

	 }//}}}
	
	 bool PositionManager::getTopLeftCluster(const NavLocation& loc, RasterCellIndex& cell) 
	 { //{{{

		RasterCellIndex topleftcell;
		if (!getTopLeftRasterCell(loc,topleftcell)) 
		  return false;
		
		RasterRect rect;
		if (quadtree->getClusterForCell(topleftcell,rect))
		  cell.set(rect.xtopleft,rect.ytopleft);
		else
		  return false;

		return true;
	 } //}}}

	 bool PositionManager::inSameCluster(const NavLocation& loc1, const NavLocation& loc2, RasterCellIndex& clusterRoot) 
	 { //{{{
		// find the adjacents clusters of loc1 
		vector<RasterCellIndex> loc1_adjClusters;
		if (!getAdjClusters(loc1,loc1_adjClusters)) {
		  return false;
		}
		// find the adjacents clusters of loc2 
		vector<RasterCellIndex> loc2_adjClusters;
		if (!getAdjClusters(loc2,loc2_adjClusters)) {
		  return false;
		}

		// find if the intersection with two adj clusters sets is empty or not
		unsigned int i,j;
		for (i=0;i<loc1_adjClusters.size();i++) {
		  for (j=0;j<loc2_adjClusters.size();j++) {
			 if (loc1_adjClusters[i]==loc2_adjClusters[j]) {
				clusterRoot=loc1_adjClusters[i];
				return true;
			 }
			 
		  }
		}
		return false;
	 } //}}}

	 // get the cluster for two given positions
	 bool PositionManager::getCluster(const NavLocation& loc1, const NavLocation& loc2, RasterRect& cluster) 
	 { //{{{
		RasterCellIndex clusterRoot;

		// Dealling with cases where :
		// loc1 and loc2 are not in the same cluster and linking is impossible
		// loc1 and loc2 are the same position and it's needless to add an edge here.
		// The fact that loc1 and loc2 are on the same axis should not be checked here.
		if (    (!inSameCluster(loc1,loc2,clusterRoot)) 
			 || (loc1 == loc2)
			 //|| ((loc1.pos_x == loc2.pos_x) || (loc1.pos_y == loc2.pos_y))
		   )
		{
			return false;
		}

		// Setting the common cluster in which these locations are
		DecompCluster* dCluster = quadtree->getDecomp()->at(clusterRoot.i,clusterRoot.j);
		cluster.setRect(clusterRoot.i,clusterRoot.j,dCluster->xsize,dCluster->ysize);

		return true;
	 }//}}}

	 // get the decomp cluster for two given positions
	 bool PositionManager::getCluster(const NavLocation& loc1, const NavLocation& loc2, DecompCluster** cluster) 
	 { //{{{
		RasterCellIndex clusterRoot;

		// Dealling with cases where :
		// loc1 and loc2 are not in the same cluster and linking is impossible
		// loc1 and loc2 are the same position and it's needless to add an edge here.
		// The fact that loc1 and loc2 are on the same axis should not be checked here.
		if (    (!inSameCluster(loc1,loc2,clusterRoot)) 
			 || (loc1 == loc2)
			 //|| ((loc1.pos_x == loc2.pos_x) || (loc1.pos_y == loc2.pos_y))
		   )
		{
			return false;
		}

		// Setting the common cluster in which these locations are
		(*cluster) = quadtree->getDecomp()->at(clusterRoot.i,clusterRoot.j);

		return true;
	 }//}}}

	 // Get one of the decomp cluster for one location
	 bool PositionManager::getClusterUnordered(const NavLocation& loc, DecompCluster** cluster) 
	 { //{{{
		RasterCellIndex clusterRoot;

		// Getting one of the cluster root of this location.
		vector<RasterCellIndex> loc_adjClusters;
		getAdjClusters(loc,loc_adjClusters);
		if (loc_adjClusters.size() <= 0)
		{ return false; }
		// Getting the cluster root at index 0
		clusterRoot = loc_adjClusters[0];
		// Setting the common cluster in which these locations are
		(*cluster) = quadtree->getDecomp()->at(clusterRoot.i,clusterRoot.j);

		return true;
	 }//}}}

	 // get the decomp cluster for two given positions
	 bool PositionManager::getCluster(const RasterCellIndex& clusterRoot, DecompCluster** cluster) 
	 { //{{{
		 // TODO test clusterRoot and return false if out and return false

		// Setting the common cluster in which these locations are
		(*cluster) = quadtree->getDecomp()->at(clusterRoot.i,clusterRoot.j);
		if ((*cluster) == NULL)
		{
			return false;
		}

		return true;
	 }//}}}

	 /**
	  * Get the enter positions in the graph for a given position
	  */
	 bool PositionManager::getGEnterPos(const NavLocation& loc, std::vector<NavLocation>& enterLocs) const 
	 { //{{{

		std::vector<RasterCellIndex> clusters;
		std::vector<NavLocation> curClusterInOutPoints;
  
		// QUESTION : what if loc isn't in the graph !!!?
		// ANSWER : only one cluster is found and that is the container of loc
		getAdjClusters(loc,clusters);
  
		for (unsigned int i=0;i<clusters.size();i++) {
			if (quadtree->getDecomp()->getInOutPoints(clusters[i],curClusterInOutPoints)) {
				// append locs found in adjacent clusters
				enterLocs.insert(enterLocs.end(),curClusterInOutPoints.begin(),curClusterInOutPoints.end());
				curClusterInOutPoints.clear();
			}
		}

		// Case where we are on a seldom root node
		if (clusters.size() == 0 )
		{
			if (quadtree->getDecomp()->getInOutPoints(
				   RasterCellIndex(geodata->getxroot(), geodata->getyroot()),
				   curClusterInOutPoints)) 
			{
				enterLocs.insert(enterLocs.end(),curClusterInOutPoints.begin(),curClusterInOutPoints.end());
				curClusterInOutPoints.clear();
			}
		}

		return true;
	 } //}}}


}}
