/**
 * \file PursuitGraph.cpp
 *
 * Source file for the pursuit graph class
 *
 * \author Cyril Robin
 *
 * \ingroup lgl
 *
 **/


#include <lgl/PursuitGraph.hpp>

#include <exception>
#include <sys/time.h>

#include <queue>
#include <algorithm>
#include <boost/tuple/tuple.hpp>

#ifndef VERBOSE
#define VERBOSE 0
#endif

#ifndef DISPLAY_TIME_PERF
#define DISPLAY_TIME_PERF 0
#endif

#if DISPLAY_TIME_PERF
	timeval tG;
	timeval tGd;
	double dG,dGt, dGi;
	double dG1, dG2, dGid, dGd, dGii, dGrm;
	double dGs1, dGs2 ;
#endif

namespace jafar {
	namespace lgl {

		/****************************************************************/
		/************************ PURSUIT_GRAPH *************************/
		/****************************************************************/

		//TODO : exception management and warning message
		void PursuitGraph::computeAll( int xR, int yR, int xT, int yT, const unsigned int _horizon)
		{///{{{

			#if DISPLAY_TIME_PERF
				dGt = 0 ;
			#endif

			// Check validity
			xR = xR < 0 ? 0 : xR ;
			yR = yR < 0 ? 0 : yR ;
			xT = xT < 0 ? 0 : xT ;
			yT = yT < 0 ? 0 : yT ;

			xR = (unsigned int) xR >= xsize ? xsize-2 : (xR/2) * 2 + 1 ;
			yR = (unsigned int) yR >= ysize ? ysize-2 : (yR/2) * 2 + 1 ;
			xT = (unsigned int) xT >= xsize ? xsize-2 : (xT/2) * 2 + 1 ;
			yT = (unsigned int) yT >= ysize ? ysize-2 : (yT/2) * 2 + 1 ;

			#if VERBOSE
				std::cout << "### ---III--- Robot in ("<< xR<<","<<yR<<") is tracking the target in ("<<xT<<","<<yT<<")."<<std::endl;
			#endif

			Cell_ID cellR, cellT ;

			cellR = xR * ysize + yR ;	
			cellT = xT * ysize + yT ;	
	
			// Compute the risk level of the root 
			RISK_LEVEL r ;
			r = computeRiskLevel( cellR, cellT);

			//TODO display warning message if the coordinates have been changed.

			boost::shared_ptr < PursuitGNode> _root ( new PursuitGNode (cellR, r, 0) ) ; // /!\  by default, there is no risk for the beginning 
			_root->targetPos.insert( cellT) ;

			computeAll(_root, _horizon);

		}///}}}

		//TODO : exception management and warning message
		void PursuitGraph::computeAll( Cell_ID _cellR, Cell_ID _cellT, const unsigned int _horizon)
		{///{{{
			// tMax in ms

			#if DISPLAY_TIME_PERF
				dGt = 0 ;
			#endif
			
			Cell_ID cellR, cellT ;

			// Check validity
			// Cell_ID is (2*x+1)*ysize + (2*y+1) for cell (x,y)
			unsigned xR,yR,xT,yT ;
			xR = _cellR / ysize ; 
			yR = _cellR % ysize ;
			xT = _cellT / ysize ;
			yT = _cellT % ysize ;
			
			xR = (xR/2) * 2 + 1 ;
			yR = (yR/2) * 2 + 1 ;
			xT = (xT/2) * 2 + 1 ;
			yT = (yT/2) * 2 + 1 ;

			#if VERBOSE
				std::cout << "### ---III--- Robot in ("<< xR<<","<<yR<<") is tracking the target in ("<<xT<<","<<yT<<")."<<std::endl;
			#endif

			cellR = xR * ysize + yR ;	
			cellT = xT * ysize + yT ;	

			// Compute the risk level of the root 
			RISK_LEVEL r ;
			r = computeRiskLevel( cellR, cellT);

			//TODO display warning message if the coordinates have been changed.

			boost::shared_ptr < PursuitGNode> _root ( new PursuitGNode (cellR, r, 0) ) ; // /!\  by default, there is no risk for the beginning 
			_root->targetPos.insert( cellT) ;

			computeAll(_root, _horizon);

		}///}}}

		//TODO : exception management
		void PursuitGraph::computeAll ( const boost::shared_ptr <PursuitGNode> &_root, const unsigned int _horizon ) /// TRHOW EXCEPTION
		{///{{{

			unsigned int i ;

			// Set primary parameters
			horizon = _horizon ;

			root = _root ;
			time = root->time ;

			risk.clear();
			front.clear();

			setOfNodes.insert(root);
			nodeMap.insert( std::pair<Cell_ID, boost::shared_ptr<PursuitGNode> > (root->robotPos, root) );
			front.push_back(root);
			
			timeval t;

			#if VERBOSE
				double d, dT;
			#endif

			// Compute the graph (tree) until the horizon is reached	
			try {	
				
				#if VERBOSE
					gettimeofday(&t,NULL);
					dT = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0;
				#endif
				
				for ( i = 0 ; i < horizon ; i++)
				{				
					#if DISPLAY_TIME_PERF
						dG = 0 ;
						dG1 = 0 ; dG2 = 0 ;
						dGd = 0 ;
					#endif

					gettimeofday(&t,NULL);
					d_est = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0;

					#if VERBOSE
						d = d_est;
					#endif

					//computeNewStage();
					computeNewStage();

					gettimeofday(&t,NULL);
					d_est = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d_est ;

					#if VERBOSE
						d = d_est;
					#endif

					std::set<Cell_ID> frontPos ;	
					for (unsigned int ii=0 ; ii<front.size() ; ++ii)
					{
						boost::shared_ptr <PursuitGNode> pt ( front.at(ii) );
						if (pt)
						{
							frontPos.insert(pt->robotPos);
						}
					}

					#if DISPLAY_TIME_PERF
						std::cout 	<< "### ---III--- *** Stage #" << i << " : "
									<< "front = " 	<< front.size() 	<< " (" << frontPos.size() <<") - "
									<< "risk = " 	<< risk.size()		<< " - "
									<< "danger = " 	<< danger.size() 	<< " - "
									<< "computed in " << d << " ms. ( with " << dG << " ms for visibility, ie " << (dG/d)*100 << "\t%.)" 
									<< std::endl;

						std::cout 	<< "### ---III--- *** (details of subparts of MakeNewNodes : "
									<< "[Visibility= " 	<< dG 	<< " ms - "
									<< "M= " 			<< dG1 	<< " ms - "
									<< "newNode= " 		<< dG2 	<< " ms ])"
									<< std::endl;

						dGt += dG ;
					#else 
					#if VERBOSE
						std::cout 	<< "### ---III--- Stage #" << i <<" : "
									<< "front = " 	<< front.size() 	<< " (" << frontPos.size()<< ") - "
									<< "risk = " 	<< risk.size()  	<< " - "
									<< "danger = " 	<< danger.size() 	<< " - "
									<< "computed in " << d << " ms." 
									<< std::endl;
					#endif
					#endif				
				}	

				#if VERBOSE
					gettimeofday(&t,NULL);
					dT = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - dT;
					std::cout 	<< std::endl 
								<< "### ---III--- Total for all the " << i-1 << " stages : " 
								<< dT << "ms for visibility." 
								<< std::endl << std::endl;
				#endif
			}
			catch (const std::exception& e)
			{
				horizon = i ;

				//TODO
				throw;

			}
	
			#if DISPLAY_TIME_PERF
				std::cout << "### ---III--- Total for all the stages : " << dGt << "\tms for visibility." << std::endl;
			#endif

		}///}}}

		//TODO : exception management and warning message
		void PursuitGraph::computeAllBefore( int xR, int yR, int xT, int yT, const double tMax)
		{///{{{
			// tMax in ms

			#if DISPLAY_TIME_PERF
				dGt = 0 ;
			#endif

			// Check validity
			xR = xR < 0 ? 0 : xR ;
			yR = yR < 0 ? 0 : yR ;
			xT = xT < 0 ? 0 : xT ;
			yT = yT < 0 ? 0 : yT ;

			xR = (unsigned int) xR >= xsize ? xsize-2 : (xR/2) * 2 + 1 ;
			yR = (unsigned int) yR >= ysize ? ysize-2 : (yR/2) * 2 + 1 ;
			xT = (unsigned int) xT >= xsize ? xsize-2 : (xT/2) * 2 + 1 ;
			yT = (unsigned int) yT >= ysize ? ysize-2 : (yT/2) * 2 + 1 ;

			#if VERBOSE
				std::cout << "### ---III--- Robot in ("<< xR<<","<<yR<<") is tracking the target in ("<<xT<<","<<yT<<")."<<std::endl;
			#endif

			Cell_ID cellR, cellT ;

			cellR = xR * ysize + yR ;	
			cellT = xT * ysize + yT ;	

			// Compute the risk level of the root 
			RISK_LEVEL r ;
			r = computeRiskLevel( cellR, cellT);

			//TODO display warning message if the coordinates have been changed.

			boost::shared_ptr < PursuitGNode> _root ( new PursuitGNode (cellR, r, 0) ) ; // /!\  by default, there is no risk for the beginning 
			_root->targetPos.insert( cellT) ;

			computeAllBefore(_root, tMax);

		}///}}}

		//TODO : exception management and warning message
		void PursuitGraph::computeAllBefore( Cell_ID _cellR, Cell_ID _cellT, const double tMax)
		{///{{{

			#if DISPLAY_TIME_PERF
				dGt = 0 ;
			#endif
			
			Cell_ID cellR, cellT ;

			// Check validity
			// Cell_ID is (2*x+1)*ysize + (2*y+1) for cell (x,y)
			unsigned xR,yR,xT,yT ;
			xR = _cellR / ysize ; 
			yR = _cellR % ysize ;
			xT = _cellT / ysize ;
			yT = _cellT % ysize ;
			
			xR = (xR/2) * 2 + 1 ;
			yR = (yR/2) * 2 + 1 ;
			xT = (xT/2) * 2 + 1 ;
			yT = (yT/2) * 2 + 1 ;

			#if VERBOSE
				std::cout << "### ---III--- Robot in ("<< xR<<","<<yR<<") is tracking the target in ("<<xT<<","<<yT<<")."<<std::endl;
			#endif

			cellR = xR * ysize + yR ;	
			cellT = xT * ysize + yT ;	

			// Compute the risk level of the root 
			RISK_LEVEL r ;
			r = computeRiskLevel( cellR, cellT);

			//TODO display warning message if the coordinates have been changed.

			boost::shared_ptr < PursuitGNode> _root ( new PursuitGNode (cellR, r, 0) ) ; // /!\  by default, there is no risk for the beginning 
			_root->targetPos.insert( cellT) ;

			computeAllBefore(_root, tMax);

		}///}}}

		//TODO : exception management
		// tMax in ms 
		void PursuitGraph::computeAllBefore( const boost::shared_ptr <PursuitGNode> &_root, const double tMax) /// TRHOW EXCEPTION
		{///{{{
			// tMax in ms

			unsigned int i ;

			// Set primary parameters
			root = _root ;
			time = root->time ;

			risk.clear();
			front.clear();

			setOfNodes.insert(root);
			nodeMap.insert( std::pair<Cell_ID, boost::shared_ptr<PursuitGNode> > (root->robotPos, root) );
			front.push_back(root);

			timeval t;
			double d, tTot;

			// Compute the graph (tree) until the horizon is reached	
			try {	
				
				tTot = 0, d = 0 ;
				
				for ( i = 0 ; (tTot+d) < tMax ; i++)
				{				
					#if DISPLAY_TIME_PERF
						dG = 0 ;
						dG1 = 0 ; dG2 = 0 ;
						dGd = 0 ;
					#endif

					gettimeofday(&t,NULL);
					d_est = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0;

					d = d_est ;

					//computeNewStage();
					computeNewStage();

					gettimeofday(&t,NULL);
					d_est = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d_est;

					d = d_est ;
					tTot += d ;

					std::set<Cell_ID> frontPos ;	
					for (unsigned int ii=0 ; ii<front.size() ; ++ii)
					{
						boost::shared_ptr <PursuitGNode> pt ( front.at(ii) );
						if (pt)
						{
							frontPos.insert(pt->robotPos);
						}
					}

					#if DISPLAY_TIME_PERF
						std::cout 	<< "### ---III--- *** Stage #" << i << " : "
									<< "front = " 	<< front.size() 	<< " (" << frontPos.size() <<") - "
									<< "risk = " 	<< risk.size()		<< " - "
									<< "danger = " 	<< danger.size() 	<< " - "
									<< "computed in " << d << " ms. ( with " << dG << " ms for visibility, ie " << (dG/d)*100 << "\t%.)" 
									<< std::endl;

						std::cout 	<< "### ---III--- *** (details of subparts of MakeNewNodes : "
									<< "[Visibility= " 	<< dG 	<< " ms - "
									<< "M= " 			<< dG1 	<< " ms - "
									<< "newNode= " 		<< dG2 	<< " ms ])"
									<< std::endl;

						dGt += dG ;
					#else 
					#if VERBOSE
						std::cout 	<< "### ---III--- Stage #" << i <<" : "
									<< "front = " 	<< front.size() 	<< " (" << frontPos.size()<< ") - "
									<< "risk = " 	<< risk.size()  	<< " - "
									<< "danger = " 	<< danger.size() 	<< " - "
									<< "computed in " << d << " ms." 
									<< std::endl;
					#endif
					#endif				
				}	

				#if VERBOSE
					gettimeofday(&t,NULL);
					dT = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - dT;
					std::cout 	<< std::endl 
								<< "### ---III--- Total for all the " << i-1 << " stages : " 
								<< dT << "ms for visibility." 
								<< std::endl << std::endl;
				#endif
			}
			catch (const std::exception& e)
			{
				horizon = i ;

				//TODO
				throw;

			}

			horizon = i-1 ;

	
			#if DISPLAY_TIME_PERF
				std::cout << "### ---III--- Total for all the stages : " << dGt << "\tms for visibility." << std::endl;
			#endif

		}///}}}

		//TODO : exception management
		void PursuitGraph::computeNewStage() // THROW EXCEPTION
		{///{{{

			// Updating the time in the pursuit graph
			++time;
			
			// Initialization
			std::vector< boost::shared_ptr < PursuitGNode > > newNodes ;
			std::vector< boost::shared_ptr < PursuitGNode > >::iterator it ;
			std::vector< boost::weak_ptr < PursuitGNode > >::iterator cit ;
			std::pair< 	std::multimap< Cell_ID, boost::weak_ptr < PursuitGNode > >::iterator,
						std::multimap< Cell_ID, boost::weak_ptr < PursuitGNode > >::iterator >  itMm ;

			std::set<Cell_ID> robotAccess ;
			std::set<Cell_ID> targetAccess ;

			std::map< tr1Node, std::vector< boost::shared_ptr< PursuitGNode > > > map1 ;
			std::map< tr2Node, std::vector< Cell_ID > > map2 ;

			std::pair< std::map< tr1Node, std::vector< boost::shared_ptr< PursuitGNode > > >::iterator, bool > itm1 ;
			std::pair< std::map< tr2Node, std::vector< Cell_ID > >::iterator, bool > itm2 ;

			#if DISPLAY_TIME_PERF
				timeval t;
				double d ;
				double dd, dd1=0, dd2=0, dd3 = 0;

				dGrm = 0 ;

				gettimeofday(&t,NULL);
				d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
			#endif

			// For each node of the current front (=for each leaf)
			for ( cit = front.begin() ; cit!=front.end() ; ++cit )
			{///{{{

				boost::shared_ptr < PursuitGNode > currNode = cit->lock() ;

				if (currNode )
				{
					#if DISPLAY_TIME_PERF
						gettimeofday(&t,NULL);
						dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
					#endif

					navigator->findRobotAccessibility  ( currNode->robotPos,  robotAccess  );
					navigator->findTargetAccessibility ( currNode->targetPos, targetAccess );

					#if DISPLAY_TIME_PERF
						gettimeofday(&t,NULL);
						dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - dd;
						dd1 +=dd ;

						gettimeofday(&t,NULL);
						dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
					#endif

					// Make pairing between robot position and target potential positions
					makeNewNodes( newNodes, currNode->riskTerm, currNode->robotPos, robotAccess, targetAccess ) ;

					#if DISPLAY_TIME_PERF
						gettimeofday(&t,NULL);
						dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - dd;
						dd2 +=dd ;

						gettimeofday(&t,NULL);
						dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
					#endif


					// Compute "map1" : first step to pair (robotPos,targetPos) couples with parent nodes
					for (it = newNodes.begin() ; it != newNodes.end() ; ++it )
					{
						std::set<Cell_ID>::iterator it1;
						for ( it1 = (*it)->targetPos.begin() ; it1!=(*it)->targetPos.end() ; ++it1 )
						{
							struct tr1Node n1 ;
								n1.b  = (*it)->risk;
								n1.r = (*it)->robotPos; 
								n1.t  = (*it1) ;
								n1.rt = int(n1.b != OK )*(currNode->riskTerm + 1) ; // increment the riskTerm if necessary

							std::vector< boost::shared_ptr< PursuitGNode > > v1  ;

							itm1 = map1.insert ( std::pair	<	struct  tr1Node, 
																std::vector< boost::shared_ptr< PursuitGNode> > 
															> 	(n1, v1) )  ; // add the couple (robotPos, targetPos) if it doesn't exits, else find it.

							itm1.first->second.push_back(currNode); // add parent
						}
					}

					#if DISPLAY_TIME_PERF
						gettimeofday(&t,NULL);
						dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - dd;
						dd3 +=dd;
					#endif

				}
				else
				{	
					//TODO exception management
					throw ("[computeNewStage] : Problem encountered while accessing data"); 
				}

			}///}}}

			#if DISPLAY_TIME_PERF
				gettimeofday(&t,NULL);
				d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d;
				std::cout 	<< "**** first part computed in " << d << " ms. ("
									<< "subpart1 = " << dd1 << " - "
									<< "subpart2 = " << dd2 << " - "
									<< "subpart3 = " << dd3 << ")." 
									<< std::endl;
			
				gettimeofday(&t,NULL);
				d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
			#endif

			// Transforming "map1" into "map2"
			// second step to pair (robotPos,targetPos) couples with parent nodes (here: gather couple (robotPos, targetPos) with the same parents)
			std::map< tr1Node, std::vector< boost::shared_ptr< PursuitGNode > > >::const_iterator citm1 ;
			for ( citm1 = map1.begin() ; citm1 != map1.end() ; ++citm1 )
			{
				struct tr2Node n2 ;
					n2.b  = citm1->first.b  ;
					n2.r  = citm1->first.r  ;
					n2.rt = citm1->first.rt ;
					n2.p  = citm1->second   ;

				std::vector< Cell_ID > v2  ;

				itm2 = map2.insert ( std::pair	<	struct  tr2Node, 
													std::vector< Cell_ID > 
												> 	(n2, v2) )  ;

				itm2.first->second.push_back( citm1->first.t );
			}

			// Make front and risk change
			std::map< tr2Node, std::vector< Cell_ID > >::const_iterator citm2 ;
			front.clear();
			for ( citm2 = map2.begin() ; citm2 != map2.end() ; ++citm2 )
			{
				switch ( citm2->first.b ) 		// consider the RISK_LEVEL
				{
				case OK 	: {///{{{
					// make node
					boost::shared_ptr< PursuitGNode > newNode (	new PursuitGNode( 	citm2->first.r,
																					citm2->first.b, 
																					std::set<Cell_ID> (citm2->second.begin(), citm2->second.end() ),
																					time,
																					citm2->first.rt
																				) ); 

					// Does the node already exists in the graph ?
					bool exist = false ;
					itMm = nodeMap.equal_range( newNode->robotPos );
					for ( ; itMm.first != itMm.second ; ++(itMm.first) )
					{
						boost::shared_ptr< PursuitGNode > cmpN ( itMm.first->second	);

						if ( (cmpN) && ( *(cmpN) == *newNode ) )
						{
							exist = true ;
							newNode = cmpN ;
						}
					}

					// create links	
					std::vector< boost::shared_ptr <PursuitGNode> >::iterator itf;
					std::vector< boost::shared_ptr <PursuitGNode> > parents ( citm2->first.p );

					for (itf = parents.begin(); itf != parents.end() ; ++itf )
						(*itf)->addChild( newNode ) ;

					if (! exist ) // normal futur processing
					{ 
						// set and map of Nodes
						setOfNodes.insert( newNode );
						nodeMap.insert( std::pair< Cell_ID, boost::shared_ptr<PursuitGNode> > (newNode->robotPos,newNode) );

						front.push_back( newNode );
					} 
					// (else) the node has already been processed : no need to re-process everything, just indicate the loops!
					
					break; }///}}}

				case RISK 	: {///{{{

					// Is the risk a danger ? (reach the riskTermMaxValue)
					RISK_LEVEL risk_level = RISK ;
					if ( citm2->first.rt >= riskTermMaxValue )	
						risk_level = DANGER_UNSEEN ;

					// make node
					boost::shared_ptr< PursuitGNode > newNode (	new PursuitGNode( 	citm2->first.r,
																					//citm2->first.b, 
																					risk_level, 
																					std::set<Cell_ID> (citm2->second.begin(), citm2->second.end() ),
																					time,
																					citm2->first.rt
																				) ); 

					// Does the node already exists in the graph ?
					bool exist = false ;
					itMm = nodeMap.equal_range( newNode->robotPos );
					for ( ; itMm.first != itMm.second ; ++(itMm.first) )
					{
						boost::shared_ptr< PursuitGNode > cmpN ( itMm.first->second	);

						if ( (cmpN) && ( *(cmpN) == *newNode ) )
						{
							exist = true ;
							newNode = cmpN ;
						}
					}

					// create links	
					std::vector< boost::shared_ptr <PursuitGNode> >::iterator itf;
					std::vector< boost::shared_ptr <PursuitGNode> > parents ( citm2->first.p );

					for (itf = parents.begin(); itf != parents.end() ; ++itf )
						(*itf)->addChild( newNode ) ;

					if (! exist ) // normal futur processing
					{ 
						// set and map of Nodes
						setOfNodes.insert( newNode );
						nodeMap.insert( std::pair< Cell_ID, boost::shared_ptr<PursuitGNode> > (newNode->robotPos,newNode) );

						if ( citm2->first.b == RISK ) 	// if it is a risk indeed
						{
							front.push_back( newNode );
							risk.push_back( newNode );
						}
						else							// the risk has become a danger
							danger.push_back( newNode );
							
					} 
					// (else) the node has already been processed : no need to re-process everything, just indicate the loops!

					break;}///}}}
						
				//case DANGER	: {///{{{
				case DANGER_HIDDEN		: 
				case DANGER_LOST  		: 
				case DANGER_UNSEEN		: 
				case DANGER_DISAPPEARED :
				{
					std::vector<Cell_ID>::const_iterator citV ;
					// We distinguish every single position of the target (reduce the whole graph)
					for (citV = citm2->second.begin(); citV!=citm2->second.end();) 
					{
						std::set< Cell_ID > tPos;
						tPos.insert(*citV);++citV;

						// make node
						boost::shared_ptr< PursuitGNode > newNode (	new PursuitGNode( 	citm2->first.r,
																						citm2->first.b, 
																						tPos,
																						time,
																						citm2->first.rt
																					) ); 
						
						// Does the node already exists in the graph ?
						bool exist = false ;
						itMm = nodeMap.equal_range( newNode->robotPos );
						for ( ; itMm.first != itMm.second ; ++(itMm.first) )
						{
							boost::shared_ptr< PursuitGNode > cmpN ( itMm.first->second	);

							// set and map of Nodes
							if ( (cmpN) && ( *(cmpN) == *newNode ) ) 		// don't consider time
							//if ( cmpN && ( (*cmpN).equal(*newNode) ) )	// consider time
							{
								exist = true ;
								newNode = cmpN ;
							}
						}

						// create links	
						std::vector< boost::shared_ptr <PursuitGNode> >::iterator itf;
						std::vector< boost::shared_ptr <PursuitGNode> > parents ( citm2->first.p );

						for (itf = parents.begin(); itf != parents.end() ; ++itf )
							(*itf)->addChild( newNode ) ;

						if (! exist ) // normal futur processing
						{ 
							// set and map of Nodes
							setOfNodes.insert( newNode );
							nodeMap.insert( std::pair< Cell_ID, boost::shared_ptr<PursuitGNode> > (newNode->robotPos,newNode) );

							danger.push_back( newNode );
						} 
						// (else) the node has already been processed : no need to re-process everything, just indicate the loops!

					} // end for
					break; }///}}}

				default: ///{{{
					; //throw exception TODO
				///}}}
				}
			}

			#if DISPLAY_TIME_PERF
				gettimeofday(&t,NULL);
				d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d;
				std::cout << "**** second part computed in " << d << " ms. " << std::endl;
			#endif

		}///}}}

		//TODO : exception management and handdling with un-expected value of arguments
		bool PursuitGraph::updatePG( int _xT, int _yT, int &xR, int &yR) 
		{///{{{

			#if DISPLAY_TIME_PERF
				timeval t;
				double d;

				gettimeofday(&t,NULL);
				d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
			#endif


			int xT=_xT, yT=_yT ;

			// beware of the Cell_ID value (position may be different than the center of the cell currently use in the graph...)
			
			Cell_ID cellT ;

			// Check validity
			xT = xT < 0 ? 0 : xT ;
			yT = yT < 0 ? 0 : yT ;

			xT = (unsigned int) xT >= xsize ? xsize-2 : (xT/2) * 2 + 1 ;
			yT = (unsigned int) yT >= ysize ? ysize-2 : (yT/2) * 2 + 1 ;

			cellT = xT * ysize + yT ;	

			//TODO display warning message if the coordinates have been changed.

			bool found = false ;

			//Search if the new situation had been predicted in the PursuitGraph
			std::vector< boost::weak_ptr<PursuitGNode> >::const_iterator itb, ite ;

			itb = root->getChildren().begin() ; 
			ite = root->getChildren().end()   ; 

			std::set<Cell_ID>::const_iterator itP, itD ;

			#if VERBOSE
			std::cout << "### ---III--- Seeking for (" << xT<<","<<yT<<") (= "<< cellT <<") : "  ; 
			#endif

			for ( ; itb != ite ; ++itb ) 
			{
				boost::shared_ptr <PursuitGNode> ptr  =  (*itb).lock() ;

				if (ptr)
				{
					itP = ptr->targetPos.find( cellT ) ;

					if ( itP != ptr->targetPos.end() ) 
					{
						#if VERBOSE
						std::cout << "found! :-) " << std::endl;
						#endif

						found = true ;

						xR = ptr->robotPos / ysize ;
						yR = ptr->robotPos % ysize ;

						root = ptr ;
						time = root->time ;

						// re-center/re-orient the graph 
						// (and by the way, update risk, front and setOfNodes, and clean the graph)
						reorient();

						// TODO : handle the case when the new root is a risk!

						break;
					}
				}
			}

			// If the new situation had not been predicted
			if (!found) //TODO : what must the robot do ? what to do with the graph (clean ?)
			{
				std::cout << " NOT FOUND! :-( " << std::endl;

				// TODO set an adequate newrobotPos
				xR = 0 ; yR = 0 ;
			}

			#if DISPLAY_TIME_PERF
				gettimeofday(&t,NULL);
				d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d;
				std::cout << " UpdatePG computed in " << d << " ms." <<std::endl;
			#endif

			return (found);

		}///}}}

		//TODO : exception management and handdling with un-expected value of arguments
		bool PursuitGraph::updatePG( Cell_ID newtargetPos, Cell_ID &newrobotPos) 
		{///{{{

			#if DISPLAY_TIME_PERF
				timeval t;
				double d;

				gettimeofday(&t,NULL);
				d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
			#endif
			

			// beware of the Cell_ID value (position may be different than the center of the cell currently use in the graph...)
			
			Cell_ID cellT ;

			// Check validity
			// Cell_ID is (2*x+1)*ysize + (2*y+1) for cell (x,y)
			unsigned xT,yT ;
			xT = newtargetPos / ysize ;
			yT = newtargetPos % ysize ;
			
			xT = (xT/2) * 2 + 1 ;
			yT = (yT/2) * 2 + 1 ;

			cellT = xT * ysize + yT ;	

			//TODO display warning message if the coordinates have been changed.

			bool found = false ;

			//Search if the new situation had been predicted in the PursuitGraph
			std::vector< boost::weak_ptr<PursuitGNode> >::const_iterator itb, ite ;

			itb = root->getChildren().begin() ; 
			ite = root->getChildren().end()   ; 

			std::set<Cell_ID>::const_iterator itP, itD ;

			#if VERBOSE
			std::cout << "### ---III--- Seeking for " << cellT << "(near : "<< newtargetPos <<") : "  ; 
			#endif

			for ( ; itb != ite ; ++itb ) 
			{
				boost::shared_ptr <PursuitGNode> ptr (*itb) ;

				if (ptr)
				{
					itP = ptr->targetPos.find( cellT ) ;

					if ( itP != ptr->targetPos.end() ) 
					{
						#if VERBOSE
						std::cout << "found! :-) " << std::endl;
						#endif

						found = true ;

						//if (root->risk != OK && ptr->risk == OK )	// if we succeed in managing the risk
						//TODO : manage the deprecated risk/hypothesis

						newrobotPos = ptr->robotPos ;
						root = ptr ;
						time = root->time ;

						// re-center/re-orient the graph 
						// (and by the way, update risk, front and setOfNodes, and clean the graph)
						reorient();

						// TODO : handle the case when the new root is a risk!

						break;
					}
				}
			}

			// If the new situation had not been predicted
			if (!found) //TODO : what must the robot do ? what to do with the graph (clean ?)
			{
				std::cout << " NOT FOUND! :-( " << std::endl;

				// TODO set an adequate newrobotPos
				newrobotPos = 0 ;
			}

			#if DISPLAY_TIME_PERF
				gettimeofday(&t,NULL);
				d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d;
				std::cout << " UpdatePG computed in " << d << " ms." <<std::endl;
			#endif

			return (found);

		}///}}}

		//TODO : exception management
		void PursuitGraph::reorient() 
		{///{{{

			// Clear the old values
			risk.clear() ;
			danger.clear() ;
			front.clear() ;
			nodeMap.clear() ;


			std::set< boost::shared_ptr<PursuitGNode> >	visitedNodes ;
			std::pair< std::set< boost::shared_ptr< PursuitGNode > >::iterator, bool > it ;

			// structure ( Graph Node, time, risk term)
			std::queue< boost::tuple< boost::weak_ptr<PursuitGNode>, unsigned int , unsigned int> > q ;

			std::vector< boost::weak_ptr<PursuitGNode> >::const_iterator itV ;

			// root		
			q.push( boost::tuple< boost::weak_ptr<PursuitGNode>, unsigned int, unsigned int > ( boost::weak_ptr<PursuitGNode> (root),0,0) ) ;

			// for the graph, from the root, till the temporal horizon
			while (!q.empty())
			{
				//if (q.front().first.expired() )
				if (q.front().get<0>().expired() )
				{
					q.pop();
					continue;
				}

				boost::shared_ptr< PursuitGNode > ptr = (q.front().get<0>()).lock() ;

				if (ptr)
				{
					it = visitedNodes.insert(ptr);	

					if (it.second) // the element had not been visited yet
					{

						// TODO : improve the implementation (only one switch is needed)

						ptr->time = q.front().get<1>() ; 											// updating the time of the node
						//if (ptr->risk)
						if (ptr->risk==RISK)
							ptr->riskTerm = q.front().get<2>() + 1 ;

						if ( ptr->time > (root->time + horizon) ) 									// the node is beyond the temporal horizon
							continue ;
							
						if ( ptr->getChildren().empty() || ptr->time == (root->time + horizon) ) 	// the node is a terminal node 
						{
							switch ( ptr->risk)
							{
								case OK : 		front.push_back(ptr); break;
								case RISK :		front.push_back(ptr); break;
								case DANGER_HIDDEN		:
								case DANGER_LOST		: 
								case DANGER_UNSEEN 		:
								case DANGER_DISAPPEARED	:   break;
								default: ;// throw exception TODO
							}
							ptr->noMoreChildren();
						}
						else 
							//for (itV = ptr->getChildren().begin() ; itV != ptr->getChildren().end() ; ++itV )
							//	q.push( boost::tuple< boost::weak_ptr<PursuitGNode>, unsigned int, unsigned int > ( *itV , ptr->time + 1, ptr->riskTerm ) ) ;
							switch ( ptr->risk)
							{
								case OK : 		
								case RISK :	
									for (itV = ptr->getChildren().begin() ; itV != ptr->getChildren().end() ; ++itV )
										q.push( boost::tuple< boost::weak_ptr<PursuitGNode>, unsigned int, unsigned int > ( *itV , ptr->time + 1, ptr->riskTerm ) ) ;
									break;
								case DANGER_HIDDEN		:
								case DANGER_LOST		: 
								case DANGER_UNSEEN 		:
								case DANGER_DISAPPEARED	:   
									ptr->noMoreChildren();
									break;
								default: ;// throw exception TODO
							}

						switch ( ptr->risk)
						{
							case OK :
								break;
							case RISK :
								risk.push_back(ptr);
								break;
							case DANGER_HIDDEN		:
							case DANGER_LOST		: 
							case DANGER_UNSEEN 		:
							case DANGER_DISAPPEARED	:   
								danger.push_back(ptr);
								break;
							default: ;//throw exception TODO
						}
						
						nodeMap.insert( std::pair< Cell_ID, boost::weak_ptr<PursuitGNode> > (ptr->robotPos, ptr ) );

					}
					//else : the element had been visited before (not re-computed)
				}

				q.pop();	
			}

			setOfNodes = visitedNodes ;

		}///}}}

		//TODO : exception management
		void PursuitGraph::completeToHorizon()
		{///{{{

			if ( front.empty() )
				return;

			boost::shared_ptr<PursuitGNode> currNode ;

			unsigned int stage ;

			timeval t ; double d ;
			gettimeofday(&t,NULL);
			d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;

			// temporal sorting
			sort(front.begin(),front.end(), before) ;
			
			// Compute the graph (tree) until the horizon is reached	
			///{{{
			try {	
				// develop the closest in term of time 
				currNode = (front.at(0)).lock();
				if (currNode)
					stage = currNode->time;

				// Until each node in the front line belong to the temporal horizon
				while ( currNode && stage < (root->time + horizon) )
				{				
					completeStage(stage) ;

					if ( front.empty() )
						break;

					// time constraint management
					gettimeofday(&t,NULL);
					d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d;
					d_est = d;

					gettimeofday(&t,NULL);
					d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;

					// temporal sorting
					sort(front.begin(),front.end(), before) ;
			
					// develop the closest in term of time 
					currNode = (front.at(0)).lock();
					stage = currNode->time;
				}
			}
			catch (const std::exception& e)
			{
				horizon = stage - root->time ;
				//TODO
				throw;
			}
			///}}}

			d_est *= 0.9 ; // decreasing factor, in order to evoid permanent deadlock

			if (currNode)
				horizon = currNode->time - root->time ;
			else 
				horizon-- ;

			#if VERBOSE
			std::cout 	<< "### ---III--- [CompleteHorizon] "
						<< "Front = " 		<< front.size() 	<< " - "
						<< "Risk = " 		<< risk.size() 		<< " - "
						<< "Danger =  " 	<< danger.size() 	<< " - "
						<< "(NB: d_est = " 	<< d_est 			<< " ms." 
						<< std::endl;
			#endif

		}///}}}

		//TODO : exception management
		void PursuitGraph::completeToHorizonBefore( const double &tMax ) 
		{///{{{
			// tMax in ms and d is an estimation of the time to needed to compute completeStageRisk

			if ( front.empty() )
				return;

			boost::shared_ptr<PursuitGNode> currNode ; 

			unsigned int stage ;

			timeval t;
			double d = 0.0, tTot = 0.0;

			gettimeofday(&t,NULL);
			d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;

			// temporal sorting
			sort(front.begin(),front.end(), before) ;
			
			// Compute the graph (tree) until the horizon is reached	
			///{{{
			try {	
				// develop the closest in term of time 
				currNode = (front.at(0)).lock();

				for ( ; (currNode) && tTot + d_est < tMax ; )
				{				
					stage = currNode->time;

					completeStage(stage) ;

					if ( front.empty() )
						break;

					// time constraint management
					gettimeofday(&t,NULL);
					d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d;
					d_est = d;
					tTot += d_est ;

					gettimeofday(&t,NULL);
					d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;

					// temporal sorting
					sort(front.begin(),front.end(), before) ;
			
					// develop the closest in term of time 
					currNode = (front.at(0)).lock();

				}
			}
			catch (const std::exception& e)
			{
				horizon = stage - root->time ;
				//TODO
				throw;
			}
			///}}}

			d_est *= 0.9 ; // decreasing factor, in order to evoid permanent deadlock

			if (currNode)
				horizon = currNode->time - root->time ;
			else 
				horizon-- ;

			#if VERBOSE
			std::cout 	<< "### ---III--- [CompleteHorizon] "
						<< "Front = " 		<< front.size() 	<< " - "
						<< "Risk = " 		<< risk.size() 		<< " - "
						<< "Danger =  " 	<< danger.size() 	<< " - "
						<< "(NB: d_est = " 	<< d_est 			<< " ms." 
						<< std::endl;
			#endif

		}///}}}

		//TODO exception management
		void PursuitGraph::completeStage( unsigned long int stage) // THROW EXCEPTION
		{///{{{

			// Updating the time in the pursuit graph
			time = stage + 1 ;
			
			// Initialization
			std::vector< boost::shared_ptr < PursuitGNode > > newNodes ;
			std::vector< boost::shared_ptr < PursuitGNode > >::iterator it ;
			std::vector< boost::weak_ptr < PursuitGNode > >::iterator cit ;
			std::pair< 	std::multimap< Cell_ID, boost::weak_ptr < PursuitGNode > >::iterator,
						std::multimap< Cell_ID, boost::weak_ptr < PursuitGNode > >::iterator >  itMm ;

			std::set<Cell_ID> robotAccess ;
			std::set<Cell_ID> targetAccess ;

			std::map< tr1Node, std::vector< boost::shared_ptr< PursuitGNode > > > map1 ;
			std::map< tr2Node, std::vector< Cell_ID > > map2 ;

			std::pair< std::map< tr1Node, std::vector< boost::shared_ptr< PursuitGNode > > >::iterator, bool > itm1 ;
			std::pair< std::map< tr2Node, std::vector< Cell_ID > >::iterator, bool > itm2 ;

			#if DISPLAY_TIME_PERF
				timeval t;
				double d ;
				double dd, dd1=0, dd2=0, dd3 = 0;

				dGrm = 0;
				dGs1 = 0 ; dGs2 = 0 ;

				gettimeofday(&t,NULL);
				d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
			#endif

			// For each node of the current front (=for each leaf)
			for ( cit = front.begin() ; cit!=front.end() ; )
			{///{{{
				
				boost::shared_ptr < PursuitGNode > currNode = cit->lock() ;

				if (currNode )
				{
					//compute only the node related to the considered stage 
					if (currNode->time != stage)
					{
						++cit;
						continue ;
					}

					front.erase(cit);
					
					#if DISPLAY_TIME_PERF
						gettimeofday(&t,NULL);
						dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
					#endif

					navigator->findRobotAccessibility  ( currNode->robotPos,  robotAccess  );
					navigator->findTargetAccessibility ( currNode->targetPos, targetAccess );

					#if DISPLAY_TIME_PERF
						gettimeofday(&t,NULL);
						dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - dd;
						dd1 +=dd ;

						gettimeofday(&t,NULL);
						dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;

					#endif

					// Make pairing between robot position and target potential positions
					makeNewNodes( newNodes, currNode->riskTerm, currNode->robotPos, robotAccess, targetAccess ) ;

					#if DISPLAY_TIME_PERF
						gettimeofday(&t,NULL);
						dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - dd;
						dd2 +=dd ;

						gettimeofday(&t,NULL);
						dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
					#endif

					// Compute "map1" : first step to pair (robotPos,targetPos) couples with parent nodes
					for (it = newNodes.begin() ; it != newNodes.end() ; ++it )
					{
						std::set<Cell_ID>::iterator it1;
						for ( it1 = (*it)->targetPos.begin() ; it1!=(*it)->targetPos.end() ; ++it1 )
						{
							struct tr1Node n1 ;
								n1.b  = (*it)->risk;
								n1.r = (*it)->robotPos; 
								n1.t  = (*it1) ;
								n1.rt = int(n1.b != OK )*(currNode->riskTerm + 1) ; // increment the riskTerm if necessary

							std::vector< boost::shared_ptr< PursuitGNode > > v1  ;

							itm1 = map1.insert ( std::pair	<	struct  tr1Node, 
																std::vector< boost::shared_ptr< PursuitGNode> > 
															> 	(n1, v1) )  ; // add the couple (robotPos, targetPos) if it doesn't exits, else find it.

							itm1.first->second.push_back(currNode); // add parent
						}
					}

					#if DISPLAY_TIME_PERF
						gettimeofday(&t,NULL);
						dd = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - dd;
						dd3 +=dd;
					#endif

				}
				else
				{	
					//TODO exception management
					throw ("[computeNewStage] : Problem encountered while accessing data"); 
				}

			}///}}}

			#if DISPLAY_TIME_PERF
				gettimeofday(&t,NULL);
				d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d;
				std::cout 	<< "**** first part computed in " << d << " ms. ("
									<< "subpart1 = " << dd1 << " - "
									<< "subpart2 = " << dd2 << " - "
									<< "subpart3 = " << dd3 << ")." 
									<< std::endl;
			
				gettimeofday(&t,NULL);
				d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
			#endif

			// Transforming "map1" into "map2"
			// second step to pair (robotPos,targetPos) couples with parent nodes (here: gather couple (robotPos, targetPos) with the same parents)
			std::map< tr1Node, std::vector< boost::shared_ptr< PursuitGNode > > >::const_iterator citm1 ;
			for ( citm1 = map1.begin() ; citm1 != map1.end() ; ++citm1 )
			{
				struct tr2Node n2 ;
					n2.b  = citm1->first.b  ;
					n2.r  = citm1->first.r  ;
					n2.rt = citm1->first.rt ;
					n2.p  = citm1->second   ;

				std::vector< Cell_ID > v2  ;

				itm2 = map2.insert ( std::pair	<	struct  tr2Node, 
													std::vector< Cell_ID > 
												> 	(n2, v2) )  ;

				itm2.first->second.push_back( citm1->first.t );
			}

			// Make front and risk change
			//std::map< tr2Node, std::vector< Cell_ID > >::const_iterator citm2 ;
			std::map< tr2Node, std::vector< Cell_ID > >::iterator citm2 ;
			for ( citm2 = map2.begin() ; citm2 != map2.end() ; ++citm2 )
			{
				switch ( citm2->first.b ) 		// consider the RISK_LEVEL
				{
				case OK 	: {///{{{
					// make node
					boost::shared_ptr< PursuitGNode > newNode (	new PursuitGNode( 	citm2->first.r,
																					citm2->first.b, 
																					std::set<Cell_ID> (citm2->second.begin(), citm2->second.end() ),
																					time,
																					citm2->first.rt
																				) ); 

					// Does the node already exists in the graph ?
					bool exist = false ;
					itMm = nodeMap.equal_range( newNode->robotPos );
					for ( ; itMm.first != itMm.second ; ++(itMm.first) )
					{
						boost::shared_ptr< PursuitGNode > cmpN ( itMm.first->second	);

						if ( (cmpN) && ( *(cmpN) == *newNode ) )
						{
							exist = true ;
							newNode = cmpN ;
						}
					}

					// create links	
					std::vector< boost::shared_ptr <PursuitGNode> >::iterator itf;
					std::vector< boost::shared_ptr <PursuitGNode> > parents ( citm2->first.p );

					for (itf = parents.begin(); itf != parents.end() ; ++itf )
						(*itf)->addChild( newNode ) ;

					if (! exist ) // normal futur processing
					{ 
						// set and map of Nodes
						setOfNodes.insert( newNode );
						nodeMap.insert( std::pair< Cell_ID, boost::shared_ptr<PursuitGNode> > (newNode->robotPos,newNode) );

						front.push_back( newNode );
					} 
					// (else) the node has already been processed : no need to re-process everything, just indicate the loops!
					
					break; }///}}}

				case RISK 	: {///{{{

					// Is the risk a danger ? (reach the riskTermMaxValue)
					RISK_LEVEL risk_level = RISK ;
					if ( citm2->first.rt >= riskTermMaxValue )	
						risk_level = DANGER_UNSEEN ;

					// make node
					boost::shared_ptr< PursuitGNode > newNode (	new PursuitGNode( 	citm2->first.r,
																					//citm2->first.b, 
																					risk_level, 
																					std::set<Cell_ID> (citm2->second.begin(), citm2->second.end() ),
																					time,
																					citm2->first.rt
																				) ); 

					// Does the node already exists in the graph ?
					bool exist = false ;
					itMm = nodeMap.equal_range( newNode->robotPos );
					for ( ; itMm.first != itMm.second ; ++(itMm.first) )
					{
						boost::shared_ptr< PursuitGNode > cmpN ( itMm.first->second	);

						if ( (cmpN) && ( *(cmpN) == *newNode ) )
						{
							exist = true ;
							newNode = cmpN ;
						}
						}

					// create links	
					std::vector< boost::shared_ptr <PursuitGNode> >::iterator itf;
					std::vector< boost::shared_ptr <PursuitGNode> > parents ( citm2->first.p );

					for (itf = parents.begin(); itf != parents.end() ; ++itf )
						(*itf)->addChild( newNode ) ;

					if (! exist ) // normal futur processing
					{ 
						// set and map of Nodes
						setOfNodes.insert( newNode );
						nodeMap.insert( std::pair< Cell_ID, boost::shared_ptr<PursuitGNode> > (newNode->robotPos,newNode) );

						if ( citm2->first.b == RISK ) 	// if it is a risk indeed
						{
							front.push_back( newNode );
							risk.push_back( newNode );
						}
						else							// the risk has become a danger
							danger.push_back( newNode );
							
					} 
					// (else) the node has already been processed : no need to re-process everything, just indicate the loops!

					break;}///}}}
						
				//case DANGER	: {///{{{
				case DANGER_HIDDEN		: 
				case DANGER_LOST  		: 
				case DANGER_UNSEEN		: 
				case DANGER_DISAPPEARED :
				{
					std::vector<Cell_ID>::const_iterator citV ;
					// We distinguish every single position of the target (reduce the whole graph)
					for (citV = citm2->second.begin(); citV!=citm2->second.end();) 
					{
						std::set< Cell_ID > tPos;
						tPos.insert(*citV);++citV;

						// make node
						boost::shared_ptr< PursuitGNode > newNode (	new PursuitGNode( 	citm2->first.r,
																						citm2->first.b, 
																						tPos,
																						time,
																						citm2->first.rt
																					) ); 
						
						// Does the node already exists in the graph ?
						bool exist = false ;
						itMm = nodeMap.equal_range( newNode->robotPos );
						for ( ; itMm.first != itMm.second ; ++(itMm.first) )
						{
							boost::shared_ptr< PursuitGNode > cmpN ( itMm.first->second	);

							// set and map of Nodes
							if ( (cmpN) && ( *(cmpN) == *newNode ) ) 		// don't consider time
							//if ( cmpN && ( (*cmpN).equal(*newNode) ) )	// consider time
							{
								exist = true ;
								newNode = cmpN ;
							}
						}

						// create links	
						std::vector< boost::shared_ptr <PursuitGNode> >::iterator itf;
						std::vector< boost::shared_ptr <PursuitGNode> > parents ( citm2->first.p );

						for (itf = parents.begin(); itf != parents.end() ; ++itf )
							(*itf)->addChild( newNode ) ;

						if (! exist ) // normal futur processing
						{ 
							// set and map of Nodes
							setOfNodes.insert( newNode );
							nodeMap.insert( std::pair< Cell_ID, boost::shared_ptr<PursuitGNode> > (newNode->robotPos,newNode) );

							danger.push_back( newNode );
						} 
						// (else) the node has already been processed : no need to re-process everything, just indicate the loops!

					} // end for
					break; }///}}}

				default: ///{{{
					; //throw exception TODO
					///}}}
				}
			}

			#if DISPLAY_TIME_PERF
				gettimeofday(&t,NULL);
				d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d;
				std::cout << "**** second part computed in " << d << " ms. " << std::endl;
			#endif

		}///}}}

		//TODO try to improve the M matrix computation ?
		//TODO exception management
		void PursuitGraph::makeNewNodes ( 	std::vector< boost::shared_ptr < PursuitGNode > > &newNodes,
											const unsigned int &riskTerm,
											const Cell_ID &robotPos,
											const std::set < Cell_ID > &robotAccess,
											const std::set < Cell_ID > &targetAccess
											) // THROW EXCEPTION
		{///{{{

			int m,n,i,j;
			m = robotAccess.size();
			n = targetAccess.size();

			// "translation" table
			std::vector < Cell_ID > robotID  (m);
			std::vector < Cell_ID > targetID (n);

			std::set< Cell_ID >::const_iterator it ;
			std::set< Cell_ID >::const_iterator ite ;
			i = -1 ;
			for ( it = robotAccess.begin() ; it != robotAccess.end() ; ++it) 
				robotID[++i] = (*it) ;
				
			i = -1 ;
			for ( it = targetAccess.begin() ; it != targetAccess.end() ; ++it) 
				targetID[++i] = (*it) ;

			#if DISPLAY_TIME_PERF
				gettimeofday(&tG,NULL);
				dGi = 1000.0 * tG.tv_sec + tG.tv_usec/ 1000.0;
			#endif

			// pre-compute the vibility link for the robot position
			std::map < Cell_ID, const VisibleArea* > visibilityLink ;
			visibilityMap->computeVisibilityOnDemand ( visibilityLink,robotAccess );

			#if DISPLAY_TIME_PERF
				gettimeofday(&tG,NULL);
				dGi = 1000.0 * tG.tv_sec + tG.tv_usec/ 1000.0 - dGi;
				dG+=dGi;

				gettimeofday(&tG,NULL);
				dGi = 1000.0 * tG.tv_sec + tG.tv_usec/ 1000.0;
			 #endif

			// Matrix M of vibility and distance
			std::vector< std::vector< std::pair < bool, double > > > M (m, std::vector < std::pair < bool, double > > (n, std::pair<bool, double> (false,0.0) )) ;
			std::map < Cell_ID, const VisibleArea* >::const_iterator itt ;

			//for ( i = 0 ; i < m ; i++ )
			i=0;
			for (itt = visibilityLink.begin() ; itt != visibilityLink.end() ; ++itt )
			{

				it  = itt->second->visibleArea.begin();
				ite = itt->second->visibleArea.end()  ;

				for ( j = 0 ; j < n && it!= ite ; j++ )
				{
					while ( it!= ite && *it < targetID[j] )
						++it;

					if ( it != ite && *it == targetID[j] ) 
					{
						M[i][j] = std::pair< bool, double > ( true, navigator->distCellID ( robotID[i], targetID[j]) ) ; 
						++it;
					}
				}
				i++;
			}

			#if DISPLAY_TIME_PERF
				gettimeofday(&tG,NULL);
				dGi = 1000.0 * tG.tv_sec + tG.tv_usec/ 1000.0 - dGi;
				dG1+=dGi;

				gettimeofday(&tG,NULL);
				dGi = 1000.0 * tG.tv_sec + tG.tv_usec/ 1000.0;
			#endif

			// Creating all the potential nodes
			// (including the risk one)
			newNodes.clear();

			//non risk
			for ( i = 0 ; i < m ; i++)
				newNodes.push_back( boost::shared_ptr< PursuitGNode > (new PursuitGNode( robotID[i] , OK )) ); 
			// risk
			newNodes.push_back( boost::shared_ptr< PursuitGNode > (new PursuitGNode( robotPos, RISK )) );

			// for each potential positions of the target, give the best position of the robot
			for ( j = 0 ; j < n ; j++ )
			{
				// find the appropriate robot position (ie the closest one which have a viewline on the considerd target position )
				long int imin = m ;
				double dmin = 999999 ;
				for ( i = 0 ; i < m ; i++ )
				{
					if ( M[i][j].first && M[i][j].second < dmin )
					{
						imin = i ;
						dmin = M[i][j].second ;
					}
				}

				newNodes.at(imin)->targetPos.insert( targetID[j] );
			}

			// risk management
			#if DISPLAY_TIME_PERF
				gettimeofday(&tG,NULL);
				dGii = 1000.0 * tG.tv_sec + tG.tv_usec/ 1000.0;
			#endif

			std::vector< boost::shared_ptr< PursuitGNode > > solutions ;
			manageRisk(robotPos, newNodes.at(m), solutions) ;

			newNodes.pop_back();
			newNodes.insert(newNodes.end(),solutions.begin(), solutions.end());

			#if DISPLAY_TIME_PERF
				gettimeofday(&tG,NULL);
				dGii = 1000.0 * tG.tv_sec + tG.tv_usec/ 1000.0 - dGii;
				dGrm += dGii;
			#endif
			
			// Remove the deprecated node (those who have 0 targetPos)
			std::vector< boost::shared_ptr < PursuitGNode > >::iterator itN ;
			for ( itN = newNodes.begin() ; itN != newNodes.end() ; )
				if ( (*itN)->targetPos.empty() )
					newNodes.erase(itN);		
				else
					++itN;

			#if DISPLAY_TIME_PERF
				gettimeofday(&tG,NULL);
				dGi = 1000.0 * tG.tv_sec + tG.tv_usec/ 1000.0 - dGi;
				dG2+=dGi;
			#endif

		}///}}}

		void PursuitGraph::manageRisk (	const Cell_ID &robotPos, 
										const boost::shared_ptr< PursuitGNode > &risk, 
										std::vector< boost::shared_ptr < PursuitGNode > > &solutions
										)
		{///{{{

			#if DISPLAY_TIME_PERF
				timeval t;
				double d;
			#endif

			// Distinguish low risks and real dangers
			std::set< Cell_ID > lowRisks, realDangers ;
			std::set< Cell_ID >::const_iterator citS ;
			for ( citS = risk->targetPos.begin() ; citS != risk->targetPos.end() ; ++citS )
				if ( visibilityMap->isVisible(*citS) )
					lowRisks.insert(*citS);
				else
					realDangers.insert(*citS);

			// low risks management
			///{{{
			if (!lowRisks.empty())
			{
				// find a set of positions from which the robot may handle the risks
				std::set< Cell_ID > positions ;

				#if DISPLAY_TIME_PERF
					gettimeofday(&t,NULL);
					d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
				#endif

				// Chose the espected behaviour (manage all, manage the nearest, etc.)
				// BEWARE : IT REALLY CHANGES THE BEHAVIOUR ! (and the performances...)
				findIdealPositions(risk->targetPos, positions);
				//findAnyPositions(lowRisks, positions); 			 

				#if DISPLAY_TIME_PERF
					gettimeofday(&t,NULL);
					d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d;
					dGs1 += d ;

					gettimeofday(&t,NULL);
					d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 ;
				#endif

				// find a way to this set
				std::vector< Cell_ID > path ;
				navigator->nav2Set( robotPos, positions, path, DEFAULT_PURSUIT_TEMPORAL_TOLERANCE);

				#if DISPLAY_TIME_PERF
					gettimeofday(&t,NULL);
					d = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - d;
					dGs2 += d ;
				#endif

				// chose the next move 
				if (!path.empty())
				{
					Cell_ID strategicPos ;
					if (path.size()>1)	strategicPos = path.at(path.size()-2); // first step of the path  
					else 				strategicPos = path.at(0) ;


					solutions.push_back( boost::shared_ptr< PursuitGNode > (new PursuitGNode( 	strategicPos,	
																								RISK, 
																								lowRisks
																								)) ); 
				}		
				else
				{
					//solutions.push_back( boost::shared_ptr< PursuitGNode > (new PursuitGNode( 	robotPos,
					solutions.push_back( boost::shared_ptr< PursuitGNode > (new PursuitGNode( 	0,
																								DANGER_LOST,  
																								lowRisks
																								)) ); 
				}
			}
			//}}}

			// real dangers management
			if (!realDangers.empty())
				//solutions.push_back( boost::shared_ptr< PursuitGNode > (new PursuitGNode( 	robotPos,
				solutions.push_back( boost::shared_ptr< PursuitGNode > (new PursuitGNode( 	0,
																							DANGER_HIDDEN,  
																							realDangers
																							)) ); 

		}///}}}

		void PursuitGraph::findIdealPositions( const std::set< Cell_ID > &targetPos, std::set< Cell_ID > &positions)
		{///{{{

			positions.clear();

			std::set< Cell_ID > temp ;

			/* compute visibility */
			std::map < Cell_ID, const VisibleArea* > visibilityLink ;
			visibilityMap->computeVisibilityOnDemand ( visibilityLink,targetPos );

			/* make intersection */
			if (!visibilityLink.empty() )
			{
				std::map < Cell_ID, const VisibleArea* >::const_iterator citm ;
				citm = visibilityLink.begin();
				positions = citm->second->visibleArea;

				++citm;
				for ( ; citm != visibilityLink.end() ; ++ citm )
				{
					temp.clear();

					VisibleArea visA (*citm->second);
					std::set_intersection(	visA.visibleArea.begin(), visA.visibleArea.end(),
											positions.begin(), positions.end(),
											inserter(temp,temp.begin()) 
											);
					positions =  temp ;
				}
			}

		}///}}}

		void PursuitGraph::findAnyPositions( const std::set< Cell_ID > &targetPos, std::set< Cell_ID > &positions)
		{///{{{
			positions.clear();

			std::set< Cell_ID > temp ;

			/* compute visibility */
			std::map < Cell_ID, const VisibleArea* > visibilityLink ;
			visibilityMap->computeVisibilityOnDemand ( visibilityLink,targetPos );

			/* make intersection */
			if (!visibilityLink.empty() )
			{
				std::map < Cell_ID, const VisibleArea* >::const_iterator citm ;
				citm = visibilityLink.begin();
				positions = citm->second->visibleArea;

				++citm;
				for ( ; citm != visibilityLink.end() ; ++ citm )
				{
					temp.clear();

					VisibleArea visA (*citm->second);
					std::set_union(	visA.visibleArea.begin(), visA.visibleArea.end(),
									positions.begin(), positions.end(),
									inserter(temp,temp.begin()) 
									);

					positions =  temp ;
				}
			}

		}///}}}

		RISK_LEVEL PursuitGraph::computeRiskLevel( const Cell_ID &robotPos, const Cell_ID &targetPos)
		{///{{{

			RISK_LEVEL r ;

			if ( !visibilityMap->isVisible(targetPos) )
				r = DANGER_HIDDEN;
			
			else
			{ 
				const VisibleArea* visibleArea ;
				visibleArea = visibilityMap->getPtVisibilityOnDemand( targetPos ) ;

				std::set<Cell_ID>::iterator itS ;
				itS = visibleArea->visibleArea.find(robotPos);

				if ( itS != visibleArea->visibleArea.end() ) // if there is a visiblity link
					r = OK ;
				else
				{
					std::vector< Cell_ID > path ;
					navigator->nav2Set( robotPos, visibleArea->visibleArea, path, pursuitTemporalTolerance);

					if (!path.empty())		// there is a way to a viewshed, from which we can watch the target
						r = RISK ;
					else					// Cannot access a viewshed to watch the target
						r = DANGER_LOST ; 
				}
			}

			return (r);

		}///}}}

		unsigned int PursuitGraph::gethorizon() const
		{///{{{
			return(horizon);
		}///}}}
		
		const std::vector< boost::weak_ptr <PursuitGNode> > & PursuitGraph::getrisk() const
		{///{{{
			return ( risk );
		}///}}}

		const std::vector< boost::weak_ptr <PursuitGNode> > & PursuitGraph::getdanger() const
		{///{{{
			return ( danger );
		}///}}}

		const std::vector< boost::weak_ptr <PursuitGNode> > & PursuitGraph::getfront() const
		{///{{{
			return ( front );
		}///}}}

		const std::set< boost::shared_ptr <PursuitGNode> > & PursuitGraph::getsetOfNodes() const
		{///{{{
			return ( setOfNodes );
		}///}}}

		const std::multimap< Cell_ID, boost::weak_ptr <PursuitGNode> > & PursuitGraph::getnodeMap() const
		{///{{{
			return ( nodeMap );
		}///}}}

		const boost::shared_ptr <PursuitGNode>  & PursuitGraph::getroot() const
		{///{{{
			return ( root);
		}///}}}

		PursuitNavigator* PursuitGraph::getnavigator() 
		{///{{{
			return( navigator);
		}///}}}
		
		unsigned int PursuitGraph::getysize() const
		{///{{{
			return (ysize);
		}///}}}

		unsigned int PursuitGraph::getxsize() const
		{///{{{
			return (xsize);
		}///}}}

		unsigned long int PursuitGraph::gettime() const
		{///{{{	
			return (time);
		}///}}}

		void PursuitGraph::disp()
		{///{{{

			std::set< boost::shared_ptr<PursuitGNode> >::const_iterator itS ;
			std::vector< boost::weak_ptr<PursuitGNode> >::const_iterator itV ;

			// all the nodes
			for (itS = setOfNodes.begin() ; itS != setOfNodes.end() ; ++itS )
			{
				//std::cout << "| "<< (*itS)->robotPos << " |" ;
				std::cout << "| ("	<< (*itS)->robotPos 			<<","
									<< (*itS)->risk 				<<","
									<< (*itS)->targetPos.size()	<<","
									<< (*itS)->time 				<<"," 
									<< (*itS)->riskTerm 			<< ") |" ;

				//std::cout << "*********** Node # "<< (*itS)->robotPos << std::endl ;
				//for ( itV = (*itS)->getChildren().begin() ; itV != (*itS)->getChildren().end() ; ++itV )
				//{
				//	boost::shared_ptr< PursuitGNode > ptr = (*itV).lock() ;
				//	if (ptr)
				//		std::cout << "\t\t\t\t\t sub-node # "<< (ptr)->robotPos << std::endl ;
				//}
			}

			std::cout << "FRONT (" << front.size() << ") is : ";

			// the front
			for ( itV = front.begin() ; itV != front.end() ; ++itV )
			{
				boost::shared_ptr< PursuitGNode > ptr = (*itV).lock() ;
				if (ptr)
					//std::cout << "| "<< ptr->robotPos << " |";
					std::cout << "| ("	<< ptr->robotPos				<<","
										<< ptr->risk					<<","
										<< ptr->targetPos.size() 	<<","
										<< ptr->time 					<<"," 
										<< ptr->riskTerm 				<< ") |" ;
			}
			std::cout << std::endl;
			std::cout << std::endl;
			std::cout << std::endl;

		}///}}}

		void PursuitGraph::disp_risk()
		{///{{{

			std::set< Cell_ID >::const_iterator itS ;
			std::vector< boost::weak_ptr<PursuitGNode> >::const_iterator itV ;

			std::cout<< "RISKS (" << risk.size() << ") is : ";

			// the risks
			for ( itV = risk.begin() ; itV != risk.end() ; ++itV )
			{
				boost::shared_ptr< PursuitGNode > ptr = (*itV).lock() ;
				if (ptr)
					//std::cout << "| "<< ptr->robotPos << " |";
					std::cout << "| ("	<< ptr->robotPos				<<","
										<< ptr->risk					<<","
										<< ptr->targetPos.size() 		<<","
										<< ptr->time 					<<"," 
										<< ptr->riskTerm 				<< ") |" ;

					for ( itS = ptr->targetPos.begin() ; itS != ptr->targetPos.end() ; ++itS )
						std::cout << "|  "<< *itS <<" |";
					std::cout << std::endl;
			}
			std::cout << std::endl;

		}///}}}

		void PursuitGraph::disp_danger()
		{///{{{

			std::set< Cell_ID >::const_iterator itS ;
			std::vector< boost::weak_ptr<PursuitGNode> >::const_iterator itV ;

			std::cout << "DANGER (" << danger.size() << ") is : ";

			// the dangers
			for ( itV = danger.begin() ; itV != danger.end() ; ++itV )
			{
				boost::shared_ptr< PursuitGNode > ptr = (*itV).lock() ;
				if (ptr)
					//std::cout << "| "<< ptr->robotPos << " |";
					std::cout << "| ("	<< ptr->robotPos				<<","
										<< ptr->risk					<<","
										<< ptr->targetPos.size() 		<<","
										<< ptr->time 					<<"," 
										<< ptr->riskTerm 				<< ") |" ;

					for ( itS = ptr->targetPos.begin() ; itS != ptr->targetPos.end() ; ++itS )
						std::cout << "|  "<< *itS <<" |" ;
					std::cout << std::endl;
			}
			std::cout << std::endl;

		}///}}}

		/****************************************************************/
		/*********************** / PURSUIT_GRAPH ************************/
		/****************************************************************/

	}//lgl
}//jafar

