/**
 * \file PursuitNavigator.cpp
 *
 * Source file for the pursuit navigator class
 * 
 *
 * \author Cyril Robin
 *
 * \ingroup lgl
 *
 **/


#include <lgl/PursuitNavigator.hpp>

#include <ostream> 

#ifndef DISPLAY_TIME_PERF
#define DISPLAY_TIME_PERF 0
#endif

#ifndef SWITCH_CONNEXITY
#define SWITCH_CONNEXITY 0
// use with "findAccessibilxwity()" functions
// 0 is for 9-connexity
// 1 is for 5-connexity ( '+' cross)
#endif

#ifndef DISTANCE
//type of distance used
enum { 	EUCLIDIAN,
		MANHATTAN };
#define DISTANCE EUCLIDIAN
//#define DISTANCE MANHATTAN
#endif


namespace jafar {
	namespace lgl {
		
		/****************************************************************/
		/************************ PURSUIT_NAVIGATOR *********************/
		/****************************************************************/
		
		void PursuitNavigator::findRobotAccessibility ( const Cell_ID &currentRobotPos, std::set<Cell_ID> &robotAccess ) 
		{///{{{

			robotAccess.clear();

			int i0,j0,i,j ;

			i0 = currentRobotPos / ysize ;
			j0 = currentRobotPos % ysize ; // = currentRobotPos - i0 * ysize ;
			
			// Movement model
			#if SWITCH_CONNEXITY
				// here : 5-cells connexity
				for ( i = -2 ; i <= 2 ; i+=2 )
				for ( j = -2 ; j <= 2 ; j+=2 )
					if ( i!= j || (i == 0 && j== 0 ) ) 	// 5-connexity
					if (i0+i >=0 && j0+j >= 0 ) //inside the map
					if ( robGeoData->getValueFromBand(GEODATA_TFLAT_BAND, (i+i0)/2, (j+j0)/2 ) > 0.5 ) 
						robotAccess.insert( (i0+i) * ysize + (j0+j) );
			#else
				// here : 3 x 3 cells : 9-cells connexity
				for ( i = -2 ; i <= 2 ; i+=2 )
				for ( j = -2 ; j <= 2 ; j+=2 )
					if (i0+i >=0 && j0+j >= 0 ) //inside the map
					if ( robGeoData->getValueFromBand(GEODATA_TFLAT_BAND, (i+i0)/2, (j+j0)/2 ) > 0.5 ) 
						robotAccess.insert( (i0+i) * ysize + (j0+j) );
			#endif
			
		}///}}}

		void PursuitNavigator::findTargetAccessibility ( const std::set<Cell_ID> &currentTargetPos, std::set<Cell_ID> &targetAccess )
		{///{{{
			
			targetAccess.clear();

			int i0,j0,i,j ;

			std::set<Cell_ID>::const_iterator it ;
			
			// for each potential current position of the target
			for ( it = currentTargetPos.begin(); it != currentTargetPos.end(); ++it)
			{
				i0 = (*it) / ysize ;
				j0 = (*it) % ysize ; 

				// Movement model
				#if SWITCH_CONNEXITY
					// here : 5-cells connexity
					for ( i = -2 ; i <= 2 ; i+=2 )
					for ( j = -2 ; j <= 2 ; j+=2 )
						if ( i!= j || (i == 0 && j== 0 ) ) 	// 5-connexity
						if (i0+i >=0 && j0+j >= 0 ) //inside the map
						if ( tarGeoData->getValueFromBand(GEODATA_TFLAT_BAND, (i+i0)/2, (j+j0)/2 ) > 0.5 ) 
							targetAccess.insert( (i0+i) * ysize + (j0+j) );
				#else
					// here : 3 x 3 cells : 9-cells connexity
					for ( i = -2 ; i <= 2 ; i+=2 )
					for ( j = -2 ; j <= 2 ; j+=2 )
						if (i0+i >=0 && j0+j >= 0 ) //inside the map
						if ( tarGeoData->getValueFromBand(GEODATA_TFLAT_BAND, (i+i0)/2, (j+j0)/2 ) > 0.5 ) 
							targetAccess.insert( (i0+i) * ysize + (j0+j) );
				#endif
			}

		}///}}}

		double PursuitNavigator::distCellID ( Cell_ID a, Cell_ID b)
		{///{{{

			long int x,y ;
			double d;
			x = a/ysize - b/ysize ;
			y = a%ysize - b%ysize ;

			#if	(DISTANCE == EUCLIDIAN)
				/* Euclian distance */
				d =  sqrt(  x*x + y*y ) ;

			#elif (DISTANCE == MANHATTAN)
				/* Manhattan distance */
				d =  (x >= 0? x : -x) + (y >= 0? y : -y) ; 
			#else
				d = 0 ; // default
			#endif

			return ( d ) ;

		}///}}}
	
		double PursuitNavigator::distPos ( pNavPos A, pNavPos B)
		{///{{{
			return ( distPos (A.x, A.y, B.x, B.y) ) ;
		}///}}}
	
		double PursuitNavigator::distPos ( int xA, int yA, int xB, int yB)
		{///{{{

			double x,y ;
			double d;
			x = xA - xB ;
			y = yA - yB ;

			#if	(DISTANCE == EUCLIDIAN)
				/* Euclian distance */
				d =  sqrt(  x*x + y*y ) ; 
				
			#elif (DISTANCE == MANHATTAN)
				/* Manhattan distance */
				d =  (x >= 0? x : -x) + (y >= 0? y : -y) ; 
			#else
				d = 0 ; // default
			#endif

			return ( d ) ;

		}///}}}
	
		double PursuitNavigator::heuristic ( int xA, int yA, int xB, int yB)
		{///{{{
			return (distPos(xA,yA,xB,yB));
		}///}}}
	
		bool PursuitNavigator::navAStar( Cell_ID start, Cell_ID goal, std::vector<Cell_ID> &path )
		{///{{{

			path.clear();

			bool found = false ;

			std::list< boost::shared_ptr<pNavPos> > toDevelop ; 
			std::list< boost::shared_ptr<pNavPos> > positions ; 
			std::set< boost::shared_ptr<pNavPos>, ptr_cmp > visitedPos ; 

			int xStart, yStart, xGoal, yGoal ;
				xStart = start / ysize ;
				yStart = start % ysize ;
				xGoal  = goal  / ysize ;
				yGoal  = goal  % ysize ;

			if ( xStart >=0 && yStart >= 0 ) //inside the map
			if ( robGeoData->getValueFromBand(GEODATA_TFLAT_BAND, xStart/2, yStart/2 ) > 0.5 ) 
				toDevelop.push_back( boost::shared_ptr< pNavPos > (new pNavPos ( xStart, yStart, 0.0, heuristic( xStart, yStart, xGoal, yGoal), 0 )) ); 

			// Loop
			while ( !found && !toDevelop.empty() )
			{
				toDevelop.sort(ptr_greater< boost::shared_ptr <pNavPos> > () ) ; 

				positions.push_back(toDevelop.back());
				toDevelop.pop_back();

				// Test if we have reached the goal
				if (positions.back()->x == xGoal && positions.back()->y == yGoal )
					found = true ;
				else // develop the neighbours
					addNeighbours( positions.back(),xGoal, yGoal,visitedPos, toDevelop );
			}

			// Build the Path
			if (found)
			{
				boost::shared_ptr< pNavPos > ptr (positions.back()) ;
				path.push_back( ptr->x*ysize + ptr->y );	

				while ( !ptr->prev.empty()) 
				{
					// chose the best (the shortest path)
					ptr->prev.sort( ptr_less< boost::shared_ptr< pNavPos > >() )  ; 
					ptr = ptr->prev.front() ;

					path.push_back( ptr->x*ysize + ptr->y );	
				}
			}

			return(found);
		}///}}}
	
		bool PursuitNavigator::navAStar( int xStart, int yStart, int xGoal, int yGoal, std::vector<Cell_ID> &path )
		{///{{{

			path.clear();

			bool found = false ;

			std::list< boost::shared_ptr<pNavPos> > toDevelop ; 
			std::list< boost::shared_ptr<pNavPos> > positions ; 
			std::set< boost::shared_ptr<pNavPos>, ptr_cmp > visitedPos ; 

			if ( xStart >=0 && yStart >= 0 ) //inside the map
			if ( robGeoData->getValueFromBand(GEODATA_TFLAT_BAND, xStart/2, yStart/2 ) > 0.5 ) 
				toDevelop.push_back( boost::shared_ptr< pNavPos > (new pNavPos ( xStart, yStart, 0.0, heuristic( xStart, yStart, xGoal, yGoal), 0 )) ); 

			// Loop
			while ( !found && !toDevelop.empty() )
			{
				toDevelop.sort(ptr_greater< boost::shared_ptr <pNavPos> > () ) ; 

				positions.push_back(toDevelop.back());
				toDevelop.pop_back();

				// Test if we have reached the goal
				if (positions.back()->x == xGoal && positions.back()->y == yGoal )
					found = true ;
				else // develop the neighbours
					addNeighbours( positions.back(),xGoal, yGoal,visitedPos, toDevelop );
			}

			// Build the Path
			if (found)
			{
				boost::shared_ptr< pNavPos > ptr (positions.back()) ;
				path.push_back( ptr->x*ysize + ptr->y );	

				while ( !ptr->prev.empty()) 
				{
					// chose the best (the shortest path)
					ptr->prev.sort( ptr_less< boost::shared_ptr< pNavPos > >() )  ; 
					ptr = ptr->prev.front() ;

					path.push_back( ptr->x*ysize + ptr->y );	
				}
			}

			return(found);
		}///}}}
	
		bool PursuitNavigator::nav2Set( int xStart, int yStart, const std::set< Cell_ID > &goals, std::vector<Cell_ID> &path )
		{///{{{

			path.clear();

			bool found = false ;
			if (goals.empty()) return (found);

			std::list< boost::shared_ptr<pNavPos> > toDevelop ; 
			std::list< boost::shared_ptr<pNavPos> > positions ; 
			std::set< boost::shared_ptr<pNavPos>, ptr_cmp > visitedPos ; 

			if ( xStart >=0 && yStart >= 0 ) //inside the map
			if ( robGeoData->getValueFromBand(GEODATA_TFLAT_BAND, xStart/2, yStart/2 ) > 0.5 ) 
				toDevelop.push_back( boost::shared_ptr< pNavPos > (new pNavPos ( xStart, yStart, 0.0, 0.0, 0) )); 

			// Loop
			while ( !found && !toDevelop.empty() )
			{
				toDevelop.sort(ptr_greater< boost::shared_ptr <pNavPos> > () ) ; 

				positions.push_back(toDevelop.back());
				toDevelop.pop_back();

				// Test if we have reached the goal
				if ( goals.find( positions.back()->x * ysize + positions.back()->y ) != goals.end() )
					found = true ;
				else // develop the neighbours
					addNeighbours_classic( positions.back(),visitedPos, toDevelop );
			}

			// Build the Path
			if (found)
			{
				boost::shared_ptr< pNavPos > ptr (positions.back()) ;
				path.push_back( ptr->x*ysize + ptr->y );	

				while ( !ptr->prev.empty()) 
				{
					// chose the best (the shortest path)
					ptr->prev.sort( ptr_less< boost::shared_ptr< pNavPos > >() )  ; 
					ptr = ptr->prev.front() ;

					path.push_back( ptr->x*ysize + ptr->y );	
				}
			}

			return(found);
		}///}}}
	
		bool PursuitNavigator::nav2Set( Cell_ID start, const std::set< Cell_ID > &goals, std::vector<Cell_ID> &path )
		{///{{{

			path.clear();

			bool found = false ;
			if (goals.empty()) return (found);

			std::list< boost::shared_ptr<pNavPos> > toDevelop ; 
			std::list< boost::shared_ptr<pNavPos> > positions ; 
			std::set< boost::shared_ptr<pNavPos>, ptr_cmp > visitedPos ; 

			int xStart, yStart ;
				xStart = start / ysize ;
				yStart = start % ysize ;

			if ( xStart >=0 && yStart >= 0 ) //inside the map
			if ( robGeoData->getValueFromBand(GEODATA_TFLAT_BAND, xStart/2, yStart/2 ) > 0.5 ) 
				toDevelop.push_back( boost::shared_ptr< pNavPos > (new pNavPos ( xStart, yStart, 0.0, 0.0, 0 ) )); 

			// Loop
			while ( !found && !toDevelop.empty() )
			{
				toDevelop.sort(ptr_greater< boost::shared_ptr <pNavPos> > () ) ; 

				positions.push_back(toDevelop.back());
				toDevelop.pop_back();

				// Test if we have reached the goal
				if ( goals.find( positions.back()->x * ysize + positions.back()->y ) != goals.end() )
					found = true ;
				else // develop the neighbours
					addNeighbours_classic( positions.back(),visitedPos, toDevelop );
			}

			// Build the Path
			if (found)
			{
				boost::shared_ptr< pNavPos > ptr (positions.back()) ;
				path.push_back( ptr->x*ysize + ptr->y );	

				while ( !ptr->prev.empty()) 
				{
					// chose the best (the shortest path)
					ptr->prev.sort( ptr_less< boost::shared_ptr< pNavPos > >() )  ; 
					ptr = ptr->prev.front() ;

					path.push_back( ptr->x*ysize + ptr->y );	
				}
			}

			return(found);
		}///}}}
	
		bool PursuitNavigator::nav2Set( Cell_ID start, const std::set< Cell_ID > &goals, std::vector<Cell_ID> &path, const unsigned int &pursuitTemporalTolerance )
		{///{{{

			path.clear();

			bool found = false ;
			if (goals.empty()) return (found);

			std::list< boost::shared_ptr<pNavPos> > toDevelop ; 
			std::list< boost::shared_ptr<pNavPos> > positions ; 
			std::set< boost::shared_ptr<pNavPos>, ptr_cmp > visitedPos ; 

			int xStart, yStart ;
				xStart = start / ysize ;
				yStart = start % ysize ;

			if ( xStart >=0 && yStart >= 0 ) //inside the map
			if ( robGeoData->getValueFromBand(GEODATA_TFLAT_BAND, xStart/2, yStart/2 ) > 0.5 ) 
				toDevelop.push_back( boost::shared_ptr< pNavPos > (new pNavPos ( xStart, yStart, 0.0, 0.0, 0 ) )); 

			// Loop
			while ( !found && !toDevelop.empty() )
			{
				toDevelop.sort(ptr_greater< boost::shared_ptr <pNavPos> > () ) ; 

				positions.push_back(toDevelop.back());
				toDevelop.pop_back();

				// temporal limit
				if (positions.back()->t >= pursuitTemporalTolerance )
					continue;

				// Test if we have reached the goal
				if ( goals.find( positions.back()->x * ysize + positions.back()->y ) != goals.end() )
					found = true ;
				else // develop the neighbours
					addNeighbours_classic( positions.back(),visitedPos, toDevelop );
			}

			// Build the Path
			if (found)
			{
				boost::shared_ptr< pNavPos > ptr (positions.back()) ;
				path.push_back( ptr->x*ysize + ptr->y );	

				while ( !ptr->prev.empty()) 
				{
					// chose the best (the shortest path)
					ptr->prev.sort( ptr_less< boost::shared_ptr< pNavPos > >() )  ; 
					ptr = ptr->prev.front() ;

					path.push_back( ptr->x*ysize + ptr->y );	
				}
			}

			return(found);
		}///}}}
	
		void PursuitNavigator::addNeighbours( 	boost::shared_ptr< pNavPos> &currPos, 
												const int &xGoal, const int &yGoal, 
												std::set< boost::shared_ptr<pNavPos>, ptr_cmp > &visitedPos,  
												std::list< 	boost::shared_ptr<pNavPos> > &queue) 
		{///{{{
			int i,j,ii,jj ;

			// Movement model
			#if SWITCH_CONNEXITY
				// here : 5-cells connexity
				for ( i = -2 ; i <= 2 ; i+=2 )
				for ( j = -2 ; j <= 2 ; j+=2 )
				if ( i!= j || (i == 0 && j== 0 ) ) 	// 5-connexity
				{
					ii = i + currPos->x ;
					jj = j + currPos->y ;

					if (ii >=0 && jj >= 0 ) //inside the map
					if ( robGeoData->getValueFromBand(GEODATA_TFLAT_BAND, ii/2, jj/2 ) > 0.5 ) 
					{
						boost::shared_ptr <pNavPos> newPos ( new pNavPos( 	ii, jj, 
																			currPos->g + distPos (currPos->x, currPos->y,ii,jj),
																			heuristic(ii,jj,xGoal,yGoal), 
																			currPos->t + 1, 
																			currPos	)); 
						std::pair< std::set< boost::shared_ptr<pNavPos> >::const_iterator, bool> p = visitedPos.insert(newPos) ; 

						if ( p.second ) // if not yet visited
							queue.push_back( newPos );
						else
						{
							(*p.first)->prev.push_back( currPos ) ; // add the new parent

							//if necessary, update g
							if ( (*p.first)->g > newPos-> g ) 
								(*p.first)->g = newPos-> g ; 
						}
					}
				}

			#else
				// here : 3 x 3 cells : 9-cells connexity
				for ( i = -2 ; i <= 2 ; i+=2 )
				for ( j = -2 ; j <= 2 ; j+=2 )
				{
					ii = i + currPos->x ;
					jj = j + currPos->y ;

					if (ii >=0 && jj >= 0 ) //inside the map
					if ( robGeoData->getValueFromBand(GEODATA_TFLAT_BAND, ii/2, jj/2 ) > 0.5 ) 
					{
						boost::shared_ptr <pNavPos> newPos ( new pNavPos( 	ii, jj, 
																			currPos->g + distPos (currPos->x, currPos->y,ii,jj), 
																			heuristic(ii,jj,xGoal,yGoal), 
																			currPos->t + 1,
																			currPos )); 
						std::pair< std::set< boost::shared_ptr<pNavPos> >::const_iterator, bool> p = visitedPos.insert(newPos) ; 

						if ( p.second ) // if not yet visited
							queue.push_back( newPos );
						else
						{
							(*p.first)->prev.push_back( currPos ) ; // add the new parent

							//if necessary, update g
							if ( (*p.first)->g > newPos-> g ) 
								(*p.first)->g = newPos-> g ; 
						}
					}
				}

			#endif
			
		}///}}}

		void PursuitNavigator::addNeighbours_classic( 	boost::shared_ptr< pNavPos> &currPos, 
														std::set< boost::shared_ptr<pNavPos>, ptr_cmp > &visitedPos,  
														std::list< 	boost::shared_ptr<pNavPos> > &queue) 
		{///{{{
			int i,j,ii,jj ;

			// Movement model
			#if SWITCH_CONNEXITY
				// here : 5-cells connexity
				for ( i = -2 ; i <= 2 ; i+=2 )
				for ( j = -2 ; j <= 2 ; j+=2 )
				if ( i!= j || (i == 0 && j== 0 ) ) 	// 5-connexity
				{
					ii = i + currPos->x ;
					jj = j + currPos->y ;

					if (ii >=0 && jj >= 0 ) //inside the map
					if ( robGeoData->getValueFromBand(GEODATA_TFLAT_BAND, ii/2, jj/2 ) > 0.5 ) 
					{
						boost::shared_ptr<pNavPos> newPos ( new pNavPos( 	ii, jj, 
																			currPos->g + distPos (currPos->x,currPos->y,ii,jj), 
																			0.0, 
																			currPos->t+1, 
																			currPos )); 
						std::pair< std::set< boost::shared_ptr<pNavPos> >::const_iterator, bool> p = visitedPos.insert(newPos) ; 

						if ( p.second ) // if not yet visited
							queue.push_back( newPos );
						else
						{
							(*p.first)->prev.push_back( currPos ) ; // add the new parent

							//if necessary, update g
							if ( (*p.first)->g > newPos-> g ) 
								(*p.first)->g = newPos-> g ; 
						}
					}
				}

			#else
				// here : 3 x 3 cells : 9-cells connexity
				for ( i = -2 ; i <= 2 ; i+=2 )
				for ( j = -2 ; j <= 2 ; j+=2 )
				{
					ii = i + currPos->x ;
					jj = j + currPos->y ;

					if (ii >=0 && jj >= 0 ) //inside the map
					if ( robGeoData->getValueFromBand(GEODATA_TFLAT_BAND, ii/2, jj/2 ) > 0.5 ) 
					{
						boost::shared_ptr<pNavPos> newPos ( new pNavPos( 	ii, jj, 
																			currPos->g + distPos (currPos->x,currPos->y,ii,jj),
																			0.0,
																			currPos->t+1,
																			currPos )); 
						std::pair< std::set< boost::shared_ptr<pNavPos> >::const_iterator, bool> p = visitedPos.insert(newPos) ; 

						if ( p.second ) // if not yet visited
							queue.push_back( newPos );
						else
						{
							(*p.first)->prev.push_back( currPos ) ; // add the new parent

							//if necessary, update g
							if ( (*p.first)->g > newPos-> g ) 
								(*p.first)->g = newPos-> g ; 
						}
					}
				}

			#endif
			
		}///}}}

		GeoData * PursuitNavigator::getrobGeoData() 
		{///{{{
			return( robGeoData);
		}///}}}

		GeoData * PursuitNavigator::gettarGeoData()
		{///{{{
			return( tarGeoData);
		}///}}}

		unsigned int PursuitNavigator::getysize() const
		{///{{{
			return (ysize);
		}///}}}

		unsigned int PursuitNavigator::getxsize() const
		{///{{{
			return (xsize);
		}///}}}

		/****************************************************************/
		/********************** / PURSUIT_NAVIGATOR *********************/
		/****************************************************************/

	}
}

