/**
 * \file PursuitTaskManager.cpp
 *
 * Source file for the pursuit task manager class
 *
 * \author Cyril Robin
 *
 * \ingroup lgl
 *
 **/


#include <lgl/PursuitTaskManager.hpp>

#include <exception>
#include <sys/time.h>

#ifndef VERBOSE
#define VERBOSE 0
#endif

namespace jafar {
	namespace lgl {

		/****************************************************************/
		/************************* PURSUIT TASK *************************/
		/****************************************************************/

		bool PursuitTask::operator< (const PursuitTask &rhs) const
		{///{{{
			if ( time != rhs.time) 	return ( time < rhs.time) ;
			else 					return ( place < rhs.place ) ;
		}///}}}
	
		bool PursuitTask::operator== (const PursuitTask &rhs) const
		{///{{{
			return ( 	time 	== rhs.time 
					&&	place	== rhs.place
					&&	asking_robot == rhs.asking_robot
					);
		}///}}}
	

		/****************************************************************/
		/********************** PURSUIT TASK MANAGER ********************/
		/****************************************************************/
		PursuitTaskManager::PursuitTaskManager( int _robotID ) :
			robotID(_robotID)
		{///{{{
			countTaskID = 0 ;
		}///}}}

		void PursuitTaskManager::updateTaskList( const PursuitGraph *aPG)
		{///{{{

			if (aPG == NULL)
				return ;

			newTaskList.clear();
			obsoleteTasks.clear();
			nonattributedTasks.clear();

			// declaration
			std::map< PursuitTaskKey, PursuitTask > obsoleteTaskList ;
			std::map< PursuitTaskKey, PursuitTask > brandnewTaskList ;

			std::vector< boost::weak_ptr< PursuitGNode > >::const_iterator citV ;
			std::map< PursuitTaskKey, PursuitTask >::iterator itM ;
			std::set< Cell_ID >::const_iterator citS ;
			PursuitTaskKey key ;

			// Transforming dangers in tasks ///{{{
			// for each node
			for ( citV = aPG->getdanger().begin() ; citV != aPG->getdanger().end() ; ++citV )
			{
				boost::shared_ptr< PursuitGNode > ptr = citV->lock();
				if (ptr)
				{
					key.time = ptr->time;
					//for each position
					for ( citS = ptr->targetPos.begin() ; citS!= ptr->targetPos.end() ; ++citS )
					{
						key.place = *citS ;

						//newTaskList.insert( std::make_pair (key, PursuitTask(key.time, key.place, robotID, ++countTaskID) ));
						newTaskList.insert( std::make_pair (key, PursuitTask(key.time, key.place, robotID, -2) ));
					}
				}
			}
			///}}}
			
			// finding obsolete tasks
			set_difference(		oldTaskList.begin(), oldTaskList.end(),
								newTaskList.begin(), newTaskList.end(),
								inserter( obsoleteTaskList, obsoleteTaskList.begin())
								);

			// Dealing with the obsolete tasks 
			for ( itM = obsoleteTaskList.begin() ; itM != obsoleteTaskList.end() ; ++itM )
			{
			   //printf("[Task list] obsolete task : #id = #%d \n", itM->second.id );
			   obsoleteTasks.push_back( itM->second.id );
			}

			// Computing the brand new tasks
			// TODO delete this part (exists only for debug / verbose display purpose
			set_difference(	newTaskList.begin(), newTaskList.end(),
								oldTaskList.begin(), oldTaskList.end(),
								inserter( brandnewTaskList, brandnewTaskList.begin())
								);

			// Debug output / VERBOSE
			printf( "[Task list] : new task list = %u \n", newTaskList.size());
			printf( "[Task list] : brand new task list = %u \n", brandnewTaskList.size());
			printf( "[Task list] : old task list = %u \n", oldTaskList.size());
			printf( "[Task list] : obsolete task list = %u \n", obsoleteTaskList.size());

			// Dealing with the non-attributed tasks (ie those which do not have a robot to perform them yet)
			for ( itM = newTaskList.begin() ; itM != newTaskList.end() ; ++itM )
			{
				// attributing an ID if necessary
				if (itM->second.id < 0 ) 	
					itM->second.id = ++countTaskID ;

				// check is any robot is currently in charge of the task or not
				if (itM->second.robot_in_charge < 0 ) 
					nonattributedTasks.push_back(itM->second.robot_in_charge);
			}

			printf( "[Task list] : obsolete Tasks vector = %u \n", obsoleteTasks.size());
			printf( "[Task list] : non-attributed Tasks vector = %u \n", nonattributedTasks.size());

			// updating memory
			oldTaskList = newTaskList ;

			printf( "Task list updated! :-)  There are %u tasks by now ! \n", oldTaskList.size());

		}///}}}

	}//lgl
}//jafar

