/**
 * \file RobotPursuer.cpp
 *
 * Source file for the robot pursuer class
 *
 * \author Cyril Robin
 *
 * \ingroup lgl
 *
 **/


#include <lgl/RobotPursuer.hpp>

#include <exception>
#include <sys/time.h>

#ifndef VERBOSE
#define VERBOSE 0
#endif

#ifndef TIME_CONSTRAINT
#define TIME_CONSTRAINT true
#endif

#if TIME_CONSTRAINT
	#ifndef DEFAULT_TMAX
	#define DEFAULT_TMAX 800.0
	#endif
#endif


namespace jafar {
	namespace lgl {

		/****************************************************************/
		/************************ ROBOT PURSUER *************************/
		/****************************************************************/

		RobotPursuer::RobotPursuer( 	int _id,
										std::string _robotAccessibilityMap,
										std::string _targetAccessibilityMap,
										std::string _robotVisibilityMap,
										unsigned int _horizon, // temporal horizon
										unsigned int _viewLength, // viewLength
										unsigned int _temporalTolerance, 
										unsigned int _riskTermMaxValue, 
										double _sensorsHeight,
										ROBOT_STATE _state,
										Cell_ID _position 
				): ///{{{
					id(_id),
					horizon (_horizon),
					viewLength (_viewLength),
					temporalTolerance (_temporalTolerance),
					riskTermMaxValue (_riskTermMaxValue), 
					sensorsHeight (_sensorsHeight),
					state(_state),
					position(_position)
				{
					robotAccessibilityMap 	= _robotAccessibilityMap 	;
					targetAccessibilityMap 	= _targetAccessibilityMap 	;
					robotVisibilityMap 		= _robotVisibilityMap 		;
					
					navigator = new PursuitNavigator ( robotAccessibilityMap, targetAccessibilityMap) ; 	// Navigation data

					geodata = new GeoData (robotVisibilityMap);						// Visibility data
					geodata->loadVisibility(robotVisibilityMap);
					visibilityMap = new VisibilityMap ( geodata, viewLength, sensorsHeight ) ;	

					//PursuitGraph myPG (navigator, myVisibilityMap); 
					pursuitGraph = new PursuitGraph (*navigator, *visibilityMap, temporalTolerance, riskTermMaxValue); 

				}///}}}
			
		RobotPursuer::RobotPursuer( 	int _id,
								std::string _robotAccessibilityMap,
								std::string _targetAccessibilityMap,
								std::string _robotVisibilityMap,
								unsigned int _horizon, // temporal horizon
								unsigned int _viewLength, 
								unsigned int _temporalTolerance, 
								unsigned int _riskTermMaxValue, 
								double _sensorsHeight,
								ROBOT_STATE _state,
								int x, int y
				): ///{{{
					id(_id),
					horizon (_horizon),
					viewLength (_viewLength),
					temporalTolerance (_temporalTolerance),
					riskTermMaxValue (_riskTermMaxValue), 
					sensorsHeight (_sensorsHeight),
					state(_state)
				{
					robotAccessibilityMap 	= _robotAccessibilityMap 	;
					targetAccessibilityMap 	= _targetAccessibilityMap 	;
					robotVisibilityMap 		= _robotVisibilityMap 		;
					
					navigator = new PursuitNavigator ( robotAccessibilityMap, targetAccessibilityMap) ; 	// Navigation data

					geodata = new GeoData (robotVisibilityMap);						// Visibility data
					geodata->loadVisibility(robotVisibilityMap);
					visibilityMap = new VisibilityMap ( geodata, viewLength, sensorsHeight ) ;	

					//PursuitGraph myPG (navigator, myVisibilityMap); 
					pursuitGraph = new PursuitGraph (*navigator, *visibilityMap, temporalTolerance, riskTermMaxValue); 

					//TODO make safer (check)
					position = x*visibilityMap->getysize() + y ;
				}///}}}
			
		RobotPursuer::~RobotPursuer() 
		{///{{{
			if(pursuitGraph		!= NULL) delete pursuitGraph;
			if(visibilityMap	!= NULL) delete visibilityMap;
			if(geodata 			!= NULL) delete geodata ;
			if(navigator 		!= NULL) delete navigator ;
		}///}}}

		PursuitGraph* RobotPursuer::getPG() const
		{///{{{
			return( pursuitGraph );
		}///}}}

		std::string RobotPursuer::getRobotAccessibilityMap() const
		{///{{{
			return( robotAccessibilityMap );
		}///}}}

		std::string RobotPursuer::getTargetAccessibilityMap() const
		{///{{{
			return( targetAccessibilityMap );
		}///}}}

		std::string RobotPursuer::getRobotVisibilityMap() const
		{///{{{
			return( robotVisibilityMap );
		}///}}}

		Cell_ID RobotPursuer::getPosition() const
		{///{{{
			return( position );
		}///}}}

		void RobotPursuer::getPosition(int &x, int &y) const
		{///{{{
			x = position / pursuitGraph->getysize();
			y = position % pursuitGraph->getysize();
		}///}}}

		void RobotPursuer::start( int xTarget, int yTarget) 
		{///{{{

			switch ( state )
			{///{{{
				case IDLE :
					break;

				case BUSY :
					break;

				case SUPPORT :
					state = IDLE ; //cannot be SUPPORT when stating
					break;

				case TRACKER :
					startTracking(xTarget,yTarget);
					
					// Task Manager
					pursuitTaskManager.updateTaskList(pursuitGraph); 
					
					break;

				default:
					//throw exception ?
					;
			}///}}}

		}///}}}

		void RobotPursuer::run( int xTarget, int yTarget) 
		{///{{{

			#if TIME_CONSTRAINT
				double tMax = DEFAULT_TMAX ;   
			#endif

			timeval t;
			double dl;
			
			int xRobot, yRobot;

			switch ( state )
			{///{{{
				case IDLE :
					break;

				case BUSY :
					break;

				case SUPPORT :
					break;

				case TRACKER :
					//updating the Pursuit Graph
					gettimeofday(&t,NULL); dl = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0;
						pursuitGraph->updatePG ( xTarget, yTarget, xRobot, yRobot);

					gettimeofday(&t,NULL); dl = 1000.0 * t.tv_sec + t.tv_usec/ 1000.0 - dl;

					// completing the Pursuit Graph
					#if TIME_CONSTRAINT
						pursuitGraph->completeToHorizonBefore( ( tMax - dl > 0.0 ) ? tMax - dl : 0.0 ); 
					#else
						pursuitGraph->completeToHorizon( );
					#endif

					// Task Manager
					pursuitTaskManager.updateTaskList(pursuitGraph); 
					

					// Update its position
					position = xRobot*pursuitGraph->getysize()+yRobot;

					break;

				default:
					//throw exception ?
					;
			}///}}}

		}///}}}

		void RobotPursuer::startTracking(int xTarget, int yTarget) 
		{///{{{
			pursuitGraph->computeAll( position, xTarget*pursuitGraph->getysize() + yTarget, horizon);
		}///}}}

	}//lgl
}//jafar

