/**
 * \file VisibilityMap.cpp
 *
 * Source code for VisibilityMap Class
 * 
 *
 * \author Cyril Robin
 *
 * \ingroup lgl
 *
 **/

#include <lgl/VisibilityMap.hpp>

#include <ostream>

#ifndef VERBOSE
#define VERBOSE 1
#endif

namespace jafar { //namespace jafar
	namespace lgl { //namespace lgl
	

//------------------------------------------------------------------------//
//------------------ Visibility Map Implementation -----------------------//
//{{{ --------------------------------------------------------------------//

		VisibilityMap::VisibilityMap(	GeoData * _pt_Geodata,
										int	_viewLength,
										double _sensorsHeight,
										int _maxHeight
									 ):///{{{
										pt_Geodata (_pt_Geodata),
										viewLength (_viewLength),
										sensorsHeight(_sensorsHeight * 255 / DEFAULT_MAX_HEIGHT),
										maxHeight(_maxHeight)
		{

			if (pt_Geodata !=NULL)
			{
				xsize = 2 * pt_Geodata->getxsize() + 1 ; 	// 2N+1 vertices for N cells
				ysize = 2 * pt_Geodata->getysize() + 1 ;	// 2N+1 vertices for N cells
				
				#if VERBOSE
				std::cout << "#---III---  [VisibilityMap] xsize = " << xsize << " and ysize = " << ysize << std::endl;
				#endif
			}
			
			visibilityMap.clear();
			od_visibilityMap.assign( xsize*ysize, std::pair<bool, VisibleArea> (false, VisibleArea() ) ); 

		}///}}}
		
		VisibilityMap::~VisibilityMap()
		{///{{{

		}///}}}

		bool VisibilityMap::computeWholeVisibility () 	
		{///{{{

			if (pt_Geodata == NULL)
				return(false);

			// Local variables 
			unsigned long int i,j; 
			std::vector<RasterCellIndex> visibleCells;	// contains the cells visible from another given one
			std::set<Cell_ID> visibleVertices;			// contains the vertices visible from another given one
			std::vector<double> raysRatio ;				// contains ratio constraints for visibility ( = tan(angle) )

			// Memory allocation
			visibilityMap.clear();
			visibilityMap.resize(xsize*ysize);
				
			// Initialization
			for ( i=0 ; i < xsize ; i++ ) 					
				for ( j=0 ; j < ysize ; j++ ) 		
					visibilityMap.at(i*ysize+j) = VisibleArea( (Cell_ID) (i*ysize + j), xsize, ysize );

			// Computing exploration pattern for visibility for the given viewLength
			computePattern();

			// Computing the whole map
			// for each cell in geodata
			for ( i=0 ; i < xsize/2 ; i++ )
				for ( j=0 ; j < ysize/2 ; j++ )
				{///{{{

					// find the visible cells
					findVisibleCells (i,j, visibleCells, raysRatio );

					// transform each cell of the list into its corresponding vertices
					///{{{
					visibleVertices.clear();

					std::vector<RasterCellIndex>::iterator it ;
					for ( it=visibleCells.begin() ; it!=visibleCells.end() ; ++it )
					{

							//*
							// * NB: NavGNode osbolete, use Cell_ID (= x * ysize + y) instead
							//visibleVertices.insert( ( 2*it->i  )*ysize + 2*it->j   );
							//visibleVertices.insert( ( 2*it->i+1)*ysize + 2*it->j   );
							//visibleVertices.insert( ( 2*it->i+2)*ysize + 2*it->j   );
							//visibleVertices.insert( ( 2*it->i  )*ysize + 2*it->j+1 );
							//visibleVertices.insert( ( 2*it->i+1)*ysize + 2*it->j+1 );
							//visibleVertices.insert( ( 2*it->i+2)*ysize + 2*it->j+1 );
							//visibleVertices.insert( ( 2*it->i  )*ysize + 2*it->j+2 );
							//visibleVertices.insert( ( 2*it->i+1)*ysize + 2*it->j+2 );
							//visibleVertices.insert( ( 2*it->i+2)*ysize + 2*it->j+2 );
							//*/

							/* NB : round the nodes of the pixel to the center of the corresponding pixel
							 * (for computational and, above all, memory reason)
							 */
							visibleVertices.insert( (2*it->i+1)* ysize + (2*it->j+1) ) ;

					}///}}}

					// Update the current vertices in the visibility map
					///{{{
					/*
					visibilityMap[2*i  ][2*j  ].updateArea(visibleVertices);
					visibilityMap[2*i+1][2*j  ].updateArea(visibleVertices);
					visibilityMap[2*i+2][2*j  ].updateArea(visibleVertices);
					visibilityMap[2*i  ][2*j+1].updateArea(visibleVertices);
					visibilityMap[2*i+1][2*j+1].updateArea(visibleVertices);
					visibilityMap[2*i+2][2*j+1].updateArea(visibleVertices);
					visibilityMap[2*i  ][2*j+2].updateArea(visibleVertices);
					visibilityMap[2*i+1][2*j+2].updateArea(visibleVertices);
					visibilityMap[2*i+2][2*j+2].updateArea(visibleVertices);
					*/

					/* NB : round the nodes of the pixel to the center of the corresponding pixel
					 * (for computational and, above all, memory reason)
					 */
					visibilityMap.at( (2*i+1) * ysize + (2*j+1) ).updateArea(visibleVertices);

					///}}}
					
				}///}}}

			
			return (true);
		}///}}}

		bool VisibilityMap::clearVisibility()
		{///{{{
			visibilityMap.clear();
			currentPattern.clear();

			return(true) ;
		}///}}}

		bool VisibilityMap::computePartialVisibility (std::map<Cell_ID,VisibleArea> &partialVisibilityMap, const std::set<Cell_ID> &cellSet ) 	
		{///{{{

			if (pt_Geodata == NULL)
				return(false);

			// Local variables 
			int i,j; 
			std::vector<RasterCellIndex> visibleCells;	// contains the cells visible from another given one
			std::set<Cell_ID> visibleVertices;			// contains the vertices visible from another given one
			std::vector<double> raysRatio ;				// contains ratio constraints for visibility ( = tan(angle) )
			std::set<Cell_ID>::const_iterator it1 ;		// iterator for cellSet

			// Computing exploration pattern for visibility for the given viewLength
			if ( currentPattern.empty() )
				computePattern();

			// computing visibility for each cell of cellSet
			for ( it1=cellSet.begin() ; it1!=cellSet.end() ; ++it1 )
			{///{{{
					// convert Cell_ID into coordinates
					i = (*it1 / ysize) / 2 ;
					j = (*it1 - (*it1 / ysize) * ysize) / 2 ;
					
					// find the visible cells
					findVisibleCells (i,j, visibleCells, raysRatio );

					// transform each cell of the list into its corresponding vertices
					///{{{
					visibleVertices.clear();

					std::vector<RasterCellIndex>::iterator it2 ;
					for ( it2=visibleCells.begin() ; it2!=visibleCells.end() ; ++it2 )
					{

							/*
							 * NB: NavGNode osbolete, use Cell_ID (= x * ysize + y) instead
							visibleVertices.insert( ( 2*it2->i  )*ysize + 2*it2->j   );
							visibleVertices.insert( ( 2*it2->i+1)*ysize + 2*it2->j   );
							visibleVertices.insert( ( 2*it2->i+2)*ysize + 2*it2->j   );
							visibleVertices.insert( ( 2*it2->i  )*ysize + 2*it2->j+1 );
							visibleVertices.insert( ( 2*it2->i+1)*ysize + 2*it2->j+1 );
							visibleVertices.insert( ( 2*it2->i+2)*ysize + 2*it2->j+1 );
							visibleVertices.insert( ( 2*it2->i  )*ysize + 2*it2->j+2 );
							visibleVertices.insert( ( 2*it2->i+1)*ysize + 2*it2->j+2 );
							visibleVertices.insert( ( 2*it2->i+2)*ysize + 2*it2->j+2 );
							*/

							/* NB : round the nodes of the pixel to the center of the corresponding pixel
							 * (for computational and, above all, memory reason)
							 */
							visibleVertices.insert( (2*it2->i+1)*ysize + (2*it2->j+1) ) ;

					}///}}}

					// Update the current vertices in the visibility map
					///{{{

					/* NB : round the nodes of the pixel to the center of the corresponding pixel
					 * (for computational and, above all, memory reason)
					 */
					partialVisibilityMap.insert( std::pair<Cell_ID,VisibleArea> ( *it1 , VisibleArea( *it1 , xsize, ysize ) ) ) ;
					partialVisibilityMap.find( *it1 )->second.updateArea(visibleVertices);

					///}}}
					
				}///}}}

			
			return (true);
		}///}}}

		bool VisibilityMap::computeVisibilityOnDemand (std::map<Cell_ID,VisibleArea> &partialVisibilityMap, const std::set<Cell_ID> &cellSet ) 	
		{///{{{

			if (pt_Geodata == NULL)
				return(false);

			if ( od_visibilityMap.size() < xsize*ysize )
				od_visibilityMap.assign( xsize*ysize, std::pair<bool, VisibleArea> (false, VisibleArea() ) ); 

			// Local variables 
			int i,j; 
			std::vector<RasterCellIndex> visibleCells;	// contains the cells visible from another given one
			std::set<Cell_ID> visibleVertices;			// contains the vertices visible from another given one
			std::vector<double> raysRatio ;				// contains ratio constraints for visibility ( = tan(angle) )
			std::set<Cell_ID>::const_iterator it1 ;		// iterator for cellSet

			// Computing exploration pattern for visibility for the given viewLength
			if ( currentPattern.empty() )
				computePattern();

			// computing visibility for each cell of cellSet
			for ( it1=cellSet.begin() ; it1!=cellSet.end() ; ++it1 )
			{
				if ( !(od_visibilityMap.at(*it1).first) ) // not yet computed => compute visibility
				{///{{{
					// convert Cell_ID into coordinates
					i = (*it1 / ysize) / 2 ;
					j = (*it1 - (*it1 / ysize) * ysize) / 2 ;
					

					// find the visible cells
					findVisibleCells (i,j, visibleCells, raysRatio );

					// transform each cell of the list into its corresponding vertices
					///{{{
					visibleVertices.clear();

					std::vector<RasterCellIndex>::iterator it2 ;
					for ( it2=visibleCells.begin() ; it2!=visibleCells.end() ; ++it2 )
					{

							/*
							 * NB: NavGNode osbolete, use Cell_ID (= x * ysize + y) instead
							visibleVertices.insert( ( 2*it2->i  )*ysize + 2*it2->j   );
							visibleVertices.insert( ( 2*it2->i+1)*ysize + 2*it2->j   );
							visibleVertices.insert( ( 2*it2->i+2)*ysize + 2*it2->j   );
							visibleVertices.insert( ( 2*it2->i  )*ysize + 2*it2->j+1 );
							visibleVertices.insert( ( 2*it2->i+1)*ysize + 2*it2->j+1 );
							visibleVertices.insert( ( 2*it2->i+2)*ysize + 2*it2->j+1 );
							visibleVertices.insert( ( 2*it2->i  )*ysize + 2*it2->j+2 );
							visibleVertices.insert( ( 2*it2->i+1)*ysize + 2*it2->j+2 );
							visibleVertices.insert( ( 2*it2->i+2)*ysize + 2*it2->j+2 );
							*/

							/* NB : round the nodes of the pixel to the center of the corresponding pixel
							 * (for computational and, above all, memory reason)
							 */
							visibleVertices.insert( (2*it2->i+1)*ysize + (2*it2->j+1) ) ;

					}///}}}

					// Update the current vertices in the od_visibilityMap
					///{{{

					/* NB : round the nodes of the pixel to the center of the corresponding pixel
					 * (for computational and, above all, memory reason)
					 */
					od_visibilityMap.at(*it1)= std::pair< bool,VisibleArea> ( true , VisibleArea( *it1 , xsize, ysize ) );
					od_visibilityMap.at(*it1).second.updateArea(visibleVertices);

					///}}}
					
				}///}}}

				partialVisibilityMap.insert( std::pair<Cell_ID,VisibleArea> ( *it1 , od_visibilityMap.at(*it1).second ) );
			}

			
			return (true);
		}///}}}

		bool VisibilityMap::computeVisibilityOnDemand (std::map<Cell_ID, const VisibleArea*> &partialVisibilityMap, const std::set<Cell_ID> &cellSet ) 	
		{///{{{	

			if (pt_Geodata == NULL)
				return(false);

			if ( od_visibilityMap.size() < xsize*ysize )
				od_visibilityMap.assign( xsize*ysize, std::pair<bool, VisibleArea> (false, VisibleArea() ) ); 

			// Local variables 
			int i,j; 
			std::vector<RasterCellIndex> visibleCells;	// contains the cells visible from another given one
			std::set<Cell_ID> visibleVertices;			// contains the vertices visible from another given one
			std::vector<double> raysRatio ;				// contains ratio constraints for visibility ( = tan(angle) )
			std::set<Cell_ID>::const_iterator it1 ;		// iterator for cellSet

			// Computing exploration pattern for visibility for the given viewLength
			if ( currentPattern.empty() )
				computePattern();

			// computing visibility for each cell of cellSet
			for ( it1=cellSet.begin() ; it1!=cellSet.end() ; ++it1 )
			{
				if ( !(od_visibilityMap.at(*it1).first) ) // not yet computed => compute visibility
				{///{{{
					// convert Cell_ID into coordinates
					i = (*it1 / ysize) / 2 ;
					j = (*it1 - (*it1 / ysize) * ysize) / 2 ;
					

					// find the visible cells
					findVisibleCells (i,j, visibleCells, raysRatio );

					// transform each cell of the list into its corresponding vertices
					///{{{
					visibleVertices.clear();

					std::vector<RasterCellIndex>::iterator it2 ;
					for ( it2=visibleCells.begin() ; it2!=visibleCells.end() ; ++it2 )
					{

							/*
							 * NB: NavGNode osbolete, use Cell_ID (= x * ysize + y) instead
							visibleVertices.insert( ( 2*it2->i  )*ysize + 2*it2->j   );
							visibleVertices.insert( ( 2*it2->i+1)*ysize + 2*it2->j   );
							visibleVertices.insert( ( 2*it2->i+2)*ysize + 2*it2->j   );
							visibleVertices.insert( ( 2*it2->i  )*ysize + 2*it2->j+1 );
							visibleVertices.insert( ( 2*it2->i+1)*ysize + 2*it2->j+1 );
							visibleVertices.insert( ( 2*it2->i+2)*ysize + 2*it2->j+1 );
							visibleVertices.insert( ( 2*it2->i  )*ysize + 2*it2->j+2 );
							visibleVertices.insert( ( 2*it2->i+1)*ysize + 2*it2->j+2 );
							visibleVertices.insert( ( 2*it2->i+2)*ysize + 2*it2->j+2 );
							*/

							/* NB : round the nodes of the pixel to the center of the corresponding pixel
							 * (for computational and, above all, memory reason)
							 */
							visibleVertices.insert( (2*it2->i+1)*ysize + (2*it2->j+1) ) ;

					}///}}}

					// Update the current vertices in the od_visibilityMap
					///{{{

					/* NB : round the nodes of the pixel to the center of the corresponding pixel
					 * (for computational and, above all, memory reason)
					 */
					od_visibilityMap.at(*it1)= std::pair< bool,VisibleArea> ( true , VisibleArea( *it1 , xsize, ysize ) );
					od_visibilityMap.at(*it1).second.updateArea(visibleVertices);

					///}}}
					
				}///}}}

				partialVisibilityMap.insert( std::pair<Cell_ID,VisibleArea*> ( *it1 , &od_visibilityMap.at(*it1).second ) );
			}


			
			return (true);
		}///}}}

		bool VisibilityMap::getPrecomputedVisibility(std::map<Cell_ID,const VisibleArea*> &partialVisibilityMap, const std::set<Cell_ID> &cellSet ) 	
		{///{{{

			if (pt_Geodata == NULL)
				return(false);
			
			// if the visibilityMap has not been computed yet, it does the job
			if ( visibilityMap.size() < xsize*ysize )
				computeWholeVisibility();

			// Local variables 
			std::set<Cell_ID>::const_iterator it1 ;		// iterator for cellSet

			// computing visibility for each cell of cellSet
			for ( it1=cellSet.begin() ; it1!=cellSet.end() ; ++it1 )
				partialVisibilityMap.insert( std::pair<Cell_ID,VisibleArea*> ( *it1 , &visibilityMap.at(*it1) ) );
			
			return (true);
		}///}}}

		bool VisibilityMap::getPrecomputedVisibility(std::map<Cell_ID,VisibleArea> &partialVisibilityMap, const std::set<Cell_ID> &cellSet ) 	
		{///{{{

			if (pt_Geodata == NULL)
				return(false);
			
			// if the visibilityMap has not been computed yet, it does the job
			if ( visibilityMap.size() < xsize*ysize )
				computeWholeVisibility();

			// Local variables 
			std::set<Cell_ID>::const_iterator it1 ;		// iterator for cellSet

			// computing visibility for each cell of cellSet
			for ( it1=cellSet.begin() ; it1!=cellSet.end() ; ++it1 )
				partialVisibilityMap.insert( std::pair<Cell_ID,VisibleArea> ( *it1 , visibilityMap.at(*it1) ) );
			
			return (true);
		}///}}}

		void VisibilityMap::findVisibleCells( const int i, const int j, std::vector<RasterCellIndex> &cellList, std::vector<double> &raysRatio )
		{///{{{
			
			cellList.clear();
		
			raysRatio.clear();

			// if the cell is really visible 
			// => NO! may see far away...
			//if ( pt_Geodata->getValueFromBand(GEODATA_VISIBILITY_BAND, i,j) > MIN_VISIBILITY_VALUE ) 
			{///{{{
				
				// the current cell is visible from itself
				if ( pt_Geodata->getValueFromBand(GEODATA_VISIBILITY_BAND, i,j) > MIN_VISIBILITY_VALUE ) 
					cellList.push_back(RasterCellIndex(i,j));

				// initialization
				int x,y ; 					// coordinates of the observed cell
				double vCell, hCell, h0 ;	// visibility and visibility height value of the observed cell
											// h0 is the heigth of the cell (i,j), used as reference
											// d is the distance between the two considered cells
				double cRatio ; 			// ratio for the observed cell
				bool b ; 					// tell if the observed cell is visible or not

				// iterators
				std::vector<PatternPoint>::const_iterator it ;
				std::vector<CellRay>::const_iterator itr ;

				h0 = pt_Geodata->getValueFromBand( GEODATA_VISIBILITY_HEIGHT_BAND,i,j);

				/* NB: initialization must be (-infinity)
				 * here, -1 is enough (because Ratio will always be 1> cRatio > -1)
				 */
				raysRatio.resize(n,-1) ;
				
				// Loop on the surrounding cells using the pattern
				for (it = currentPattern.begin(); it!=currentPattern.end() ; ++it )
				{///{{{

					b = false ;

					x = i + it->x ;
					y = j + it->y ;

					vCell = pt_Geodata->getValueFromBand( GEODATA_VISIBILITY_BAND,x,y);
					hCell = pt_Geodata->getValueFromBand( GEODATA_VISIBILITY_HEIGHT_BAND,x,y);

					/* inside the map : 1 >= visibility >= 0
					 * NB: here we may have vCell < MIN_VISIBILITY_VALUE
					 * because an invisible cell may still hide other visible cells.
					 */
					if (vCell >= 0) 
					{

						// for each rays intersecting the current observed surrouding cell
						for (itr = (*it).rays.begin(); itr!=(*it).rays.end() ; ++itr )
						{

							// ratio = (relative height value (including sensors height )) / relative distance value
							cRatio = (double) (hCell -h0 - ((double) sensorsHeight/255.0) ) /  itr->dist ; 

							if ( raysRatio[itr->id] <= cRatio + 0.0000000001 ) // avoid bug due to double precision
							{
								raysRatio[itr->id] = cRatio ;
								b = true;
							} 
						}
						
						/* if at least one ray crosses the cell with a sufficient ratio
						 * and if the cell has a minimal value
						 */
						if ( b  && vCell > MIN_VISIBILITY_VALUE )
							cellList.push_back(RasterCellIndex(x,y));

					}//if


				}///}}}

			}///}}}  /if (cell visible)

		}///}}}

		void VisibilityMap::computePattern()
		{///{{{

				// initialization
				currentPattern.clear();

				int k,l;		// k is the considered range of cells (circle)
								// l is the deviation from the axes

				n = ceil ( PI * ( 2 * viewLength + 1 )  / 1.0 ); 	// minimal number of rays
				
				if (n%2 !=0) n++ ; //nbr pair de rayon
				alpha = (double) (2 * PI / n) ;						// minimal angle ( = angular step between rays)

				#if VERBOSE
				std::cout 	<< "#---III---  [VisibilityMap] ViewLength = " << viewLength 
								<< " => Nbr of Rays = " << n 
								<< "  and Step Angle Alpha = " << 180 * ( alpha  / PI ) 
								<< std::endl;
				#endif

				/* NB : each ray is identified by his number (1..n) and associated to a specific angle (n*alpha)
				 */
				
				/* compute visibility for the other cells
				 * Loop on growing circle until viewLength is reached
				 */
				k = 1 ;

				while ( k <= viewLength  )  // the number of considered visibility "circles"  
				{ ///{{{

					/* /!\ NB : the order is VERY important... /!\ */

					// l = 0 : the cross
					l = 0;

					if ( sqrt ( k*k + 0*0 ) <= viewLength )
					{
						currentPattern.push_back( computeCurrentPoint(  k,  0 ) );
						currentPattern.push_back( computeCurrentPoint(  0,  k ) );
						currentPattern.push_back( computeCurrentPoint( -k,  0 ) );
						currentPattern.push_back( computeCurrentPoint(  0, -k ) );
					}

					// between cross and corners
					for ( l = 1  ; l < k ; l ++ )
					{///{{{

						if ( sqrt ( k*k + l*l ) <= viewLength )
						{
							currentPattern.push_back( computeCurrentPoint(  k,  l ) );
							currentPattern.push_back( computeCurrentPoint(  l,  k ) );
							currentPattern.push_back( computeCurrentPoint( -l,  k ) );
							currentPattern.push_back( computeCurrentPoint( -k,  l ) );
							currentPattern.push_back( computeCurrentPoint( -k, -l ) );
							currentPattern.push_back( computeCurrentPoint( -l, -k ) );
							currentPattern.push_back( computeCurrentPoint(  l, -k ) );
							currentPattern.push_back( computeCurrentPoint(  k, -l ) );
						}

					}///}}}
				
					// l = k : the corners
					if ( sqrt ( k*k + k*k ) <= viewLength )
					{
						currentPattern.push_back( computeCurrentPoint(  k,  k ) );
						currentPattern.push_back( computeCurrentPoint( -k, -k ) );
						currentPattern.push_back( computeCurrentPoint(  k, -k ) );
						currentPattern.push_back( computeCurrentPoint( -k,  k ) );
					}

					k++ ;

				} ///}}}
				
		}///}}}


		PatternPoint VisibilityMap::computeCurrentPoint( const int i, const int j) 
		{///{{{

				std::vector<CellRay> raysList ;
				double firstAngle, secondAngle, thirdAngle, d ;
				unsigned int a, b;

				int i1,j1,i2,j2,i3,j3 ; // extremities of the cell

 				/* NB: use the counterclockwise (trigonometric) direction */
				/* BEWARE of the direction of the axis in lgl ! (while computing angle : axis y is reversed) */


				// find the extremities of the cell
				///{{{
				if 		( i >  0 && j <  0 ) //top right quarter 
				{
					i1 = 2*i+1; 
					j1 = 2*j+1;
					i2 = 2*i-1;
					j2 = 2*j-1;
					i3 = 2*i-1;
					j3 = 2*j+1;
				}
				else if ( i <  0 && j <  0 ) // top left quarter
				{
					i1 = 2*i+1; 
					j1 = 2*j-1;
					i2 = 2*i-1;
					j2 = 2*j+1;
					i3 = 2*i+1;
					j3 = 2*j+1;
				}
				else if ( i <  0 && j >  0 ) // bottom left quarter
				{
					i1 = 2*i-1; 
					j1 = 2*j-1;
					i2 = 2*i+1;
					j2 = 2*j+1;
					i3 = 2*i+1;
					j3 = 2*j-1;
				}
				else if ( i >  0 && j >  0 ) // bottom right quarter
				{
					i1 = 2*i-1; 
					j1 = 2*j+1;
					i2 = 2*i+1;
					j2 = 2*j-1;
					i3 = 2*i-1;
					j3 = 2*j-1;
				}
				else if ( i == 0 && j >  0 ) // cross down
				{
					i1 = 	-1; 
					j1 = 2*j-1;
					i2 = 	+1;
					j2 = 2*j-1;
					i3 =     0;
					j3 = 2*j-1;
				} 
				else if ( i == 0 && j <  0 ) // cross up 
				{
					i1 =    +1; 
					j1 = 2*j+1;
					i2 =    -1;
					j2 = 2*j+1;
					i3 =     0;
					j3 = 2*j+1;
				} 
				else if ( i <  0 && j == 0 )  // cross left
				{
					i1 = 2*i+1; 
					j1 =    -1;
					i2 = 2*i+1;
					j2 =    +1;
					i3 = 2*i+1;
					j3 = 0  ;
				} 
				else if ( i >  0 && j == 0 ) // cross right
				{
					i1 = 2*i-1; 
					j1 = 	+1;
					i2 = 2*i-1;
					j2 =    -1;
					i3 = 2*i-1;
					j3 =     0;
				} 
				///}}}

				// compute angles
				firstAngle 	= computeAngle( i1, -j1 ); // "-j" because y-axis is inversed
				secondAngle	= computeAngle( i2, -j2 ); // "-j" because y-axis is inversed		
				thirdAngle	= computeAngle( i3, -j3 ); // "-j" because y-axis is inversed		

				// find the rays intersecting the cell (i,j), considering its extremities
				a = (unsigned int) ceil  ( firstAngle / alpha) ;
				b = (unsigned int) floor (secondAngle / alpha) ;

				d = computeDistance( i1, -j1, i2, -j2, i3, -j3,thirdAngle, a ) ; // "-j" because y-axis is inversed
				raysList.push_back( CellRay( a % n, d) );

				if (a <=b ) 		// manage the continuity between 2PI and 0
					while ( a < b) 		// find the rays between the cell extremities	
					{
						a++;
						d = computeDistance( i1, -j1, i2, -j2, i3, -j3,thirdAngle, a ) ; // "-j" because y-axis is inversed
						raysList.push_back( CellRay( a % n, d) );
					}
				else 
				{
					b+= n;
					while ( a < b)		// find the rays between the cell extremities	
					{
						a++;
						d = computeDistance( i1, -j1, i2, -j2, i3, -j3,thirdAngle, a ) ; // "-j" because y-axis is inversed
						raysList.push_back( CellRay( a % n, d) );
					}
				}
					
				return( PatternPoint(i,j,raysList) );

		}///}}}

		double VisibilityMap::computeAngle(const int i, const int j) 
		{///{{{
			double result ;

			result = acos (  i / sqrt ( i*i + j*j) ) ;

			if ( j < 0 ) 
				result= -result ;

			// angle between 0 and 2Pi
			if (result >= 0 ) 
				return (result) ;
			else 
				return (result + 2*PI);
		}///}}}

		double VisibilityMap::computeDistance( 	const int i1, const int j1,
												const int i2, const int j2,
												const int i3, const int j3,
												const double thirdAngle,
												const int r
											)
		{///{{{
	
			/* NB: use euclidian distance */

			double x,y;		// the intersection between the ray and the cell

			if ( alpha * r < thirdAngle)
			{
				// between 1 and 3
				findIntersection ( x, y, i1, j1, i3, j3, alpha * r );

			} else {
				// between 3 and 2
				findIntersection ( x, y, i3, j3, i2, j2, alpha * r );

			}	
				
			return ( sqrt (x*x + y*y) );
		}///}}}

		bool VisibilityMap::findIntersection( double &x, double &y, const int i1, const int j1, const int i2, const int j2, const double angle)
		{///{{{

			double a, b, c ;

			if (angle != PI/2 && angle != 3*PI/2 ) // tan(angle) is defined
			{			
				c = tan (angle) ;

				if ( i1 != i2 )
				{
					a =  (j2-j1) / ( (double) (i2-i1) ) ;
					b =  j2 - a * i2 ;

					x =  b / (c - a); // y = c.x and y = a.x + b 
					y =  a * x + b ;

				} else {
					
					x = i1 ;

					if (j1 == j2) 
					{
						std::cerr << "---WWW---  Problem with the arguments ; P1 and P2 are the same point ("<<i1<<","<<j1<<")."<<std::endl;
						y=j1 ;
						return(false) ;
					}
					else
					{
						y = c * x ;
					}
				}
			} 
			else
			{
				x = 0 ;
				y = (j1 + j2 ) / 2 ;
			}

			return (true);
		}///}}}

		/** Get functions 
		 */

		VisibleArea VisibilityMap::getVisibility(int i, int j) const
		{///{{{

			unsigned long int ii , jj ;

			if (i  >= 0) 						ii = i ; 		else 	ii = 0 ;
			if (j  >= 0) 						jj = j ;		else 	jj = 0 ;
			if (ii >= xsize && xsize > 0)	 	ii = xsize-1 ;
			if (jj >= ysize && ysize > 0) 	 	jj = ysize-1 ;

			#if VERBOSE
			if (i != int (i) || j != int (jj) )
				std::cout << "#---WWW---  Node ("<<i<<","<<j<<") is outside the map - Replace by (" << ii << "," << jj << ")" << std::endl;
			#endif

			/* real point
			 */
			//return( visibilityMap.at( ii * ysize +  jj ) );
			
			/* approximate point
			 * NB : round the node to the center of the corresponding pixel (for computational and, above all, memory reason)
			 */
			#if VERBOSE
			std::cout << "#---WWW---  Node ("<<ii<<","<<jj<<") in the visibility Map has been rounded to (" ;
			#endif

				ii = 2 * (ii/2) + 1 ;
				jj = 2 * (jj/2) + 1 ;
				if (ii > xsize-2) 	ii = xsize-2 ;
				if (jj > ysize-2) 	jj = ysize-2 ;

			#if VERBOSE
			std::cout << ii << "," << jj << ")" << std::endl;
			#endif

			return( visibilityMap.at( ii * ysize +  jj ) );

		}///}}}	

		const VisibleArea* VisibilityMap::getPtVisibility(int i, int j) const
		{///{{{

			unsigned long int ii , jj ;

			if (i  >= 0) 						ii = i ; 		else 	ii = 0 ;
			if (j  >= 0) 						jj = j ;		else 	jj = 0 ;
			if (ii >= xsize && xsize > 0)	 	ii = xsize-1 ;
			if (jj >= ysize && ysize > 0) 	 	jj = ysize-1 ;

			#if VERBOSE
			if (i != int (i) || j != int (jj) )
				std::cout << "#---WWW---  Node ("<<i<<","<<j<<") is outside the map - Replace by (" << ii << "," << jj << ")" << std::endl;
			#endif

			/* real point
			 */
			//return( &visibilityMap.at( ii * ysize +  jj ) );
			
			/* approximate point
			 * NB : round the node to the center of the corresponding pixel (for computational and, above all, memory reason)
			 */
			#if VERBOSE
			std::cout << "#---WWW---  Node ("<<ii<<","<<jj<<") in the visibility Map has been rounded to (" ;
			#endif

				ii = 2 * (ii/2) + 1 ;
				jj = 2 * (jj/2) + 1 ;
				if (ii > xsize-2) 	ii = xsize-2 ;
				if (jj > ysize-2) 	jj = ysize-2 ;
			
			#if VERBOSE
			std::cout << ii << "," << jj << ")" << std::endl;
			#endif

			return( &visibilityMap.at( ii * ysize +  jj ) );

		}///}}}

		const VisibleArea* VisibilityMap::getPtVisibilityOnDemand(Cell_ID cell) 
		{///{{{	

			if (pt_Geodata == NULL)
				return(false);

			if ( od_visibilityMap.size() < xsize*ysize )
				od_visibilityMap.assign( xsize*ysize, std::pair<bool, VisibleArea> (false, VisibleArea() ) ); 

			// Local variables 
			int i,j; 
			std::vector<RasterCellIndex> visibleCells;	// contains the cells visible from another given one
			std::set<Cell_ID> visibleVertices;			// contains the vertices visible from another given one
			std::vector<double> raysRatio ;				// contains ratio constraints for visibility ( = tan(angle) )

			// Computing exploration pattern for visibility for the given viewLength
			if ( currentPattern.empty() )
				computePattern();

			// computing visibility the cell
			if ( !(od_visibilityMap.at(cell).first) ) // not yet computed => compute visibility
			{///{{{
				// convert Cell_ID into coordinates
				i = (cell / ysize) / 2 ;
				j = (cell % ysize) / 2 ;

				// find the visible cells
				findVisibleCells (i,j, visibleCells, raysRatio );

				// transform each cell of the list into its corresponding vertices
				///{{{
				visibleVertices.clear();

				std::vector<RasterCellIndex>::iterator it2 ;
				for ( it2=visibleCells.begin() ; it2!=visibleCells.end() ; ++it2 )
				{

						/*
						 * NB: NavGNode osbolete, use Cell_ID (= x * ysize + y) instead
						visibleVertices.insert( ( 2*it2->i  )*ysize + 2*it2->j   );
						visibleVertices.insert( ( 2*it2->i+1)*ysize + 2*it2->j   );
						visibleVertices.insert( ( 2*it2->i+2)*ysize + 2*it2->j   );
						visibleVertices.insert( ( 2*it2->i  )*ysize + 2*it2->j+1 );
						visibleVertices.insert( ( 2*it2->i+1)*ysize + 2*it2->j+1 );
						visibleVertices.insert( ( 2*it2->i+2)*ysize + 2*it2->j+1 );
						visibleVertices.insert( ( 2*it2->i  )*ysize + 2*it2->j+2 );
						visibleVertices.insert( ( 2*it2->i+1)*ysize + 2*it2->j+2 );
						visibleVertices.insert( ( 2*it2->i+2)*ysize + 2*it2->j+2 );
						*/

						/* NB : round the nodes of the pixel to the center of the corresponding pixel
						 * (for computational and, above all, memory reason)
						 */
						visibleVertices.insert( (2*it2->i+1)*ysize + (2*it2->j+1) ) ;

				}///}}}

				// Update the current vertices in the od_visibilityMap
				///{{{

				/* NB : round the nodes of the pixel to the center of the corresponding pixel
				 * (for computational and, above all, memory reason)
				 */
				od_visibilityMap.at(cell)= std::pair< bool,VisibleArea> ( true , VisibleArea( cell , xsize, ysize ) );
				od_visibilityMap.at(cell).second.updateArea(visibleVertices);

				///}}}
					
				}///}}}

			return ( &od_visibilityMap.at(cell).second );

		}///}}}

		bool VisibilityMap::isVisible(const Cell_ID cell) const
		{///{{{
			return ( pt_Geodata->getValueFromBand(GEODATA_VISIBILITY_BAND, (cell/ysize)/2,(cell%ysize)/2) > MIN_VISIBILITY_VALUE ); 
		}///}}}

		int VisibilityMap::getxsize() const
		{//{{{
			return( xsize );
		}//}}}

		int VisibilityMap::getysize() const
		{//{{{
			return( ysize );
		}//}}}

		int VisibilityMap::getviewLength() const
		{//{{{
			return( viewLength );
		}//}}}

		int VisibilityMap::getn() const
		{//{{{
			return( n );
		}//}}}

		double VisibilityMap::getalpha() const
		{//{{{
			return( alpha );
		}//}}}

//}}} ------------------ /Visibility Map Implementation --------------------//

//------------------------------------------------------------------------//
//--------------------------- OUTPUT STREAM  -----------------------------//
//{{{ --------------------------------------------------------------------//

	//	std::ostream& operator<<(std::ostream& out, const VisibilityMap &_visibilityMap);
	
//}}} ---------------------- /OUTPUT STREAM  -----------------------------//

	} // end of namespace lgl
} // end of namespace jafar

